
 !c *************************************************************************************

!C                            APPENDIX 1                          1/15/81
!C
!C        SUBROUTINES FOR SOLVING SPARSE NONSYMMETRIC SYSTEMS
!C        OF LINEAR EQUATIONS  (UNCOMPRESSED POINTER STORAGE)
!C
!C
!C*** SUBROUTINE zndrv
!C*** DRIVER FOR SUBROUTINES FOR SOLVING SPARSE NONSYMMETRIC SYSTEMS OF
!C       LINEAR EQUATIONS (UNCOMPRESSED POINTER STORAGE)
!C
        SUBROUTINE  zndrv
     .     (N, R,C,IC, IA,JA,A, B, Z, NSP,ISP,RSP,ESP, PATH, FLAG)
!C
!C    PARAMETERS
!C    CLASS ABBREVIATIONS ARE --
!C       N - INTEGER VARIABLE
!C       F - REAL VARIABLE
!C       V - SUPPLIES A VALUE TO THE DRIVER
!C       R - RETURNS A RESULT FROM THE DRIVER
!C       I - USED INTERNALLY BY THE DRIVER
!C       A - ARRAY
!C
!C CLASS | PARAMETER
!C ------+----------
!C       |
!C         THE NONZERO ENTRIES OF THE COEFFICIENT MATRIX M ARE STORED
!C    ROW-BY-ROW IN THE ARRAY A.  TO IDENTIFY THE INDIVIDUAL NONZERO
!C    ENTRIES IN EACH ROW, WE NEED TO KNOW IN WHICH COLUMN EACH ENTRY
!C    LIES.  THE COLUMN INDICES WHICH CORRESPOND TO THE NONZERO ENTRIES
!C    OF M ARE STORED IN THE ARRAY JA;  I.E., IF  A(K) = M(I,J),  THEN
!C    JA(K) = J.  IN ADDITION, WE NEED TO KNOW WHERE EACH ROW STARTS AND
!C    HOW LONG IT IS.  THE INDEX POSITIONS IN JA AND A WHERE THE ROWS OF
!C    M BEGIN ARE STORED IN THE ARRAY IA;  I.E., IF M(I,J) IS THE FIRST
!C    NONZERO ENTRY (STORED) IN THE I-TH ROW AND A(K) = M(I,J),  THEN
!C    IA(I) = K.  MOREOVER, THE INDEX IN JA AND A OF THE FIRST LOCATION
!C    FOLLOWING THE LAST ELEMENT IN THE LAST ROW IS STORED IN IA(N+1).
!C    THUS, THE NUMBER OF ENTRIES IN THE I-TH ROW IS GIVEN BY
!C    IA(I+1) - IA(I),  THE NONZERO ENTRIES OF THE I-TH ROW ARE STORED
!C    CONSECUTIVELY IN
!C            A(IA(I)),  A(IA(I)+1),  ..., A(IA(I+1)-1),
!C    AND THE CORRESPONDING COLUMN INDICES ARE STORED CONSECUTIVELY IN
!C            JA(IA(I)), JA(IA(I)+1), ..., JA(IA(I+1)-1).
!C    FOR EXAMPLE, THE 5 BY 5 MATRIX
!C                ( 1. 0. 2. 0. 0.)
!C                ( 0. 3. 0. 0. 0.)
!C            M = ( 0. 4. 5. 6. 0.)
!C                ( 0. 0. 0. 7. 0.)
!C                ( 0. 0. 0. 8. 9.)
!C    WOULD BE STORED AS
!C               | 1  2  3  4  5  6  7  8  9
!C            ---+--------------------------
!C            IA | 1  3  4  7  8 10
!C            JA | 1  3  2  2  3  4  4  4  5
!C             A | 1. 2. 3. 4. 5. 6. 7. 8. 9.         .
!C
!C NV    | N     - NUMBER OF VARIABLES/EQUATIONS.
!C FVA   | A     - NONZERO ENTRIES OF THE COEFFICIENT MATRIX M, STORED
!C       |           BY ROWS.
!C       |           SIZE = NUMBER OF NONZERO ENTRIES IN M.
!C NVA   | IA    - POINTERS TO DELIMIT THE ROWS IN A.
!C       |           SIZE = N+1.
!C NVA   | JA    - COLUMN NUMBERS CORRESPONDING TO THE ELEMENTS OF A.
!C       |           SIZE = SIZE OF A.
!C FVA   | B     - RIGHT-HAND SIDE B;  B AND Z CAN THE SAME ARRAY.
!C       |           SIZE = N.
!C FRA   | Z     - SOLUTION X;  B AND Z CAN BE THE SAME ARRAY.
!C       |           SIZE = N.
!C
!C         THE ROWS AND COLUMNS OF THE ORIGINAL MATRIX M CAN BE
!C    REORDERED (E.G., TO REDUCE FILLIN OR ENSURE NUMERICAL STABILITY)
!C    BEFORE CALLING THE DRIVER.  IF NO REORDERING IS DONE, THEN SET
!C    R(I) = C(I) = IC(I) = I  FOR I=1,...,N.  THE SOLUTION Z IS RETURNED
!C    IN THE ORIGINAL ORDER.
!C
!C NVA   | R     - ORDERING OF THE ROWS OF M.
!C       |           SIZE = N.
!C NVA   | C     - ORDERING OF THE COLUMNS OF M.
!C       |           SIZE = N.
!C NVA   | IC    - INVERSE OF THE ORDERING OF THE COLUMNS OF M;  I.E.,
!C       |           IC(C(I)) = I  FOR I=1,...,N.
!C       |           SIZE = N.
!C
!C         THE SOLUTION OF THE SYSTEM OF LINEAR EQUATIONS IS DIVIDED INTO
!C    THREE STAGES --
!C      znsf -- THE MATRIX M IS PROCESSED SYMBOLICALLY TO DETERMINE WHERE
!C              FILLIN WILL OCCUR DURING THE NUMERIC FACTORIZATION.
!C      znnf -- THE MATRIX M IS FACTORED NUMERICALLY INTO THE PRODUCT LDU
!C              OF A UNIT LOWER TRIANGULAR MATRIX L, A DIAGONAL MATRIX D,
!C              AND A UNIT UPPER TRIANGULAR MATRIX U, AND THE SYSTEM
!C              MX = B  IS SOLVED.
!C      znns -- THE LINEAR SYSTEM  MX = B  IS SOLVED USING THE LDU
!C  OR          FACTORIZATION FROM znnf.
!C      znnt -- THE TRANSPOSED LINEAR SYSTEM  MT X = B  IS SOLVED USING
!C              THE LDU FACTORIZATION FROM znnf.
!C    FOR SEVERAL SYSTEMS WHOSE COEFFICIENT MATRICES HAVE THE SAME
!C    NONZERO STRUCTURE, znsf NEED BE DONE ONLY ONCE (FOR THE FIRST
!C    SYSTEM);  THEN znnf IS DONE ONCE FOR EACH ADDITIONAL SYSTEM.  FOR
!C    SEVERAL SYSTEMS WITH THE SAME COEFFICIENT MATRIX, znsf AND znnf NEED
!C    BE DONE ONLY ONCE (FOR THE FIRST SYSTEM);  THEN znns OR znnt IS DONE
!C    ONCE FOR EACH ADDITIONAL RIGHT-HAND SIDE.
!C
!C NV    | PATH  - PATH SPECIFICATION;  VALUES AND THEIR MEANINGS ARE --
!C       |           1  PERFORM znsf AND znnf.
!C       |           2  PERFORM znnf ONLY  (znsf IS ASSUMED TO HAVE BEEN
!C       |               DONE IN A MANNER COMPATIBLE WITH THE STORAGE
!C       |               ALLOCATION USED IN THE DRIVER).
!C       |           3  PERFORM znns ONLY  (znsf AND znnf ARE ASSUMED TO
!C       |               HAVE BEEN DONE IN A MANNER COMPATIBLE WITH THE
!C       |               STORAGE ALLOCATION USED IN THE DRIVER).
!C       |           4  PERFORM znnt ONLY  (znsf AND znnf ARE ASSUMED TO
!C       |               HAVE BEEN DONE IN A MANNER COMPATIBLE WITH THE
!C       |               STORAGE ALLOCATION USED IN THE DRIVER).
!C       |           5  PERFORM znsf ONLY.
!C
!C         VARIOUS ERRORS ARE DETECTED BY THE DRIVER AND THE INDIVIDUAL
!C    SUBROUTINES.
!C
!C NR    | FLAG  - ERROR FLAG;  VALUES AND THEIR MEANINGS ARE --
!C       |             0     NO ERRORS DETECTED
!C       |             N+K   NULL ROW IN A  --  ROW = K
!C       |            2N+K   DUPLICATE ENTRY IN A  --  ROW = K
!C       |            3N+K   INSUFFICIENT STORAGE IN znsf  --  ROW = K
!C       |            4N+1   INSUFFICIENT STORAGE IN znnf
!C       |            5N+K   NULL PIVOT  --  ROW = K
!C       |            6N+K   INSUFFICIENT STORAGE IN znsf  --  ROW = K
!C       |            7N+1   INSUFFICIENT STORAGE IN znnf
!C       |            8N+K   ZERO PIVOT  --  ROW = K
!C       |           10N+1   INSUFFICIENT STORAGE IN zndrv
!C       |           11N+1   ILLEGAL PATH SPECIFICATION
!C
!C         WORKING STORAGE IS NEEDED FOR THE FACTORED FORM OF THE MATRIX
!C    M PLUS VARIOUS TEMPORARY VECTORS.  THE ARRAYS ISP AND RSP SHOULD BE
!C    EQUIVALENCED;  INTEGER STORAGE IS ALLOCATED FROM THE BEGINNING OF
!C    ISP AND REAL STORAGE FROM THE END OF RSP.
!C
!C NV    | NSP   - DECLARED DIMENSION OF RSP;  NSP GENERALLY MUST
!C       |           BE LARGER THAN  5N+3 + 2K  (WHERE  K = (NUMBER OF
!C       |           NONZERO ENTRIES IN M)).
!C NVIRA | ISP   - INTEGER WORKING STORAGE DIVIDED UP INTO VARIOUS ARRAYS
!C       |           NEEDED BY THE SUBROUTINES;  ISP AND RSP SHOULD BE
!C       |           EQUIVALENCED.
!C       |           SIZE = 4*NSP
!C FVIRA | RSP   - REAL WORKING STORAGE DIVIDED UP INTO VARIOUS ARRAYS
!C       |           NEEDED BY THE SUBROUTINES;  ISP AND RSP SHOULD BE
!C       |           EQUIVALENCED.
!C       |           SIZE = NSP.
!C NR    | ESP   - IF SUFFICIENT STORAGE WAS AVAILABLE TO PERFORM THE
!C       |           SYMBOLIC FACTORIZATION (znsf), THEN ESP IS SET TO THE
!C       |           AMOUNT OF EXCESS STORAGE PROVIDED (NEGATIVE IF
!C       |           INSUFFICIENT STORAGE WAS AVAILABLE TO PERFORM THE
!C       |           NUMERIC FACTORIZATION (znnf)).
!C
!C
!C  CONVERSION TO DOUBLE PRECISION
!C
!C    TO CONVERT THESE ROUTINES FOR DOUBLE PRECISION ARRAYS, SIMPLY USE
!C    THE DOUBLE PRECISION DECLARATIONS IN PLACE OF THE REAL DECLARATIONS
!C    IN EACH SUBPROGRAM;  IN ADDITION, THE DATA VALUE OF THE INTEGER
!C    VARIABLE 4 MUST BE SET AS INDICATED IN SUBROUTINE zndrv
!C
        INTEGER  R(1), C(1), IC(1),  IA(1), JA(1),  ISP(1), ESP,
     *     PATH, FLAG,  Q, IM, D, U, ROW, TMP,  UMAX
        double complex    A(1),  B(1),  Z(1),  RSP(1)
        INTEGER  JU, JUMAX, JUTMP, LMAX, L, J, JLMAX, IL, NSP, N, MAX,
     *           JL, IU
!C
!C  SET 4 EQUAL TO THE RATIO BETWEEN THE LENGTH OF FLOATING POINT
!C  AND INTEGER ARRAY DATA;  E. G., 4 = 1 FOR (REAL, INTEGER),
!C  4 = 2 FOR (DOUBLE PRECISION, INTEGER)
!C
!c       DATA 4/2/
!C
        IF (PATH.LT.1 .OR. 5.LT.PATH)  GO TO 111
!C  ******  INITIALIZE AND DIVIDE UP TEMPORARY STORAGE  *****************
        IL = 1
        IU = IL + N+1
        JL = IU + N+1
!C
!C  ******  CALL znsf IF FLAG IS SET  ************************************
        IF ((PATH-1) * (PATH-5) .NE. 0)  GO TO 2
          MAX = (4*NSP + 1 - JL) - (N+1) - N
          JLMAX = MAX/2
          Q     = JL  + JLMAX
          IM    = Q   + (N+1)
          JUTMP = IM  +   N
          JUMAX = 4*NSP + 1 - JUTMP
          ESP = MAX/4
          IF (JLMAX.LE.0 .OR. JUMAX.LE.0)  GO TO 110
          CALL  znsf
     *       (N,  R, IC,  IA, JA,
     *        ISP(IL), ISP(JL), JLMAX,  ISP(IU), ISP(JUTMP), JUMAX,
     *        ISP(Q),  ISP(IM),  FLAG)
          IF (FLAG.NE.0)  GO TO 100
!C  ******  MOVE JU NEXT TO JL  *****************************************
          JLMAX = ISP(IL+N)-1
          JU    = JL + JLMAX
          JUMAX = ISP(IU+N)-1
          IF (JUMAX.LE.0)  GO TO 2
          DO 1 J=1,JUMAX
   1        ISP(JU+J-1) = ISP(JUTMP+J-1)
!C
!C  ******  CALL REMAINING SUBROUTINES  *********************************
   2    JLMAX = ISP(IL+N)-1
        JU    = JL  + JLMAX
        JUMAX = ISP(IU+N)-1
        L     = (JU + JUMAX - 2 + 4)  /  4    +    1
        LMAX  = JLMAX
        D     = L   + LMAX
        U     = D   + N
        ROW   = NSP + 1 - N
        TMP   = ROW - N
        UMAX  = TMP - U
        ESP = UMAX - JUMAX
!C
        IF ((PATH-1) * (PATH-2) .NE. 0)  GO TO 3
          IF (UMAX.LE.0)  GO TO 110
          CALL  znnf
     *       (N,  R, C, IC,  IA, JA, A,  Z,  B,
     *        ISP(IL), ISP(JL), RSP(L), LMAX,   RSP(D),
     *           ISP(IU), ISP(JU), RSP(U), UMAX,
     *        RSP(ROW),  RSP(TMP),  FLAG)
          IF (FLAG.NE.0)  GO TO 100
          RETURN
!C
   3    IF ((PATH-3) .NE. 0)  GO TO 4
          CALL  znns
     *       (N,  R, C,
     *        ISP(IL), ISP(JL), RSP(L),  RSP(D),
     *          ISP(IU), ISP(JU), RSP(U),
     *        Z,  B,  RSP(TMP))
!C
   4    IF ((PATH-4) .NE. 0)  GO TO 5
          CALL  znnt
     *       (N,  R, C,
     *        ISP(IL), ISP(JL), RSP(L),  RSP(D),
     *          ISP(IU), ISP(JU), RSP(U),
     *        Z,  B,  RSP(TMP))
   5    RETURN
!C
!C ** ERROR:  ERROR DETECTED IN znsf, znnf, znns, OR znnt
 100    RETURN
!C ** ERROR:  INSUFFICIENT STORAGE
 110    FLAG = 10*N + 1
        RETURN
!C ** ERROR:  ILLEGAL PATH SPECIFICATION
 111    FLAG = 11*N + 1
        RETURN
        END
!C
!C       ----------------------------------------------------------------
!C
!C*** SUBROUTINE znnf
!C*** NUMERIC LDU-FACTORIZATION OF SPARSE NONSYMMETRIC MATRIX AND
!C      SOLUTION OF SYSTEM OF LINEAR EQUATIONS (UNCOMPRESSED POINTER
!C      STORAGE)
!C
        SUBROUTINE  znnf
     *     (N, R,C,IC, IA,JA,A, Z, B, IL,JL,L,LMAX, D, IU,JU,U,UMAX,
     *      ROW, TMP, FLAG)
!C
!C       INPUT VARIABLES:   N, R,C,IC, IA,JA,A, B, IL,JL,LMAX, IU,JU,UMAX
!C       OUTPUT VARIABLES:  Z, L,D,U, FLAG
!C
!C       PARAMETERS USED INTERNALLY:
!C FIA   | ROW   - HOLDS INTERMEDIATE VALUES IN CALCULATION OF L, D, U.
!C       |           SIZE = N.
!C FIA   | TMP   - HOLDS NEW RIGHT-HAND SIDE B' FOR SOLUTION OF THE
!C       |           EQUATION  UX = B'.
!C       |           SIZE = N.
!C
        INTEGER  R(1), C(1), IC(1),  IA(1), JA(1),
     *     IL(1), JL(1), LMAX,  IU(1), JU(1), UMAX,  FLAG
        double complex    A(1), Z(1), B(1),  L(1), D(1), U(1),
     *     ROW(1), TMP(1),  LI, SUM, DK
        INTEGER  IMIN, J, I, IMAX, K, N, JMAX, JMIN
!C
!C  ******  CHECK STORAGE  **********************************************
        IF (IL(N+1)-1 .GT. LMAX)  GO TO 104
        IF (IU(N+1)-1 .GT. UMAX)  GO TO 107
!C
!C  ******  FOR EACH ROW  ***********************************************
        DO 10 K=1,N
!C  ******  SET THE INITIAL STRUCTURE OF ROW  ***************************
          JMIN = IL(K)
          JMAX = IL(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 2
!C  ******  IF L(K,M) .NE. 0, ROW(M)=0  *********************************
          DO 1 J=JMIN,JMAX
   1        ROW(JL(J)) = 0
   2      ROW(K) = 0
          JMIN = IU(K)
          JMAX = IU(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 4
!C  ******  IF U(K,M) .NE. 0, ROW(M)=0  *********************************
          DO 3 J=JMIN,JMAX
   3        ROW(JU(J)) = 0
   4      JMIN = IA(R(K))
          JMAX = IA(R(K)+1) - 1
!C  ******  SET ROW TO KTH ROW OF REORDERED A  **************************
          DO 5 J=JMIN,JMAX
   5        ROW(IC(JA(J))) = A(J)
!C  ******  INITIALIZE SUM  *********************************************
          SUM = B(R(K))
!C
!C  ******  ASSIGN THE KTH ROW OF L AND ADJUST ROW, SUM  ****************
          IMIN = IL(K)
          IMAX = IL(K+1) - 1
          IF (IMIN.GT.IMAX)  GO TO 8
          DO 7 I=IMIN,IMAX
            LI = - ROW(JL(I))
!C  ******  IF L IS NOT REQUIRED, THEN COMMENT OUT THE FOLLOWING LINE  **
            L(I) = - LI
            SUM = SUM + LI * TMP(JL(I))
            JMIN = IU(JL(I))
            JMAX = IU(JL(I)+1) - 1
            IF (JMIN.GT.JMAX)  GO TO 7
            DO 6 J=JMIN,JMAX
   6          ROW(JU(J)) = ROW(JU(J)) + LI * U(J)
   7        CONTINUE
!C
!C  ******  ASSIGN DIAGONAL D AND KTH ROW OF U, SET TMP(K)  *************
   8      IF (ROW(K).EQ.0)  GO TO 108
          DK = 1 / ROW(K)
          D(K) = DK
          TMP(K) = SUM * DK
          JMIN = IU(K)
          JMAX = IU(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 10
          DO 9 J=JMIN,JMAX
   9        U(J) = ROW(JU(J)) * DK
  10      CONTINUE
!C
!C  ******  SOLVE  UX = TMP  BY BACK SUBSTITUTION  **********************
        K = N
        DO 13 I=1,N
          SUM = TMP(K)
          JMIN = IU(K)
          JMAX = IU(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 12
          DO 11 J=JMIN,JMAX
  11        SUM = SUM - U(J) * TMP(JU(J))
  12      TMP(K) = SUM
          Z(C(K)) = SUM
  13      K = K-1
!C
        FLAG = 0
        RETURN
!C
!C ** ERROR:  INSUFFICIENT STORAGE FOR L
 104    FLAG = 4*N + 1
        RETURN
!C ** ERROR:  INSUFFICIENT STORAGE FOR U
 107    FLAG = 7*N + 1
        RETURN
!C ** ERROR:  ZERO PIVOT
 108    FLAG = 8*N + K
        RETURN
        END
!C
!C       ----------------------------------------------------------------
!C
!C*** SUBROUTINE znns
!C*** NUMERIC SOLUTION OF A SPARSE NONSYMMETRIC SYSTEM OF LINEAR
!C      EQUATIONS GIVEN LDU-FACTORIZATION (UNCOMPRESSED POINTER STORAGE)
!C
        SUBROUTINE  znns
     *     (N, R,C, IL,JL,L, D, IU,JU,U, Z, B, TMP)
!C
!C       INPUT VARIABLES:   N, R,C, IL,JL,L, D, IU,JU,U, B
!C       OUTPUT VARIABLES:  Z
!C
!C       PARAMETERS USED INTERNALLY:
!C FIA   | TMP   - HOLDS NEW RIGHT-HAND SIDE B' FOR SOLUTION OF THE
!C       |           EQUATION UX = B'.
!C       |           SIZE = N.
!C
        INTEGER  R(1), C(1),  IL(1), JL(1),  IU(1), JU(1)
        double complex    L(1), D(1), U(1),  Z(1), B(1),
     *           TMP(1), SUM
        INTEGER  JMAX, J, I, N, K, JMIN
!C
!C  ******  SOLVE LDY = B  BY FORWARD SUBSTITUTION  *********************
        DO 2 K=1,N
          SUM = B(R(K))
          JMIN = IL(K)
          JMAX = IL(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 2
          DO 1 J=JMIN,JMAX
   1        SUM = SUM - L(J) * TMP(JL(J))
   2      TMP(K) = SUM * D(K)
!C
!C  ******  SOLVE  UX = Y  BY BACK SUBSTITUTION  ************************
        K = N
        DO 5 I=1,N
          SUM = TMP(K)
          JMIN = IU(K)
          JMAX = IU(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 4
          DO 3 J=JMIN,JMAX
   3        SUM = SUM - U(J) * TMP(JU(J))
   4      TMP(K) = SUM
          Z(C(K)) = SUM
   5      K = K-1
        RETURN
        END
!C       ----------------------------------------------------------------
!C
!C*** SUBROUTINE znnt
!C*** NUMERIC SOLUTION OF THE TRANSPOSE OF A SPARSE NONSYMMETRIC SYSTEM
!C      OF LINEAR EQUATIONS GIVEN LDU-FACTORIZATION (UNCOMPRESSED POINTER
!C      STORAGE)
C
        SUBROUTINE  znnt
     *     (N, R,C, IL,JL,L, D, IU,JU,U, Z, B, TMP)
!C
!C       INPUT VARIABLES:   N, R,C, IL,JL,L, D, IU,JU,U, B
!C       OUTPUT VARIABLES:  Z
!C
!C       PARAMETERS USED INTERNALLY:
!C FIA   | TMP   - HOLDS NEW RIGHT-HAND SIDE B' FOR SOLUTION OF THE
!C       |           EQUATION LX = B'.
!C       |           SIZE = N.
!C
        INTEGER  R(1), C(1),  IL(1), JL(1),  IU(1), JU(1)
        double complex    L(1), D(1), U(1),  Z(1), B(1),
     *           TMP(1), TMPK
        INTEGER  JMAX, J, I, N, K, JMIN
!C
!C  ******  SOLVE  UT Y = B  BY FORWARD SUBSTITUTION  *******************
        DO 1 K=1,N
   1      TMP(K) = B(C(K))
        DO 3 K=1,N
          TMPK = - TMP(K)
          JMIN = IU(K)
          JMAX = IU(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 3
          DO 2 J=JMIN,JMAX
   2        TMP(JU(J)) = TMP(JU(J)) + U(J) * TMPK
   3      CONTINUE
!C
!C  ******  SOLVE  D LT X = Y  BY BACK SUBSTITUTION  ********************
        K = N
        DO 6 I=1,N
          TMPK = - TMP(K) * D(K)
          JMIN = IL(K)
          JMAX = IL(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 5
          DO 4 J=JMIN,JMAX
   4        TMP(JL(J)) = TMP(JL(J)) + L(J) * TMPK
   5      Z(R(K)) = - TMPK
   6      K = K-1
        RETURN
        END
!C
!C       ----------------------------------------------------------------
!C
!C               YALE SPARSE MATRIX PACKAGE - NONSYMMETRIC CODES
!C                    SOLVING THE SYSTEM OF EQUATIONS MX = B
!C                        (UNCOMPRESSED POINTER STORAGE)
!C
!C    I.   CALLING SEQUENCES
!C         THE COEFFICIENT MATRIX CAN BE PROCESSED BY AN ORDERING ROUTINE
!C    (E.G., TO REDUCE FILLIN OR ENSURE NUMERICAL STABILITY) BEFORE USING
!C    THE REMAINING SUBROUTINES.  IF NO REORDERING IS DONE, THEN SET
!C    R(I) = C(I) = IC(I) = I  FOR I=1,...,N.  THE CALLING SEQUENCE IS --
!C        (      (MATRIX ORDERING))
!C         znsf   (SYMBOLIC FACTORIZATION TO DETERMINE WHERE FILLIN WILL
!C                 OCCUR DURING NUMERIC FACTORIZATION)
!C         znnf   (NUMERIC FACTORIZATION INTO PRODUCT LDU OF UNIT LOWER
!C                 TRIANGULAR MATRIX L, DIAGONAL MATRIX D, AND UNIT UPPER
!C                 TRIANGULAR MATRIX U, AND SOLUTION OF LINEAR SYSTEM)
!C         znns   (SOLUTION OF LINEAR SYSTEM FOR ADDITIONAL RIGHT-HAND
!C     OR          SIDE USING LDU FACTORIZATION FROM znnf)
!C         znnt   (SOLUTION OF TRANSPOSED LINEAR SYSTEM FOR ADDITIONAL
!C                 RIGHT-HAND SIDE USING LDU FACTORIZATION FROM znnf)
!C
!C    II.  STORAGE OF SPARSE MATRICES
!C         THE NONZERO ENTRIES OF THE COEFFICIENT MATRIX M ARE STORED
!C    ROW-BY-ROW IN THE ARRAY A.  TO IDENTIFY THE INDIVIDUAL NONZERO
!C    ENTRIES IN EACH ROW, WE NEED TO KNOW IN WHICH COLUMN EACH ENTRY
!C    LIES.  THE COLUMN INDICES WHICH CORRESPOND TO THE NONZERO ENTRIES
!C    OF M ARE STORED IN THE ARRAY JA;  I.E., IF  A(K) = M(I,J),  THEN
!C    JA(K) = J.  IN ADDITION, WE NEED TO KNOW WHERE EACH ROW STARTS AND
!C    HOW LONG IT IS.  THE INDEX POSITIONS IN JA AND A WHERE THE ROWS OF
!C    M BEGIN ARE STORED IN THE ARRAY IA;  I.E., IF M(I,J) IS THE FIRST
!C    NONZERO ENTRY (STORED) IN THE I-TH ROW AND A(K) = M(I,J),  THEN
!C    IA(I) = K.  MOREOVER, THE INDEX IN JA AND A OF THE FIRST LOCATION
!C    FOLLOWING THE LAST ELEMENT IN THE LAST ROW IS STORED IN IA(N+1).
!C    THUS, THE NUMBER OF ENTRIES IN THE I-TH ROW IS GIVEN BY
!C    IA(I+1) - IA(I),  THE NONZERO ENTRIES OF THE I-TH ROW ARE STORED
!C    CONSECUTIVELY IN
!C            A(IA(I)),  A(IA(I)+1),  ..., A(IA(I+1)-1),
!C    AND THE CORRESPONDING COLUMN INDICES ARE STORED CONSECUTIVELY IN
!C            JA(IA(I)), JA(IA(I)+1), ..., JA(IA(I+1)-1).
!C    FOR EXAMPLE, THE 5 BY 5 MATRIX
!C                ( 1. 0. 2. 0. 0.)
!C                ( 0. 3. 0. 0. 0.)
!C            M = ( 0. 4. 5. 6. 0.)
!C                ( 0. 0. 0. 7. 0.)
!C                ( 0. 0. 0. 8. 9.)
!C    WOULD BE STORED AS
!C               | 1  2  3  4  5  6  7  8  9
!C            ---+--------------------------
!C            IA | 1  3  4  7  8 10
!C            JA | 1  3  2  2  3  4  4  4  5
!C             A | 1. 2. 3. 4. 5. 6. 7. 8. 9.         .
!C
!C         THE STRICT TRIANGULAR PORTIONS OF THE MATRICES L AND U ARE
!C    STORED IN THE SAME FASHION USING THE ARRAYS  IL, JL, L  AND
!C    IU, JU, U  RESPECTIVELY.  THE DIAGONAL ENTRIES OF L AND U ARE
!C    ASSUMED TO BE EQUAL TO ONE AND ARE NOT STORED.  THE ARRAY D
!C    CONTAINS THE RECIPROCALS OF THE DIAGONAL ENTRIES OF THE MATRIX D.
!C
!C    III. ADDITIONAL STORAGE SAVINGS
!C         IN znsf, R AND IC CAN BE THE SAME ARRAY IN THE CALLING
!C    SEQUENCE IF NO REORDERING OF THE COEFFICIENT MATRIX HAS BEEN DONE.
!C         IN znnf, R, C AND IC CAN ALL BE THE SAME ARRAY IF NO REORDERING
!C    HAS BEEN DONE.  IF ONLY THE ROWS HAVE BEEN REORDERED, THEN C AND IC
!C    CAN BE THE SAME ARRAY.  IF THE ROW AND COLUMN ORDERINGS ARE THE
!C    SAME, THEN R AND C CAN BE THE SAME ARRAY.  Z AND ROW CAN BE THE
!C    SAME ARRAY.
!C         IN znns OR znnt, R AND C CAN BE THE SAME ARRAY IF NO REORDERING
!C    HAS BEEN DONE OR IF THE ROW AND COLUMN ORDERINGS ARE THE SAME.  Z
!C    AND B CAN BE THE SAME ARRAY;  HOWEVER, THEN B WILL BE DESTROYED.
!C
!C    IV.  PARAMETERS
!C         FOLLOWING IS A LIST OF PARAMETERS TO THE PROGRAMS.  NAMES ARE
!C    UNIFORM AMONG THE VARIOUS SUBROUTINES.  CLASS ABBREVIATIONS ARE --
!C       N - INTEGER VARIABLE
!C       F - REAL VARIABLE
!C       V - SUPPLIES A VALUE TO A SUBROUTINE
!C       R - RETURNS A RESULT FROM A SUBROUTINE
!C       I - USED INTERNALLY BY A SUBROUTINE
!C       A - ARRAY
!C
!C CLASS | PARAMETER
!C ------+----------
!C FVA   | A     - NONZERO ENTRIES OF THE COEFFICIENT MATRIX M, STORED
!C       |           BY ROWS.
!C       |           SIZE = NUMBER OF NONZERO ENTRIES IN M.
!C FVA   | B     - RIGHT-HAND SIDE B.
!C       |           SIZE = N.
!C NVA   | C     - ORDERING OF THE COLUMNS OF M.
!C       |           SIZE = N.
!C FVRA  | D     - RECIPROCALS OF THE DIAGONAL ENTRIES OF THE MATRIX D.
!C       |           SIZE = N.
!C NR    | FLAG  - ERROR FLAG;  VALUES AND THEIR MEANINGS ARE --
!C       |            0     NO ERRORS DETECTED
!C       |            N+K   NULL ROW IN A  --  ROW = K
!C       |           3N+K   INSUFFICIENT STORAGE FOR JL  --  ROW = K
!C       |           4N+1   INSUFFICIENT STORAGE FOR L
!C       |           5N+K   NULL PIVOT  --  ROW = K
!C       |           6N+K   INSUFFICIENT STORAGE FOR JU  --  ROW = K
!C       |           7N+1   INSUFFICIENT STORAGE FOR U
!C       |           8N+K   ZERO PIVOT  --  ROW = K
!C NVA   | IA    - POINTERS TO DELIMIT THE ROWS IN A.
!C       |           SIZE = N+1.
!C NVA   | IC    - INVERSE OF THE ORDERING OF THE COLUMNS OF M;  I.E.,
!C       |           IC(C(I) = I  FOR I=1,...N.
!C       |           SIZE = N.
!C NVRA  | IL    - POINTERS TO DELIMIT THE ROWS IN L.
!C       |           SIZE = N+1.
!C NVRA  | IU    - POINTERS TO DELIMIT THE ROWS IN U.
!C       |           SIZE = N+1.
!C NVA   | JA    - COLUMN NUMBERS CORRESPONDING TO THE ELEMENTS OF A.
!C       |           SIZE = SIZE OF A.
!C NVRA  | JL    - COLUMN NUMBERS CORRESPONDING TO THE ELEMENTS OF L.
!C       |           SIZE = JLMAX.
!C NV    | JLMAX - DECLARED DIMENSION OF JL;  JLMAX MUST BE LARGER THAN
!C       |           THE NUMBER OF NONZERO ENTRIES IN THE STRICT LOWER
!C       |           TRIANGLE OF M PLUS FILLIN  (IL(N+1)-1 AFTER znsf).
!C NVRA  | JU    - COLUMN NUMBERS CORRESPONDING TO THE ELEMENTS OF U.
!C       |           SIZE = JUMAX.
!C NV    | JUMAX - DECLARED DIMENSION OF JU;  JUMAX MUST BE LARGER THAN
!C       |           THE NUMBER OF NONZERO ENTRIES IN THE STRICT UPPER
!C       |           TRIANGLE OF M PLUS FILLIN  (IU(N+1)-1 AFTER znsf).
!C FVRA  | L     - NONZERO ENTRIES IN THE STRICT LOWER TRIANGULAR PORTION
!C       |           OF THE MATRIX L, STORED BY ROWS.
!C       |           SIZE = LMAX
!C NV    | LMAX  - DECLARED DIMENSION OF L;  LMAX MUST BE LARGER THAN
!C       |           THE NUMBER OF NONZERO ENTRIES IN THE STRICT LOWER
!C       |           TRIANGLE OF M PLUS FILLIN  (IL(N+1)-1 AFTER znsf).
!C NV    | N     - NUMBER OF VARIABLES/EQUATIONS.
!C NVA   | R     - ORDERING OF THE ROWS OF M.
!C       |           SIZE = N.
!C FVRA  | U     - NONZERO ENTRIES IN THE STRICT UPPER TRIANGULAR PORTION
!C       |           OF THE MATRIX U, STORED BY ROWS.
!C       |           SIZE = UMAX.
!C NV    | UMAX  - DECLARED DIMENSION OF U;  UMAX MUST BE LARGER THAN
!C       |           THE NUMBER OF NONZERO ENTRIES IN THE STRICT UPPER
!C       |           TRIANGLE OF M PLUS FILLIN  (IU(N+1)-1 AFTER znsf).
!C FRA   | Z     - SOLUTION X.
!C       |           SIZE = N.
!C
!C
!C       ----------------------------------------------------------------
!C
!C*** SUBROUTINE znsf
!C*** SYMBOLIC LDU-FACTORIZATION OF A NONSYMMETRIC SPARSE MATRIX
!C      (UNCOMPRESSED POINTER STORAGE)
!C
        SUBROUTINE  znsf
     *     (N, R,IC, IA,JA, IL,JL,JLMAX, IU,JU,JUMAX, Q, IM, FLAG)
!C
!C       INPUT VARIABLES:   N, R,IC, IA,JA, JLMAX, JUMAX.
!C       OUTPUT VARIABLES:  IL,JL, IU,JU, FLAG.
!C
!C       PARAMETERS USED INTERNALLY:
!C NIA   | Q     - SUPPOSE M' IS THE RESULT OF REORDERING M;  IF
!C       |           PROCESSING OF THE KTH ROW OF M' (HENCE THE KTH ROWS
!C       |           OF L AND U) IS BEING DONE, THEN Q(J) IS INITIALLY
!C       |           NONZERO IF M'(K,J) IS NONZERO;  SINCE VALUES NEED
!C       |           NOT BE STORED, EACH ENTRY POINTS TO THE NEXT
!C       |           NONZERO;  FOR EXAMPLE, IF  N=9  AND THE 5TH ROW OF
!C       |           M' IS
!C       |                   0 X X 0 X 0 0 X 0,
!C       |           THEN Q WILL INITIALLY BE
!C       |                   A 3 5 A 8 A A 10 A 2        (A - ARBITRARY);
!C       |           Q(N+1) POINTS TO THE FIRST NONZERO IN THE ROW AND
!C       |           THE LAST NONZERO POINTS TO  N+1;  AS THE ALGORITHM
!C       |           PROCEEDS, OTHER ELEMENTS OF Q ARE INSERTED IN THE
!C       |           LIST BECAUSE OF FILLIN.
!C       |           SIZE = N+1.
!C NIA   | IM    - AT EACH STEP IN THE FACTORIZATION, IM(I) IS THE LAST
!C       |           ELEMENT IN THE ITH ROW OF U WHICH NEEDS TO BE
!C       |           CONSIDERED IN COMPUTING FILLIN.
!C       |           SIZE = N.
!C
!C  INTERNAL VARIABLES--
!C    JLPTR - POINTS TO THE LAST POSITION USED IN  JL.
!C    JUPTR - POINTS TO THE LAST POSITION USED IN  JU.
!C
        INTEGER  R(1), IC(1),  IA(1), JA(1),  IL(1), JL(1),
     *     IU(1), JU(1),  Q(1),  IM(1),  FLAG,  QM, VJ
        INTEGER  JMAX, JMIN, J, I, M, K, JLMAX, N, JUMAX, JUPTR, JLPTR
!C
!C  ******  INITIALIZE POINTERS  ****************************************
        JLPTR = 0
        IL(1) = 1
        JUPTR = 0
        IU(1) = 1
!C
!C  ******  FOR EACH ROW OF L AND U  ************************************
        DO 10 K=1,N
!C  ******  SET Q TO THE REORDERED ROW OF A  ****************************
          Q(N+1) = N+1
          JMIN = IA(R(K))
          JMAX = IA(R(K)+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 101
          DO 2 J=JMIN,JMAX
            VJ = IC(JA(J))
            QM = N+1
   1        M = QM
            QM = Q(M)
            IF (QM.LT.VJ)  GO TO 1
            IF (QM.EQ.VJ)  GO TO 102
              Q(M) = VJ
              Q(VJ) = QM
   2        CONTINUE
!C
!C  ******  FOR EACH ENTRY IN THE LOWER TRIANGLE  ***********************
          I = N+1
   3      I = Q(I)
          IF (I.GE.K)  GO TO 7
!C  ******  L(K,I) WILL BE NONZERO, SO ADD IT TO JL  ********************
            JLPTR = JLPTR+1
            IF (JLPTR.GT.JLMAX)  GO TO 103
            JL(JLPTR) = I
            QM = I
!C  ******  INSPECT ITH ROW FOR FILLIN, ADJUST IM IF POSSIBLE  **********
            JMIN = IU(I)
            JMAX = IM(I)
            IF (JMIN.GT.JMAX)  GO TO 6
            DO 5 J=JMIN,JMAX
              VJ = JU(J)
              IF (VJ.EQ.K)  IM(I) = J
   4          M = QM
              QM = Q(M)
              IF (QM.LT.VJ)  GO TO 4
              IF (QM.EQ.VJ)  GO TO 5
                Q(M) = VJ
                Q(VJ) = QM
                QM = VJ
   5          CONTINUE
   6        GO TO 3
!C
!C  ******  CHECK FOR NULL PIVOT  ***************************************
   7      IF (I.NE.K)  GO TO 105
!C  ******  REMAINING ELEMENTS OF Q DEFINE STRUCTURE OF U(K, )  *********
   8      I = Q(I)
          IF (I.GT.N)  GO TO 9
            JUPTR = JUPTR+1
            IF (JUPTR.GT.JUMAX)  GO TO 106
            JU(JUPTR) = I
            GO TO 8
!C  ******  GET READY FOR NEXT ROW  *************************************
   9      IM(K) = JUPTR
          IL(K+1) = JLPTR+1
  10      IU(K+1) = JUPTR+1
!C
        FLAG = 0
        RETURN
!C
!C ** ERROR:  NULL ROW IN A
 101    FLAG = N + R(K)
        RETURN
!C ** ERROR:  DUPLICATE ENTRY IN A
 102    FLAG = 2*N + R(K)
        RETURN
!C ** ERROR:  INSUFFICIENT STORAGE FOR JL
 103    FLAG = 3*N + K
        RETURN
!C ** ERROR:  NULL PIVOT
 105    FLAG = 5*N + K
        RETURN
!C ** ERROR:  INSUFFICIENT STORAGE FOR JU
 106    FLAG = 6*N + K
        RETURN
        END
!C                            APPENDIX 1                          1/15/81
!C
!C        SUBROUTINES FOR SOLVING SPARSE NONSYMMETRIC SYSTEMS
!C        OF LINEAR EQUATIONS  (UNCOMPRESSED POINTER STORAGE)
!C
!C
!C*** SUBROUTINE cndrv
!C*** DRIVER FOR SUBROUTINES FOR SOLVING SPARSE NONSYMMETRIC SYSTEMS OF
!C       LINEAR EQUATIONS (UNCOMPRESSED POINTER STORAGE)
!C
        SUBROUTINE  cndrv
     *     (N, R,C,IC, IA,JA,A, B, Z, NSP,ISP,RSP,ESP, PATH, FLAG)
!C
!C    PARAMETERS
!C    CLASS ABBREVIATIONS ARE --
!C       N - INTEGER VARIABLE
!C       F - REAL VARIABLE
!C       V - SUPPLIES A VALUE TO THE DRIVER
!C       R - RETURNS A RESULT FROM THE DRIVER
!C       I - USED INTERNALLY BY THE DRIVER
!C       A - ARRAY
!C
!C CLASS | PARAMETER
!C ------+----------
!C       |
!C         THE NONZERO ENTRIES OF THE COEFFICIENT MATRIX M ARE STORED
!C    ROW-BY-ROW IN THE ARRAY A.  TO IDENTIFY THE INDIVIDUAL NONZERO
!C    ENTRIES IN EACH ROW, WE NEED TO KNOW IN WHICH COLUMN EACH ENTRY
!C    LIES.  THE COLUMN INDICES WHICH CORRESPOND TO THE NONZERO ENTRIES
!C    OF M ARE STORED IN THE ARRAY JA;  I.E., IF  A(K) = M(I,J),  THEN
!C    JA(K) = J.  IN ADDITION, WE NEED TO KNOW WHERE EACH ROW STARTS AND
!C    HOW LONG IT IS.  THE INDEX POSITIONS IN JA AND A WHERE THE ROWS OF
!C    M BEGIN ARE STORED IN THE ARRAY IA;  I.E., IF M(I,J) IS THE FIRST
!C    NONZERO ENTRY (STORED) IN THE I-TH ROW AND A(K) = M(I,J),  THEN
!C    IA(I) = K.  MOREOVER, THE INDEX IN JA AND A OF THE FIRST LOCATION
!C    FOLLOWING THE LAST ELEMENT IN THE LAST ROW IS STORED IN IA(N+1).
!C    THUS, THE NUMBER OF ENTRIES IN THE I-TH ROW IS GIVEN BY
!C    IA(I+1) - IA(I),  THE NONZERO ENTRIES OF THE I-TH ROW ARE STORED
!C    CONSECUTIVELY IN
!C            A(IA(I)),  A(IA(I)+1),  ..., A(IA(I+1)-1),
!C    AND THE CORRESPONDING COLUMN INDICES ARE STORED CONSECUTIVELY IN
!C            JA(IA(I)), JA(IA(I)+1), ..., JA(IA(I+1)-1).
!C    FOR EXAMPLE, THE 5 BY 5 MATRIX
!C                ( 1. 0. 2. 0. 0.)
!C                ( 0. 3. 0. 0. 0.)
!C            M = ( 0. 4. 5. 6. 0.)
!C                ( 0. 0. 0. 7. 0.)
!C                ( 0. 0. 0. 8. 9.)
!C    WOULD BE STORED AS
!C               | 1  2  3  4  5  6  7  8  9
!C            ---+--------------------------
!C            IA | 1  3  4  7  8 10
!C            JA | 1  3  2  2  3  4  4  4  5
!C             A | 1. 2. 3. 4. 5. 6. 7. 8. 9.         .
!C
!C NV    | N     - NUMBER OF VARIABLES/EQUATIONS.
!C FVA   | A     - NONZERO ENTRIES OF THE COEFFICIENT MATRIX M, STORED
!C       |           BY ROWS.
!C       |           SIZE = NUMBER OF NONZERO ENTRIES IN M.
!C NVA   | IA    - POINTERS TO DELIMIT THE ROWS IN A.
!C       |           SIZE = N+1.
!C NVA   | JA    - COLUMN NUMBERS CORRESPONDING TO THE ELEMENTS OF A.
!C       |           SIZE = SIZE OF A.
!C FVA   | B     - RIGHT-HAND SIDE B;  B AND Z CAN THE SAME ARRAY.
!C       |           SIZE = N.
!C FRA   | Z     - SOLUTION X;  B AND Z CAN BE THE SAME ARRAY.
!C       |           SIZE = N.
!C
!C         THE ROWS AND COLUMNS OF THE ORIGINAL MATRIX M CAN BE
!C    REORDERED (E.G., TO REDUCE FILLIN OR ENSURE NUMERICAL STABILITY)
!C    BEFORE CALLING THE DRIVER.  IF NO REORDERING IS DONE, THEN SET
!C    R(I) = C(I) = IC(I) = I  FOR I=1,...,N.  THE SOLUTION Z IS RETURNED
!C    IN THE ORIGINAL ORDER.
!C
!C NVA   | R     - ORDERING OF THE ROWS OF M.
!C       |           SIZE = N.
!C NVA   | C     - ORDERING OF THE COLUMNS OF M.
!C       |           SIZE = N.
!C NVA   | IC    - INVERSE OF THE ORDERING OF THE COLUMNS OF M;  I.E.,
!C       |           IC(C(I)) = I  FOR I=1,...,N.
!C       |           SIZE = N.
!C
!C         THE SOLUTION OF THE SYSTEM OF LINEAR EQUATIONS IS DIVIDED INTO
!C    THREE STAGES --
!C      cnsf -- THE MATRIX M IS PROCESSED SYMBOLICALLY TO DETERMINE WHERE
!C              FILLIN WILL OCCUR DURING THE NUMERIC FACTORIZATION.
!C      cnnf -- THE MATRIX M IS FACTORED NUMERICALLY INTO THE PRODUCT LDU
!C              OF A UNIT LOWER TRIANGULAR MATRIX L, A DIAGONAL MATRIX D,
!C              AND A UNIT UPPER TRIANGULAR MATRIX U, AND THE SYSTEM
!C              MX = B  IS SOLVED.
!C      cnns -- THE LINEAR SYSTEM  MX = B  IS SOLVED USING THE LDU
!C  OR          FACTORIZATION FROM cnnf.
!C      cnnt -- THE TRANSPOSED LINEAR SYSTEM  MT X = B  IS SOLVED USING
!C              THE LDU FACTORIZATION FROM cnnf.
!C    FOR SEVERAL SYSTEMS WHOSE COEFFICIENT MATRICES HAVE THE SAME
!C    NONZERO STRUCTURE, cnsf NEED BE DONE ONLY ONCE (FOR THE FIRST
!C    SYSTEM);  THEN cnnf IS DONE ONCE FOR EACH ADDITIONAL SYSTEM.  FOR
!C    SEVERAL SYSTEMS WITH THE SAME COEFFICIENT MATRIX, cnsf AND cnnf NEED
!C    BE DONE ONLY ONCE (FOR THE FIRST SYSTEM);  THEN cnns OR cnnt IS DONE
!C    ONCE FOR EACH ADDITIONAL RIGHT-HAND SIDE.
!C
!C NV    | PATH  - PATH SPECIFICATION;  VALUES AND THEIR MEANINGS ARE --
!C       |           1  PERFORM cnsf AND cnnf.
!C       |           2  PERFORM cnnf ONLY  (cnsf IS ASSUMED TO HAVE BEEN
!C       |               DONE IN A MANNER COMPATIBLE WITH THE STORAGE
!C       |               ALLOCATION USED IN THE DRIVER).
!C       |           3  PERFORM cnns ONLY  (cnsf AND cnnf ARE ASSUMED TO
!C       |               HAVE BEEN DONE IN A MANNER COMPATIBLE WITH THE
!C       |               STORAGE ALLOCATION USED IN THE DRIVER).
!C       |           4  PERFORM cnnt ONLY  (cnsf AND cnnf ARE ASSUMED TO
!C       |               HAVE BEEN DONE IN A MANNER COMPATIBLE WITH THE
!C       |               STORAGE ALLOCATION USED IN THE DRIVER).
!C       |           5  PERFORM cnsf ONLY.
!C
!C         VARIOUS ERRORS ARE DETECTED BY THE DRIVER AND THE INDIVIDUAL
!C    SUBROUTINES.
!C
!C NR    | FLAG  - ERROR FLAG;  VALUES AND THEIR MEANINGS ARE --
!C       |             0     NO ERRORS DETECTED
!C       |             N+K   NULL ROW IN A  --  ROW = K
!C       |            2N+K   DUPLICATE ENTRY IN A  --  ROW = K
!C       |            3N+K   INSUFFICIENT STORAGE IN cnsf  --  ROW = K
!C       |            4N+1   INSUFFICIENT STORAGE IN cnnf
!C       |            5N+K   NULL PIVOT  --  ROW = K
!C       |            6N+K   INSUFFICIENT STORAGE IN cnsf  --  ROW = K
!C       |            7N+1   INSUFFICIENT STORAGE IN cnnf
!C       |            8N+K   ZERO PIVOT  --  ROW = K
!C       |           10N+1   INSUFFICIENT STORAGE IN cndrv
!C       |           11N+1   ILLEGAL PATH SPECIFICATION
!C
!C         WORKING STORAGE IS NEEDED FOR THE FACTORED FORM OF THE MATRIX
!C    M PLUS VARIOUS TEMPORARY VECTORS.  THE ARRAYS ISP AND RSP SHOULD BE
!C    EQUIVALENCED;  INTEGER STORAGE IS ALLOCATED FROM THE BEGINNING OF
!C    ISP AND REAL STORAGE FROM THE END OF RSP.
!C
!C NV    | NSP   - DECLARED DIMENSION OF RSP;  NSP GENERALLY MUST
!C       |           BE LARGER THAN  5N+3 + 2K  (WHERE  K = (NUMBER OF
!C       |           NONZERO ENTRIES IN M)).
!C NVIRA | ISP   - INTEGER WORKING STORAGE DIVIDED UP INTO VARIOUS ARRAYS
!C       |           NEEDED BY THE SUBROUTINES;  ISP AND RSP SHOULD BE
!C       |           EQUIVALENCED.
!C       |           SIZE = 2*NSP
!C FVIRA | RSP   - REAL WORKING STORAGE DIVIDED UP INTO VARIOUS ARRAYS
!C       |           NEEDED BY THE SUBROUTINES;  ISP AND RSP SHOULD BE
!C       |           EQUIVALENCED.
!C       |           SIZE = NSP.
!C NR    | ESP   - IF SUFFICIENT STORAGE WAS AVAILABLE TO PERFORM THE
!C       |           SYMBOLIC FACTORIZATION (cnsf), THEN ESP IS SET TO THE
!C       |           AMOUNT OF EXCESS STORAGE PROVIDED (NEGATIVE IF
!C       |           INSUFFICIENT STORAGE WAS AVAILABLE TO PERFORM THE
!C       |           NUMERIC FACTORIZATION (cnnf)).
!C
!C
!C  CONVERSION TO DOUBLE PRECISION
!C
!C    TO CONVERT THESE ROUTINES FOR DOUBLE PRECISION ARRAYS, SIMPLY USE
!C    THE DOUBLE PRECISION DECLARATIONS IN PLACE OF THE REAL DECLARATIONS
!C    IN EACH SUBPROGRAM;  IN ADDITION, THE DATA VALUE OF THE INTEGER
!C    VARIABLE 2 MUST BE SET AS INDICATED IN SUBROUTINE cndrv
!C
        INTEGER  R(1), C(1), IC(1),  IA(1), JA(1),  ISP(1), ESP,
     *     PATH, FLAG,  Q, IM, D, U, ROW, TMP,  UMAX
        double complex    A(1),  B(1),  Z(1),  RSP(1)
        INTEGER  JU, JUMAX, JUTMP, LMAX, L, J, JLMAX, IL, NSP, N, MAX,
     *           JL, IU
!C
!C  SET 2 EQUAL TO THE RATIO BETWEEN THE LENGTH OF FLOATING POINT
!C  AND INTEGER ARRAY DATA;  E. G., 2 = 1 FOR (REAL, INTEGER),
!C  2 = 2 FOR (DOUBLE PRECISION, INTEGER)
!C
!c       DATA 2/2/
!!C
        IF (PATH.LT.1 .OR. 5.LT.PATH)  GO TO 111
!C  ******  INITIALIZE AND DIVIDE UP TEMPORARY STORAGE  *****************
        IL = 1
        IU = IL + N+1
        JL = IU + N+1
!C
!C  ******  CALL cnsf IF FLAG IS SET  ************************************
        IF ((PATH-1) * (PATH-5) .NE. 0)  GO TO 2
          MAX = (2*NSP + 1 - JL) - (N+1) - N
          JLMAX = MAX/2
          Q     = JL  + JLMAX
          IM    = Q   + (N+1)
          JUTMP = IM  +   N
          JUMAX = 2*NSP + 1 - JUTMP
          ESP = MAX/2
          IF (JLMAX.LE.0 .OR. JUMAX.LE.0)  GO TO 110
          CALL  cnsf
     *       (N,  R, IC,  IA, JA,
     *        ISP(IL), ISP(JL), JLMAX,  ISP(IU), ISP(JUTMP), JUMAX,
     *        ISP(Q),  ISP(IM),  FLAG)
          IF (FLAG.NE.0)  GO TO 100
!C  ******  MOVE JU NEXT TO JL  *****************************************
          JLMAX = ISP(IL+N)-1
          JU    = JL + JLMAX
          JUMAX = ISP(IU+N)-1
          IF (JUMAX.LE.0)  GO TO 2
          DO 1 J=1,JUMAX
   1        ISP(JU+J-1) = ISP(JUTMP+J-1)
!C
!C  ******  CALL REMAINING SUBROUTINES  *********************************
   2    JLMAX = ISP(IL+N)-1
        JU    = JL  + JLMAX
        JUMAX = ISP(IU+N)-1
        L     = (JU + JUMAX - 2 + 2)  /  2    +    1
        LMAX  = JLMAX
        D     = L   + LMAX
        U     = D   + N
        ROW   = NSP + 1 - N
        TMP   = ROW - N
        UMAX  = TMP - U
        ESP = UMAX - JUMAX
!C
        IF ((PATH-1) * (PATH-2) .NE. 0)  GO TO 3
          IF (UMAX.LE.0)  GO TO 110
          CALL  cnnf
     *       (N,  R, C, IC,  IA, JA, A,  Z,  B,
     *        ISP(IL), ISP(JL), RSP(L), LMAX,   RSP(D),
     *           ISP(IU), ISP(JU), RSP(U), UMAX,
     *        RSP(ROW),  RSP(TMP),  FLAG)
          IF (FLAG.NE.0)  GO TO 100
          RETURN
!C
   3    IF ((PATH-3) .NE. 0)  GO TO 4
          CALL  cnns
     *       (N,  R, C,
     *        ISP(IL), ISP(JL), RSP(L),  RSP(D),
     *          ISP(IU), ISP(JU), RSP(U),
     *        Z,  B,  RSP(TMP))
!C
   4    IF ((PATH-4) .NE. 0)  GO TO 5
          CALL  cnnt
     *       (N,  R, C,
     *        ISP(IL), ISP(JL), RSP(L),  RSP(D),
     *          ISP(IU), ISP(JU), RSP(U),
     *        Z,  B,  RSP(TMP))
   5    RETURN
!C
!C ** ERROR:  ERROR DETECTED IN cnsf, cnnf, cnns, OR cnnt
 100    RETURN
!C ** ERROR:  INSUFFICIENT STORAGE
 110    FLAG = 10*N + 1
        RETURN
!C ** ERROR:  ILLEGAL PATH SPECIFICATION
 111    FLAG = 11*N + 1
        RETURN
        END
!C
!C       ----------------------------------------------------------------
!C
!C*** SUBROUTINE cnnf
!C*** NUMERIC LDU-FACTORIZATION OF SPARSE NONSYMMETRIC MATRIX AND
!C      SOLUTION OF SYSTEM OF LINEAR EQUATIONS (UNCOMPRESSED POINTER
!C      STORAGE)
!C
        SUBROUTINE  cnnf
     *     (N, R,C,IC, IA,JA,A, Z, B, IL,JL,L,LMAX, D, IU,JU,U,UMAX,
     *      ROW, TMP, FLAG)
!C
!C       INPUT VARIABLES:   N, R,C,IC, IA,JA,A, B, IL,JL,LMAX, IU,JU,UMAX
!C       OUTPUT VARIABLES:  Z, L,D,U, FLAG
!C
!C       PARAMETERS USED INTERNALLY:
!C FIA   | ROW   - HOLDS INTERMEDIATE VALUES IN CALCULATION OF L, D, U.
!C       |           SIZE = N.
!C FIA   | TMP   - HOLDS NEW RIGHT-HAND SIDE B' FOR SOLUTION OF THE
!C       |           EQUATION  UX = B'.
!C       |           SIZE = N.
!C
        INTEGER  R(1), C(1), IC(1),  IA(1), JA(1),
     *     IL(1), JL(1), LMAX,  IU(1), JU(1), UMAX,  FLAG
        double complex    A(1), Z(1), B(1),  L(1), D(1), U(1),
     *     ROW(1), TMP(1),  LI, SUM, DK
        INTEGER  IMIN, J, I, IMAX, K, N, JMAX, JMIN
!C
!C  ******  CHECK STORAGE  **********************************************
        IF (IL(N+1)-1 .GT. LMAX)  GO TO 104
        IF (IU(N+1)-1 .GT. UMAX)  GO TO 107
!C
!C  ******  FOR EACH ROW  ***********************************************
        DO 10 K=1,N
!C  ******  SET THE INITIAL STRUCTURE OF ROW  ***************************
          JMIN = IL(K)
          JMAX = IL(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 2
!C  ******  IF L(K,M) .NE. 0, ROW(M)=0  *********************************
          DO 1 J=JMIN,JMAX
   1        ROW(JL(J)) = 0
   2      ROW(K) = 0
          JMIN = IU(K)
          JMAX = IU(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 4
!C  ******  IF U(K,M) .NE. 0, ROW(M)=0  *********************************
          DO 3 J=JMIN,JMAX
   3        ROW(JU(J)) = 0
   4      JMIN = IA(R(K))
          JMAX = IA(R(K)+1) - 1
!C  ******  SET ROW TO KTH ROW OF REORDERED A  **************************
          DO 5 J=JMIN,JMAX
   5        ROW(IC(JA(J))) = A(J)
!C  ******  INITIALIZE SUM  *********************************************
          SUM = B(R(K))
!C
!C  ******  ASSIGN THE KTH ROW OF L AND ADJUST ROW, SUM  ****************
          IMIN = IL(K)
          IMAX = IL(K+1) - 1
          IF (IMIN.GT.IMAX)  GO TO 8
          DO 7 I=IMIN,IMAX
            LI = - ROW(JL(I))
!C  ******  IF L IS NOT REQUIRED, THEN COMMENT OUT THE FOLLOWING LINE  **
            L(I) = - LI
            SUM = SUM + LI * TMP(JL(I))
            JMIN = IU(JL(I))
            JMAX = IU(JL(I)+1) - 1
            IF (JMIN.GT.JMAX)  GO TO 7
            DO 6 J=JMIN,JMAX
   6          ROW(JU(J)) = ROW(JU(J)) + LI * U(J)
   7        CONTINUE
!C
!C  ******  ASSIGN DIAGONAL D AND KTH ROW OF U, SET TMP(K)  *************
   8      IF (ROW(K).EQ.0)  GO TO 108
          DK = 1 / ROW(K)
          D(K) = DK
          TMP(K) = SUM * DK
          JMIN = IU(K)
          JMAX = IU(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 10
          DO 9 J=JMIN,JMAX
   9        U(J) = ROW(JU(J)) * DK
  10      CONTINUE
!C
!C  ******  SOLVE  UX = TMP  BY BACK SUBSTITUTION  **********************
        K = N
        DO 13 I=1,N
          SUM = TMP(K)
          JMIN = IU(K)
          JMAX = IU(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 12
          DO 11 J=JMIN,JMAX
  11        SUM = SUM - U(J) * TMP(JU(J))
  12      TMP(K) = SUM
          Z(C(K)) = SUM
  13      K = K-1
!C
        FLAG = 0
        RETURN
!C
!C ** ERROR:  INSUFFICIENT STORAGE FOR L
 104    FLAG = 4*N + 1
        RETURN
!C ** ERROR:  INSUFFICIENT STORAGE FOR U
 107    FLAG = 7*N + 1
        RETURN
!C ** ERROR:  ZERO PIVOT
 108    FLAG = 8*N + K
        RETURN
        END
!C
!C       ----------------------------------------------------------------
!C
!C*** SUBROUTINE cnns
!C*** NUMERIC SOLUTION OF A SPARSE NONSYMMETRIC SYSTEM OF LINEAR
!C      EQUATIONS GIVEN LDU-FACTORIZATION (UNCOMPRESSED POINTER STORAGE)
!C
        SUBROUTINE  cnns
     *     (N, R,C, IL,JL,L, D, IU,JU,U, Z, B, TMP)
!C
!C       INPUT VARIABLES:   N, R,C, IL,JL,L, D, IU,JU,U, B
!C       OUTPUT VARIABLES:  Z
!C
!C       PARAMETERS USED INTERNALLY:
!C FIA   | TMP   - HOLDS NEW RIGHT-HAND SIDE B' FOR SOLUTION OF THE
!C       |           EQUATION UX = B'.
!C       |           SIZE = N.
!C
        INTEGER  R(1), C(1),  IL(1), JL(1),  IU(1), JU(1)
        double complex    L(1), D(1), U(1),  Z(1), B(1),
     *           TMP(1), SUM
        INTEGER  JMAX, J, I, N, K, JMIN
!C
!C  ******  SOLVE LDY = B  BY FORWARD SUBSTITUTION  *********************
        DO 2 K=1,N
          SUM = B(R(K))
          JMIN = IL(K)
          JMAX = IL(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 2
          DO 1 J=JMIN,JMAX
   1        SUM = SUM - L(J) * TMP(JL(J))
   2      TMP(K) = SUM * D(K)
!C
!C  ******  SOLVE  UX = Y  BY BACK SUBSTITUTION  ************************
        K = N
        DO 5 I=1,N
          SUM = TMP(K)
          JMIN = IU(K)
          JMAX = IU(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 4
          DO 3 J=JMIN,JMAX
   3        SUM = SUM - U(J) * TMP(JU(J))
   4      TMP(K) = SUM
          Z(C(K)) = SUM
   5      K = K-1
        RETURN
        END
!C       ----------------------------------------------------------------
!C
!C*** SUBROUTINE cnnt
!C*** NUMERIC SOLUTION OF THE TRANSPOSE OF A SPARSE NONSYMMETRIC SYSTEM
!C      OF LINEAR EQUATIONS GIVEN LDU-FACTORIZATION (UNCOMPRESSED POINTER
!C      STORAGE)
!C
        SUBROUTINE  cnnt
     *     (N, R,C, IL,JL,L, D, IU,JU,U, Z, B, TMP)
!C
!C       INPUT VARIABLES:   N, R,C, IL,JL,L, D, IU,JU,U, B
!C       OUTPUT VARIABLES:  Z
!C
!C       PARAMETERS USED INTERNALLY:
!C FIA   | TMP   - HOLDS NEW RIGHT-HAND SIDE B' FOR SOLUTION OF THE
!C       |           EQUATION LX = B'.
!C       |           SIZE = N.
!C
        INTEGER  R(1), C(1),  IL(1), JL(1),  IU(1), JU(1)
        complex*16    L(1), D(1), U(1),  Z(1), B(1),
     *           TMP(1), TMPK
        INTEGER  JMAX, J, I, N, K, JMIN
!C
!C  ******  SOLVE  UT Y = B  BY FORWARD SUBSTITUTION  *******************
        DO 1 K=1,N
   1      TMP(K) = B(C(K))
        DO 3 K=1,N
          TMPK = - TMP(K)
          JMIN = IU(K)
          JMAX = IU(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 3
          DO 2 J=JMIN,JMAX
   2        TMP(JU(J)) = TMP(JU(J)) + U(J) * TMPK
   3      CONTINUE
!C
!C  ******  SOLVE  D LT X = Y  BY BACK SUBSTITUTION  ********************
        K = N
        DO 6 I=1,N
          TMPK = - TMP(K) * D(K)
          JMIN = IL(K)
          JMAX = IL(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 5
          DO 4 J=JMIN,JMAX
   4        TMP(JL(J)) = TMP(JL(J)) + L(J) * TMPK
   5      Z(R(K)) = - TMPK
   6      K = K-1
        RETURN
        END
!C
!C       ----------------------------------------------------------------
!C
!C               YALE SPARSE MATRIX PACKAGE - NONSYMMETRIC CODES
!C                    SOLVING THE SYSTEM OF EQUATIONS MX = B
!C                        (UNCOMPRESSED POINTER STORAGE)
!C
!C    I.   CALLING SEQUENCES
!C         THE COEFFICIENT MATRIX CAN BE PROCESSED BY AN ORDERING ROUTINE
!C    (E.G., TO REDUCE FILLIN OR ENSURE NUMERICAL STABILITY) BEFORE USING
!C    THE REMAINING SUBROUTINES.  IF NO REORDERING IS DONE, THEN SET
!C    R(I) = C(I) = IC(I) = I  FOR I=1,...,N.  THE CALLING SEQUENCE IS --
!C        (      (MATRIX ORDERING))
!C         cnsf   (SYMBOLIC FACTORIZATION TO DETERMINE WHERE FILLIN WILL
!C                 OCCUR DURING NUMERIC FACTORIZATION)
!C         cnnf   (NUMERIC FACTORIZATION INTO PRODUCT LDU OF UNIT LOWER
!C                 TRIANGULAR MATRIX L, DIAGONAL MATRIX D, AND UNIT UPPER
!C                 TRIANGULAR MATRIX U, AND SOLUTION OF LINEAR SYSTEM)
!C         cnns   (SOLUTION OF LINEAR SYSTEM FOR ADDITIONAL RIGHT-HAND
!C     OR          SIDE USING LDU FACTORIZATION FROM cnnf)
!C         cnnt   (SOLUTION OF TRANSPOSED LINEAR SYSTEM FOR ADDITIONAL
!C                 RIGHT-HAND SIDE USING LDU FACTORIZATION FROM cnnf)
!C
!C    II.  STORAGE OF SPARSE MATRICES
!C         THE NONZERO ENTRIES OF THE COEFFICIENT MATRIX M ARE STORED
!C    ROW-BY-ROW IN THE ARRAY A.  TO IDENTIFY THE INDIVIDUAL NONZERO
!C    ENTRIES IN EACH ROW, WE NEED TO KNOW IN WHICH COLUMN EACH ENTRY
!C    LIES.  THE COLUMN INDICES WHICH CORRESPOND TO THE NONZERO ENTRIES
!C    OF M ARE STORED IN THE ARRAY JA;  I.E., IF  A(K) = M(I,J),  THEN
!C    JA(K) = J.  IN ADDITION, WE NEED TO KNOW WHERE EACH ROW STARTS AND
!C    HOW LONG IT IS.  THE INDEX POSITIONS IN JA AND A WHERE THE ROWS OF
!C    M BEGIN ARE STORED IN THE ARRAY IA;  I.E., IF M(I,J) IS THE FIRST
!C    NONZERO ENTRY (STORED) IN THE I-TH ROW AND A(K) = M(I,J),  THEN
!C    IA(I) = K.  MOREOVER, THE INDEX IN JA AND A OF THE FIRST LOCATION
!C    FOLLOWING THE LAST ELEMENT IN THE LAST ROW IS STORED IN IA(N+1).
!C    THUS, THE NUMBER OF ENTRIES IN THE I-TH ROW IS GIVEN BY
!C    IA(I+1) - IA(I),  THE NONZERO ENTRIES OF THE I-TH ROW ARE STORED
!C    CONSECUTIVELY IN
!C            A(IA(I)),  A(IA(I)+1),  ..., A(IA(I+1)-1),
!C    AND THE CORRESPONDING COLUMN INDICES ARE STORED CONSECUTIVELY IN
!C            JA(IA(I)), JA(IA(I)+1), ..., JA(IA(I+1)-1).
!C    FOR EXAMPLE, THE 5 BY 5 MATRIX
!C                ( 1. 0. 2. 0. 0.)
!C                ( 0. 3. 0. 0. 0.)
!C            M = ( 0. 4. 5. 6. 0.)
!C                ( 0. 0. 0. 7. 0.)
!C                ( 0. 0. 0. 8. 9.)
!C    WOULD BE STORED AS
!C               | 1  2  3  4  5  6  7  8  9
!C            ---+--------------------------
!C            IA | 1  3  4  7  8 10
!C            JA | 1  3  2  2  3  4  4  4  5
!C             A | 1. 2. 3. 4. 5. 6. 7. 8. 9.         .
!C
!C         THE STRICT TRIANGULAR PORTIONS OF THE MATRICES L AND U ARE
!C    STORED IN THE SAME FASHION USING THE ARRAYS  IL, JL, L  AND
!C    IU, JU, U  RESPECTIVELY.  THE DIAGONAL ENTRIES OF L AND U ARE
!C    ASSUMED TO BE EQUAL TO ONE AND ARE NOT STORED.  THE ARRAY D
!C    CONTAINS THE RECIPROCALS OF THE DIAGONAL ENTRIES OF THE MATRIX D.
!C
!C    III. ADDITIONAL STORAGE SAVINGS
!C         IN cnsf, R AND IC CAN BE THE SAME ARRAY IN THE CALLING
!C    SEQUENCE IF NO REORDERING OF THE COEFFICIENT MATRIX HAS BEEN DONE.
!C         IN cnnf, R, C AND IC CAN ALL BE THE SAME ARRAY IF NO REORDERING
!C    HAS BEEN DONE.  IF ONLY THE ROWS HAVE BEEN REORDERED, THEN C AND IC
!C    CAN BE THE SAME ARRAY.  IF THE ROW AND COLUMN ORDERINGS ARE THE
!C    SAME, THEN R AND C CAN BE THE SAME ARRAY.  Z AND ROW CAN BE THE
!C    SAME ARRAY.
!C         IN cnns OR cnnt, R AND C CAN BE THE SAME ARRAY IF NO REORDERING
!C    HAS BEEN DONE OR IF THE ROW AND COLUMN ORDERINGS ARE THE SAME.  Z
!C    AND B CAN BE THE SAME ARRAY;  HOWEVER, THEN B WILL BE DESTROYED.
!C
!C    IV.  PARAMETERS
!C         FOLLOWING IS A LIST OF PARAMETERS TO THE PROGRAMS.  NAMES ARE
!C    UNIFORM AMONG THE VARIOUS SUBROUTINES.  CLASS ABBREVIATIONS ARE --
!C       N - INTEGER VARIABLE
!C       F - REAL VARIABLE
!C       V - SUPPLIES A VALUE TO A SUBROUTINE
!C       R - RETURNS A RESULT FROM A SUBROUTINE
!C       I - USED INTERNALLY BY A SUBROUTINE
!C       A - ARRAY
!C
!C CLASS | PARAMETER
!C ------+----------
!C FVA   | A     - NONZERO ENTRIES OF THE COEFFICIENT MATRIX M, STORED
!C       |           BY ROWS.
!C       |           SIZE = NUMBER OF NONZERO ENTRIES IN M.
!C FVA   | B     - RIGHT-HAND SIDE B.
!C       |           SIZE = N.
!C NVA   | C     - ORDERING OF THE COLUMNS OF M.
!C       |           SIZE = N.
!C FVRA  | D     - RECIPROCALS OF THE DIAGONAL ENTRIES OF THE MATRIX D.
!C       |           SIZE = N.
!C NR    | FLAG  - ERROR FLAG;  VALUES AND THEIR MEANINGS ARE --
!C       |            0     NO ERRORS DETECTED
!C       |            N+K   NULL ROW IN A  --  ROW = K
!C       |           2N+K   DUPLICATE ENTRY IN A  --  ROW = K
!C       |           3N+K   INSUFFICIENT STORAGE FOR JL  --  ROW = K
!C       |           4N+1   INSUFFICIENT STORAGE FOR L
!C       |           5N+K   NULL PIVOT  --  ROW = K
!C       |           6N+K   INSUFFICIENT STORAGE FOR JU  --  ROW = K
!C       |           7N+1   INSUFFICIENT STORAGE FOR U
!C       |           8N+K   ZERO PIVOT  --  ROW = K
!C NVA   | IA    - POINTERS TO DELIMIT THE ROWS IN A.
!C       |           SIZE = N+1.
!C NVA   | IC    - INVERSE OF THE ORDERING OF THE COLUMNS OF M;  I.E.,
!C       |           IC(C(I) = I  FOR I=1,...N.
!C       |           SIZE = N.
!C NVRA  | IL    - POINTERS TO DELIMIT THE ROWS IN L.
!C       |           SIZE = N+1.
!C NVRA  | IU    - POINTERS TO DELIMIT THE ROWS IN U.
!C       |           SIZE = N+1.
!C NVA   | JA    - COLUMN NUMBERS CORRESPONDING TO THE ELEMENTS OF A.
!C       |           SIZE = SIZE OF A.
!C NVRA  | JL    - COLUMN NUMBERS CORRESPONDING TO THE ELEMENTS OF L.
!C       |           SIZE = JLMAX.
!C NV    | JLMAX - DECLARED DIMENSION OF JL;  JLMAX MUST BE LARGER THAN
!C       |           THE NUMBER OF NONZERO ENTRIES IN THE STRICT LOWER
!C       |           TRIANGLE OF M PLUS FILLIN  (IL(N+1)-1 AFTER cnsf).
!C NVRA  | JU    - COLUMN NUMBERS CORRESPONDING TO THE ELEMENTS OF U.
!C       |           SIZE = JUMAX.
!C NV    | JUMAX - DECLARED DIMENSION OF JU;  JUMAX MUST BE LARGER THAN
!C       |           THE NUMBER OF NONZERO ENTRIES IN THE STRICT UPPER
!C       |           TRIANGLE OF M PLUS FILLIN  (IU(N+1)-1 AFTER cnsf).
!C FVRA  | L     - NONZERO ENTRIES IN THE STRICT LOWER TRIANGULAR PORTION
!C       |           OF THE MATRIX L, STORED BY ROWS.
!C       |           SIZE = LMAX
!C NV    | LMAX  - DECLARED DIMENSION OF L;  LMAX MUST BE LARGER THAN
!C       |           THE NUMBER OF NONZERO ENTRIES IN THE STRICT LOWER
!C       |           TRIANGLE OF M PLUS FILLIN  (IL(N+1)-1 AFTER cnsf).
!C NV    | N     - NUMBER OF VARIABLES/EQUATIONS.
!C NVA   | R     - ORDERING OF THE ROWS OF M.
!C       |           SIZE = N.
!C FVRA  | U     - NONZERO ENTRIES IN THE STRICT UPPER TRIANGULAR PORTION
!C       |           OF THE MATRIX U, STORED BY ROWS.
!C       |           SIZE = UMAX.
!C NV    | UMAX  - DECLARED DIMENSION OF U;  UMAX MUST BE LARGER THAN
!C       |           THE NUMBER OF NONZERO ENTRIES IN THE STRICT UPPER
!C       |           TRIANGLE OF M PLUS FILLIN  (IU(N+1)-1 AFTER cnsf).
!C FRA   | Z     - SOLUTION X.
!C       |           SIZE = N.
!C
!C
!C       ----------------------------------------------------------------
!C
!C*** SUBROUTINE cnsf
!C*** SYMBOLIC LDU-FACTORIZATION OF A NONSYMMETRIC SPARSE MATRIX
!C      (UNCOMPRESSED POINTER STORAGE)
!C
        SUBROUTINE  cnsf
     *     (N, R,IC, IA,JA, IL,JL,JLMAX, IU,JU,JUMAX, Q, IM, FLAG)
!C
!C       INPUT VARIABLES:   N, R,IC, IA,JA, JLMAX, JUMAX.
!C       OUTPUT VARIABLES:  IL,JL, IU,JU, FLAG.
!C
!C       PARAMETERS USED INTERNALLY:
!C NIA   | Q     - SUPPOSE M' IS THE RESULT OF REORDERING M;  IF
!C       |           PROCESSING OF THE KTH ROW OF M' (HENCE THE KTH ROWS
!C       |           OF L AND U) IS BEING DONE, THEN Q(J) IS INITIALLY
!C       |           NONZERO IF M'(K,J) IS NONZERO;  SINCE VALUES NEED
!C       |           NOT BE STORED, EACH ENTRY POINTS TO THE NEXT
!C       |           NONZERO;  FOR EXAMPLE, IF  N=9  AND THE 5TH ROW OF
!C       |           M' IS
!C       |                   0 X X 0 X 0 0 X 0,
!C       |           THEN Q WILL INITIALLY BE
!C       |                   A 3 5 A 8 A A 10 A 2        (A - ARBITRARY);
!C       |           Q(N+1) POINTS TO THE FIRST NONZERO IN THE ROW AND
!C       |           THE LAST NONZERO POINTS TO  N+1;  AS THE ALGORITHM
!C       |           PROCEEDS, OTHER ELEMENTS OF Q ARE INSERTED IN THE
!C       |           LIST BECAUSE OF FILLIN.
!C       |           SIZE = N+1.
!C NIA   | IM    - AT EACH STEP IN THE FACTORIZATION, IM(I) IS THE LAST
!C       |           ELEMENT IN THE ITH ROW OF U WHICH NEEDS TO BE
!C       |           CONSIDERED IN COMPUTING FILLIN.
!C       |           SIZE = N.
!C
!C  INTERNAL VARIABLES--
!C    JLPTR - POINTS TO THE LAST POSITION USED IN  JL.
!C    JUPTR - POINTS TO THE LAST POSITION USED IN  JU.
!C
        INTEGER  R(1), IC(1),  IA(1), JA(1),  IL(1), JL(1),
     *     IU(1), JU(1),  Q(1),  IM(1),  FLAG,  QM, VJ
        INTEGER  JMAX, JMIN, J, I, M, K, JLMAX, N, JUMAX, JUPTR, JLPTR
!C
!C  ******  INITIALIZE POINTERS  ****************************************
        JLPTR = 0
        IL(1) = 1
        JUPTR = 0
        IU(1) = 1
!C
!C  ******  FOR EACH ROW OF L AND U  ************************************
        DO 10 K=1,N
!C  ******  SET Q TO THE REORDERED ROW OF A  ****************************
          Q(N+1) = N+1
          JMIN = IA(R(K))
          JMAX = IA(R(K)+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 101
          DO 2 J=JMIN,JMAX
            VJ = IC(JA(J))
            QM = N+1
   1        M = QM
            QM = Q(M)
            IF (QM.LT.VJ)  GO TO 1
            IF (QM.EQ.VJ)  GO TO 102
              Q(M) = VJ
              Q(VJ) = QM
   2        CONTINUE
!C
!C  ******  FOR EACH ENTRY IN THE LOWER TRIANGLE  ***********************
          I = N+1
   3      I = Q(I)
          IF (I.GE.K)  GO TO 7
!C  ******  L(K,I) WILL BE NONZERO, SO ADD IT TO JL  ********************
            JLPTR = JLPTR+1
            IF (JLPTR.GT.JLMAX)  GO TO 103
            JL(JLPTR) = I
            QM = I
!C  ******  INSPECT ITH ROW FOR FILLIN, ADJUST IM IF POSSIBLE  **********
            JMIN = IU(I)
            JMAX = IM(I)
            IF (JMIN.GT.JMAX)  GO TO 6
            DO 5 J=JMIN,JMAX
              VJ = JU(J)
              IF (VJ.EQ.K)  IM(I) = J
   4          M = QM
              QM = Q(M)
              IF (QM.LT.VJ)  GO TO 4
              IF (QM.EQ.VJ)  GO TO 5
                Q(M) = VJ
                Q(VJ) = QM
                QM = VJ
   5          CONTINUE
   6        GO TO 3
!C
!C  ******  CHECK FOR NULL PIVOT  ***************************************
   7      IF (I.NE.K)  GO TO 105
!C  ******  REMAINING ELEMENTS OF Q DEFINE STRUCTURE OF U(K, )  *********
   8      I = Q(I)
          IF (I.GT.N)  GO TO 9
            JUPTR = JUPTR+1
            IF (JUPTR.GT.JUMAX)  GO TO 106
            JU(JUPTR) = I
            GO TO 8
!C  ******  GET READY FOR NEXT ROW  *************************************
   9      IM(K) = JUPTR
          IL(K+1) = JLPTR+1
  10      IU(K+1) = JUPTR+1
!C
        FLAG = 0
        RETURN
!C
!C ** ERROR:  NULL ROW IN A
 101    FLAG = N + R(K)
        RETURN
!C ** ERROR:  DUPLICATE ENTRY IN A
 102    FLAG = 2*N + R(K)
        RETURN
!C ** ERROR:  INSUFFICIENT STORAGE FOR JL
 103    FLAG = 3*N + K
        RETURN
!C ** ERROR:  NULL PIVOT
 105    FLAG = 5*N + K
        RETURN
!C ** ERROR:  INSUFFICIENT STORAGE FOR JU
 106    FLAG = 6*N + K
        RETURN
        END
!C                            APPENDIX 1                          1/15/81
!C
!C        SUBROUTINES FOR SOLVING SPARSE NONSYMMETRIC SYSTEMS
!C        OF LINEAR EQUATIONS  (UNCOMPRESSED POINTER STORAGE)
!C
!C
!C*** SUBROUTINE dndrv
!C*** DRIVER FOR SUBROUTINES FOR SOLVING SPARSE NONSYMMETRIC SYSTEMS OF
!C       LINEAR EQUATIONS (UNCOMPRESSED POINTER STORAGE)
!C
        SUBROUTINE  dndrv
     *     (N, R,C,IC, IA,JA,A, B, Z, NSP,ISP,RSP,ESP, PATH, FLAG)
!C
!C    PARAMETERS
!C    CLASS ABBREVIATIONS ARE --
!C       N - INTEGER VARIABLE
!C       F - REAL VARIABLE
!C       V - SUPPLIES A VALUE TO THE DRIVER
!C       R - RETURNS A RESULT FROM THE DRIVER
!C       I - USED INTERNALLY BY THE DRIVER
!C       A - ARRAY
!C
!C CLASS | PARAMETER
!C ------+----------
!C       |
!C         THE NONZERO ENTRIES OF THE COEFFICIENT MATRIX M ARE STORED
!C    ROW-BY-ROW IN THE ARRAY A.  TO IDENTIFY THE INDIVIDUAL NONZERO
!C    ENTRIES IN EACH ROW, WE NEED TO KNOW IN WHICH COLUMN EACH ENTRY
!C    LIES.  THE COLUMN INDICES WHICH CORRESPOND TO THE NONZERO ENTRIES
!C    OF M ARE STORED IN THE ARRAY JA;  I.E., IF  A(K) = M(I,J),  THEN
!C    JA(K) = J.  IN ADDITION, WE NEED TO KNOW WHERE EACH ROW STARTS AND
!C    HOW LONG IT IS.  THE INDEX POSITIONS IN JA AND A WHERE THE ROWS OF
!C    M BEGIN ARE STORED IN THE ARRAY IA;  I.E., IF M(I,J) IS THE FIRST
!C    NONZERO ENTRY (STORED) IN THE I-TH ROW AND A(K) = M(I,J),  THEN
!C    IA(I) = K.  MOREOVER, THE INDEX IN JA AND A OF THE FIRST LOCATION
!C    FOLLOWING THE LAST ELEMENT IN THE LAST ROW IS STORED IN IA(N+1).
!C    THUS, THE NUMBER OF ENTRIES IN THE I-TH ROW IS GIVEN BY
!C    IA(I+1) - IA(I),  THE NONZERO ENTRIES OF THE I-TH ROW ARE STORED
!C    CONSECUTIVELY IN
!C            A(IA(I)),  A(IA(I)+1),  ..., A(IA(I+1)-1),
!C    AND THE CORRESPONDING COLUMN INDICES ARE STORED CONSECUTIVELY IN
!C            JA(IA(I)), JA(IA(I)+1), ..., JA(IA(I+1)-1).
!C    FOR EXAMPLE, THE 5 BY 5 MATRIX
!C                ( 1. 0. 2. 0. 0.)
!C                ( 0. 3. 0. 0. 0.)
!C            M = ( 0. 4. 5. 6. 0.)
!C                ( 0. 0. 0. 7. 0.)
!C                ( 0. 0. 0. 8. 9.)
!C    WOULD BE STORED AS
!C               | 1  2  3  4  5  6  7  8  9
!C            ---+--------------------------
!C            IA | 1  3  4  7  8 10
!C            JA | 1  3  2  2  3  4  4  4  5
!C             A | 1. 2. 3. 4. 5. 6. 7. 8. 9.         .
!C
!C NV    | N     - NUMBER OF VARIABLES/EQUATIONS.
!C FVA   | A     - NONZERO ENTRIES OF THE COEFFICIENT MATRIX M, STORED
!C       |           BY ROWS.
!C       |           SIZE = NUMBER OF NONZERO ENTRIES IN M.
!C NVA   | IA    - POINTERS TO DELIMIT THE ROWS IN A.
!C       |           SIZE = N+1.
!C NVA   | JA    - COLUMN NUMBERS CORRESPONDING TO THE ELEMENTS OF A.
!C       |           SIZE = SIZE OF A.
!C FVA   | B     - RIGHT-HAND SIDE B;  B AND Z CAN THE SAME ARRAY.
!C       |           SIZE = N.
!C FRA   | Z     - SOLUTION X;  B AND Z CAN BE THE SAME ARRAY.
!C       |           SIZE = N.
!C
!C         THE ROWS AND COLUMNS OF THE ORIGINAL MATRIX M CAN BE
!C    REORDERED (E.G., TO REDUCE FILLIN OR ENSURE NUMERICAL STABILITY)
!C    BEFORE CALLING THE DRIVER.  IF NO REORDERING IS DONE, THEN SET
!C    R(I) = C(I) = IC(I) = I  FOR I=1,...,N.  THE SOLUTION Z IS RETURNED
!C    IN THE ORIGINAL ORDER.
!C
!C NVA   | R     - ORDERING OF THE ROWS OF M.
!C       |           SIZE = N.
!C NVA   | C     - ORDERING OF THE COLUMNS OF M.
!C       |           SIZE = N.
!C NVA   | IC    - INVERSE OF THE ORDERING OF THE COLUMNS OF M;  I.E.,
!C       |           IC(C(I)) = I  FOR I=1,...,N.
!C       |           SIZE = N.
!C
!C         THE SOLUTION OF THE SYSTEM OF LINEAR EQUATIONS IS DIVIDED INTO
!C    THREE STAGES --
!C      dnsf -- THE MATRIX M IS PROCESSED SYMBOLICALLY TO DETERMINE WHERE
!C              FILLIN WILL OCCUR DURING THE NUMERIC FACTORIZATION.
!C      dnnf -- THE MATRIX M IS FACTORED NUMERICALLY INTO THE PRODUCT LDU
!C              OF A UNIT LOWER TRIANGULAR MATRIX L, A DIAGONAL MATRIX D,
!C              AND A UNIT UPPER TRIANGULAR MATRIX U, AND THE SYSTEM
!C              MX = B  IS SOLVED.
!C      dnns -- THE LINEAR SYSTEM  MX = B  IS SOLVED USING THE LDU
!C  OR          FACTORIZATION FROM dnnf.
!C      dnnt -- THE TRANSPOSED LINEAR SYSTEM  MT X = B  IS SOLVED USING
!C              THE LDU FACTORIZATION FROM dnnf.
!C    FOR SEVERAL SYSTEMS WHOSE COEFFICIENT MATRICES HAVE THE SAME
!C    NONZERO STRUCTURE, dnsf NEED BE DONE ONLY ONCE (FOR THE FIRST
!C    SYSTEM);  THEN dnnf IS DONE ONCE FOR EACH ADDITIONAL SYSTEM.  FOR
!C    SEVERAL SYSTEMS WITH THE SAME COEFFICIENT MATRIX, dnsf AND dnnf NEED
!C    BE DONE ONLY ONCE (FOR THE FIRST SYSTEM);  THEN dnns OR dnnt IS DONE
!C    ONCE FOR EACH ADDITIONAL RIGHT-HAND SIDE.
!C
!C NV    | PATH  - PATH SPECIFICATION;  VALUES AND THEIR MEANINGS ARE --
!C       |           1  PERFORM dnsf AND dnnf.
!C       |           2  PERFORM dnnf ONLY  (dnsf IS ASSUMED TO HAVE BEEN
!C       |               DONE IN A MANNER COMPATIBLE WITH THE STORAGE
!C       |               ALLOCATION USED IN THE DRIVER).
!C       |           3  PERFORM dnns ONLY  (dnsf AND dnnf ARE ASSUMED TO
!C       |               HAVE BEEN DONE IN A MANNER COMPATIBLE WITH THE
!C       |               STORAGE ALLOCATION USED IN THE DRIVER).
!C       |           4  PERFORM dnnt ONLY  (dnsf AND dnnf ARE ASSUMED TO
!C       |               HAVE BEEN DONE IN A MANNER COMPATIBLE WITH THE
!C       |               STORAGE ALLOCATION USED IN THE DRIVER).
!C       |           5  PERFORM dnsf ONLY.
!C
!C         VARIOUS ERRORS ARE DETECTED BY THE DRIVER AND THE INDIVIDUAL
!C    SUBROUTINES.
!C
!C NR    | FLAG  - ERROR FLAG;  VALUES AND THEIR MEANINGS ARE --
!C       |             0     NO ERRORS DETECTED
!C       |             N+K   NULL ROW IN A  --  ROW = K
!C       |            2N+K   DUPLICATE ENTRY IN A  --  ROW = K
!C       |            3N+K   INSUFFICIENT STORAGE IN dnsf  --  ROW = K
!C       |            4N+1   INSUFFICIENT STORAGE IN dnnf
!C       |            5N+K   NULL PIVOT  --  ROW = K
!c       |            6N+K   INSUFFICIENT STORAGE IN dnsf  --  ROW = K
!C       |            7N+1   INSUFFICIENT STORAGE IN dnnf
!C       |            8N+K   ZERO PIVOT  --  ROW = K
!C       |           10N+1   INSUFFICIENT STORAGE IN dndrv
!C       |           11N+1   ILLEGAL PATH SPECIFICATION
!C
!C         WORKING STORAGE IS NEEDED FOR THE FACTORED FORM OF THE MATRIX
!C    M PLUS VARIOUS TEMPORARY VECTORS.  THE ARRAYS ISP AND RSP SHOULD BE
!C    EQUIVALENCED;  INTEGER STORAGE IS ALLOCATED FROM THE BEGINNING OF
!C    ISP AND REAL STORAGE FROM THE END OF RSP.
!C
!C NV    | NSP   - DECLARED DIMENSION OF RSP;  NSP GENERALLY MUST
!C       |           BE LARGER THAN  5N+3 + 2K  (WHERE  K = (NUMBER OF
!C       |           NONZERO ENTRIES IN M)).
!C NVIRA | ISP   - INTEGER WORKING STORAGE DIVIDED UP INTO VARIOUS ARRAYS
!C       |           NEEDED BY THE SUBROUTINES;  ISP AND RSP SHOULD BE
!C       |           EQUIVALENCED.
!C       |           SIZE = 2*NSP
!C FVIRA | RSP   - REAL WORKING STORAGE DIVIDED UP INTO VARIOUS ARRAYS
!C       |           NEEDED BY THE SUBROUTINES;  ISP AND RSP SHOULD BE
!C       |           EQUIVALENCED.
!C       |           SIZE = NSP.
!C NR    | ESP   - IF SUFFICIENT STORAGE WAS AVAILABLE TO PERFORM THE
!C       |           SYMBOLIC FACTORIZATION (dnsf), THEN ESP IS SET TO THE
!C       |           AMOUNT OF EXCESS STORAGE PROVIDED (NEGATIVE IF
!C       |           INSUFFICIENT STORAGE WAS AVAILABLE TO PERFORM THE
!C       |           NUMERIC FACTORIZATION (dnnf)).
!C
!C
!C  CONVERSION TO DOUBLE PRECISION
!C
!C    TO CONVERT THESE ROUTINES FOR DOUBLE PRECISION ARRAYS, SIMPLY USE
!C    THE DOUBLE PRECISION DECLARATIONS IN PLACE OF THE REAL DECLARATIONS
!C    IN EACH SUBPROGRAM;  IN ADDITION, THE DATA VALUE OF THE INTEGER
!C    VARIABLE 2 MUST BE SET AS INDICATED IN SUBROUTINE dndrv
!C
        INTEGER  R(1), C(1), IC(1),  IA(1), JA(1),  ISP(1), ESP,
     *     PATH, FLAG,  Q, IM, D, U, ROW, TMP,  UMAX
        double precision    A(1),  B(1),  Z(1),  RSP(1)
        INTEGER  JU, JUMAX, JUTMP, LMAX, L, J, JLMAX, IL, NSP, N, MAX,
     *           JL, IU
!C
!C  SET 2 EQUAL TO THE RATIO BETWEEN THE LENGTH OF FLOATING POINT
!C  AND INTEGER ARRAY DATA;  E. G., 2 = 1 FOR (REAL, INTEGER),
!C  2 = 2 FOR (DOUBLE PRECISION, INTEGER)
!C
!c       DATA 2/2/
!C
        IF (PATH.LT.1 .OR. 5.LT.PATH)  GO TO 111
!C  ******  INITIALIZE AND DIVIDE UP TEMPORARY STORAGE  *****************
        IL = 1
        IU = IL + N+1
        JL = IU + N+1
!C
!C  ******  CALL dnsf IF FLAG IS SET  ************************************
        IF ((PATH-1) * (PATH-5) .NE. 0)  GO TO 2
          MAX = (2*NSP + 1 - JL) - (N+1) - N
          JLMAX = MAX/2
          Q     = JL  + JLMAX
          IM    = Q   + (N+1)
          JUTMP = IM  +   N
          JUMAX = 2*NSP + 1 - JUTMP
          ESP = MAX/2
          IF (JLMAX.LE.0 .OR. JUMAX.LE.0)  GO TO 110
          CALL  dnsf
     *       (N,  R, IC,  IA, JA,
     *        ISP(IL), ISP(JL), JLMAX,  ISP(IU), ISP(JUTMP), JUMAX,
     *        ISP(Q),  ISP(IM),  FLAG)
          IF (FLAG.NE.0)  GO TO 100
!C  ******  MOVE JU NEXT TO JL  *****************************************
          JLMAX = ISP(IL+N)-1
          JU    = JL + JLMAX
          JUMAX = ISP(IU+N)-1
          IF (JUMAX.LE.0)  GO TO 2
          DO 1 J=1,JUMAX
   1        ISP(JU+J-1) = ISP(JUTMP+J-1)
!C
!C  ******  CALL REMAINING SUBROUTINES  *********************************
   2    JLMAX = ISP(IL+N)-1
        JU    = JL  + JLMAX
        JUMAX = ISP(IU+N)-1
        L     = (JU + JUMAX - 2 + 2)  /  2    +    1
        LMAX  = JLMAX
        D     = L   + LMAX
        U     = D   + N
        ROW   = NSP + 1 - N
        TMP   = ROW - N
        UMAX  = TMP - U
        ESP = UMAX - JUMAX
!C
        IF ((PATH-1) * (PATH-2) .NE. 0)  GO TO 3
          IF (UMAX.LE.0)  GO TO 110
          CALL  dnnf
     *       (N,  R, C, IC,  IA, JA, A,  Z,  B,
     *        ISP(IL), ISP(JL), RSP(L), LMAX,   RSP(D),
     *           ISP(IU), ISP(JU), RSP(U), UMAX,
     *        RSP(ROW),  RSP(TMP),  FLAG)
          IF (FLAG.NE.0)  GO TO 100
          RETURN
!C
   3    IF ((PATH-3) .NE. 0)  GO TO 4
          CALL  dnns
     *       (N,  R, C,
     *        ISP(IL), ISP(JL), RSP(L),  RSP(D),
     *          ISP(IU), ISP(JU), RSP(U),
     *        Z,  B,  RSP(TMP))
!C
   4    IF ((PATH-4) .NE. 0)  GO TO 5
          CALL  dnnt
     *       (N,  R, C,
     *        ISP(IL), ISP(JL), RSP(L),  RSP(D),
     *          ISP(IU), ISP(JU), RSP(U),
     *        Z,  B,  RSP(TMP))
   5    RETURN
!C
!C ** ERROR:  ERROR DETECTED IN dnsf, dnnf, dnns, OR dnnt
 100    RETURN
!C ** ERROR:  INSUFFICIENT STORAGE
 110    FLAG = 10*N + 1
        RETURN
!C ** ERROR:  ILLEGAL PATH SPECIFICATION
 111    FLAG = 11*N + 1
        RETURN
        END
!C
!C       ----------------------------------------------------------------
!C
!C*** SUBROUTINE dnnf
!C*** NUMERIC LDU-FACTORIZATION OF SPARSE NONSYMMETRIC MATRIX AND
!C      SOLUTION OF SYSTEM OF LINEAR EQUATIONS (UNCOMPRESSED POINTER
!C      STORAGE)
!C
        SUBROUTINE  dnnf
     *     (N, R,C,IC, IA,JA,A, Z, B, IL,JL,L,LMAX, D, IU,JU,U,UMAX,
     *      ROW, TMP, FLAG)
!C
!C       INPUT VARIABLES:   N, R,C,IC, IA,JA,A, B, IL,JL,LMAX, IU,JU,UMAX
!C       OUTPUT VARIABLES:  Z, L,D,U, FLAG
!C
!C       PARAMETERS USED INTERNALLY:
!C FIA   | ROW   - HOLDS INTERMEDIATE VALUES IN CALCULATION OF L, D, U.
!C       |           SIZE = N.
!C FIA   | TMP   - HOLDS NEW RIGHT-HAND SIDE B' FOR SOLUTION OF THE
!C       |           EQUATION  UX = B'.
!C       |           SIZE = N.
!C
        INTEGER  R(1), C(1), IC(1),  IA(1), JA(1),
     *     IL(1), JL(1), LMAX,  IU(1), JU(1), UMAX,  FLAG
        double precision    A(1), Z(1), B(1),  L(1), D(1), U(1),
     *     ROW(1), TMP(1),  LI, SUM, DK
        INTEGER  IMIN, J, I, IMAX, K, N, JMAX, JMIN
!C
!C  ******  CHECK STORAGE  **********************************************
        IF (IL(N+1)-1 .GT. LMAX)  GO TO 104
        IF (IU(N+1)-1 .GT. UMAX)  GO TO 107
!C
!C  ******  FOR EACH ROW  ***********************************************
        DO 10 K=1,N
!C  ******  SET THE INITIAL STRUCTURE OF ROW  ***************************
          JMIN = IL(K)
          JMAX = IL(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 2
!C  ******  IF L(K,M) .NE. 0, ROW(M)=0  *********************************
          DO 1 J=JMIN,JMAX
   1        ROW(JL(J)) = 0
   2      ROW(K) = 0
          JMIN = IU(K)
          JMAX = IU(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 4
!C  ******  IF U(K,M) .NE. 0, ROW(M)=0  *********************************
          DO 3 J=JMIN,JMAX
   3        ROW(JU(J)) = 0
   4      JMIN = IA(R(K))
          JMAX = IA(R(K)+1) - 1
!C  ******  SET ROW TO KTH ROW OF REORDERED A  **************************
          DO 5 J=JMIN,JMAX
   5        ROW(IC(JA(J))) = A(J)
!C  ******  INITIALIZE SUM  *********************************************
          SUM = B(R(K))
!C
!C  ******  ASSIGN THE KTH ROW OF L AND ADJUST ROW, SUM  ****************
          IMIN = IL(K)
          IMAX = IL(K+1) - 1
          IF (IMIN.GT.IMAX)  GO TO 8
          DO 7 I=IMIN,IMAX
            LI = - ROW(JL(I))
!C  ******  IF L IS NOT REQUIRED, THEN COMMENT OUT THE FOLLOWING LINE  **
            L(I) = - LI
            SUM = SUM + LI * TMP(JL(I))
            JMIN = IU(JL(I))
            JMAX = IU(JL(I)+1) - 1
            IF (JMIN.GT.JMAX)  GO TO 7
            DO 6 J=JMIN,JMAX
   6          ROW(JU(J)) = ROW(JU(J)) + LI * U(J)
   7        CONTINUE
!C
!C  ******  ASSIGN DIAGONAL D AND KTH ROW OF U, SET TMP(K)  *************
   8      IF (ROW(K).EQ.0)  GO TO 108
          DK = 1 / ROW(K)
          D(K) = DK
          TMP(K) = SUM * DK
          JMIN = IU(K)
          JMAX = IU(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 10
          DO 9 J=JMIN,JMAX
   9        U(J) = ROW(JU(J)) * DK
  10      CONTINUE
!C
!C  ******  SOLVE  UX = TMP  BY BACK SUBSTITUTION  **********************
        K = N
        DO 13 I=1,N
          SUM = TMP(K)
          JMIN = IU(K)
          JMAX = IU(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 12
          DO 11 J=JMIN,JMAX
  11        SUM = SUM - U(J) * TMP(JU(J))
  12      TMP(K) = SUM
          Z(C(K)) = SUM
  13      K = K-1
!C
        FLAG = 0
        RETURN
!C
!C ** ERROR:  INSUFFICIENT STORAGE FOR L
 104    FLAG = 4*N + 1
        RETURN
!C ** ERROR:  INSUFFICIENT STORAGE FOR U
 107    FLAG = 7*N + 1
        RETURN
!C ** ERROR:  ZERO PIVOT
 108    FLAG = 8*N + K
        RETURN
        END
!C
!C       ----------------------------------------------------------------
!C
!C*** SUBROUTINE dnns
!C*** NUMERIC SOLUTION OF A SPARSE NONSYMMETRIC SYSTEM OF LINEAR
!C      EQUATIONS GIVEN LDU-FACTORIZATION (UNCOMPRESSED POINTER STORAGE)
!C
        SUBROUTINE  dnns
     *     (N, R,C, IL,JL,L, D, IU,JU,U, Z, B, TMP)
!C
!C       INPUT VARIABLES:   N, R,C, IL,JL,L, D, IU,JU,U, B
!C       OUTPUT VARIABLES:  Z
!C
!C       PARAMETERS USED INTERNALLY:
!C FIA   | TMP   - HOLDS NEW RIGHT-HAND SIDE B' FOR SOLUTION OF THE
!C       |           EQUATION UX = B'.
!C       |           SIZE = N.
!C
        INTEGER  R(1), C(1),  IL(1), JL(1),  IU(1), JU(1)
        double precision    L(1), D(1), U(1),  Z(1), B(1),
     *           TMP(1), SUM
        INTEGER  JMAX, J, I, N, K, JMIN
!C
!C  ******  SOLVE LDY = B  BY FORWARD SUBSTITUTION  *********************
        DO 2 K=1,N
          SUM = B(R(K))
          JMIN = IL(K)
          JMAX = IL(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 2
          DO 1 J=JMIN,JMAX
   1        SUM = SUM - L(J) * TMP(JL(J))
   2      TMP(K) = SUM * D(K)
!C
!C  ******  SOLVE  UX = Y  BY BACK SUBSTITUTION  ************************
        K = N
        DO 5 I=1,N
          SUM = TMP(K)
          JMIN = IU(K)
          JMAX = IU(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 4
          DO 3 J=JMIN,JMAX
   3        SUM = SUM - U(J) * TMP(JU(J))
   4      TMP(K) = SUM
          Z(C(K)) = SUM
   5      K = K-1
        RETURN
        END
!C       ----------------------------------------------------------------
!C
!C*** SUBROUTINE dnnt
!C*** NUMERIC SOLUTION OF THE TRANSPOSE OF A SPARSE NONSYMMETRIC SYSTEM
!C      OF LINEAR EQUATIONS GIVEN LDU-FACTORIZATION (UNCOMPRESSED POINTER
!C      STORAGE)
!C
        SUBROUTINE  dnnt
     *     (N, R,C, IL,JL,L, D, IU,JU,U, Z, B, TMP)
!C
!C       INPUT VARIABLES:   N, R,C, IL,JL,L, D, IU,JU,U, B
!C       OUTPUT VARIABLES:  Z
!C
!C       PARAMETERS USED INTERNALLY:
!C FIA   | TMP   - HOLDS NEW RIGHT-HAND SIDE B' FOR SOLUTION OF THE
!C       |           EQUATION LX = B'.
!C       |           SIZE = N.
!C
        INTEGER  R(1), C(1),  IL(1), JL(1),  IU(1), JU(1)
        double precision    L(1), D(1), U(1),  Z(1), B(1),
     *           TMP(1), TMPK
        INTEGER  JMAX, J, I, N, K, JMIN
!C
!C  ******  SOLVE  UT Y = B  BY FORWARD SUBSTITUTION  *******************
        DO 1 K=1,N
   1      TMP(K) = B(C(K))
        DO 3 K=1,N
          TMPK = - TMP(K)
          JMIN = IU(K)
          JMAX = IU(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 3
          DO 2 J=JMIN,JMAX
   2        TMP(JU(J)) = TMP(JU(J)) + U(J) * TMPK
   3      CONTINUE
!C
!C  ******  SOLVE  D LT X = Y  BY BACK SUBSTITUTION  ********************
        K = N
        DO 6 I=1,N
          TMPK = - TMP(K) * D(K)
          JMIN = IL(K)
          JMAX = IL(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 5
          DO 4 J=JMIN,JMAX
   4        TMP(JL(J)) = TMP(JL(J)) + L(J) * TMPK
   5      Z(R(K)) = - TMPK
   6      K = K-1
        RETURN
        END
!C
!C       ----------------------------------------------------------------
!C
!C               YALE SPARSE MATRIX PACKAGE - NONSYMMETRIC CODES
!C                    SOLVING THE SYSTEM OF EQUATIONS MX = B
!C                        (UNCOMPRESSED POINTER STORAGE)
!C
!C    I.   CALLING SEQUENCES
!C         THE COEFFICIENT MATRIX CAN BE PROCESSED BY AN ORDERING ROUTINE
!C    (E.G., TO REDUCE FILLIN OR ENSURE NUMERICAL STABILITY) BEFORE USING
!C    THE REMAINING SUBROUTINES.  IF NO REORDERING IS DONE, THEN SET
!C    R(I) = C(I) = IC(I) = I  FOR I=1,...,N.  THE CALLING SEQUENCE IS --
!C        (      (MATRIX ORDERING))
!C         dnsf   (SYMBOLIC FACTORIZATION TO DETERMINE WHERE FILLIN WILL
!C                 OCCUR DURING NUMERIC FACTORIZATION)
!C         dnnf   (NUMERIC FACTORIZATION INTO PRODUCT LDU OF UNIT LOWER
!C                 TRIANGULAR MATRIX L, DIAGONAL MATRIX D, AND UNIT UPPER
!C                 TRIANGULAR MATRIX U, AND SOLUTION OF LINEAR SYSTEM)
!C         dnns   (SOLUTION OF LINEAR SYSTEM FOR ADDITIONAL RIGHT-HAND
!C     OR          SIDE USING LDU FACTORIZATION FROM dnnf)
!C         dnnt   (SOLUTION OF TRANSPOSED LINEAR SYSTEM FOR ADDITIONAL
!C                 RIGHT-HAND SIDE USING LDU FACTORIZATION FROM dnnf)
!C
!C    II.  STORAGE OF SPARSE MATRICES
!C         THE NONZERO ENTRIES OF THE COEFFICIENT MATRIX M ARE STORED
!C    ROW-BY-ROW IN THE ARRAY A.  TO IDENTIFY THE INDIVIDUAL NONZERO
!C    ENTRIES IN EACH ROW, WE NEED TO KNOW IN WHICH COLUMN EACH ENTRY
!C    LIES.  THE COLUMN INDICES WHICH CORRESPOND TO THE NONZERO ENTRIES
!C    OF M ARE STORED IN THE ARRAY JA;  I.E., IF  A(K) = M(I,J),  THEN
!C    JA(K) = J.  IN ADDITION, WE NEED TO KNOW WHERE EACH ROW STARTS AND
!C    HOW LONG IT IS.  THE INDEX POSITIONS IN JA AND A WHERE THE ROWS OF
!C    M BEGIN ARE STORED IN THE ARRAY IA;  I.E., IF M(I,J) IS THE FIRST
!C    NONZERO ENTRY (STORED) IN THE I-TH ROW AND A(K) = M(I,J),  THEN
!C    IA(I) = K.  MOREOVER, THE INDEX IN JA AND A OF THE FIRST LOCATION
!C    FOLLOWING THE LAST ELEMENT IN THE LAST ROW IS STORED IN IA(N+1).
!C    THUS, THE NUMBER OF ENTRIES IN THE I-TH ROW IS GIVEN BY
!C    IA(I+1) - IA(I),  THE NONZERO ENTRIES OF THE I-TH ROW ARE STORED
!C    CONSECUTIVELY IN
!C            A(IA(I)),  A(IA(I)+1),  ..., A(IA(I+1)-1),
!C    AND THE CORRESPONDING COLUMN INDICES ARE STORED CONSECUTIVELY IN
!C            JA(IA(I)), JA(IA(I)+1), ..., JA(IA(I+1)-1).
!C    FOR EXAMPLE, THE 5 BY 5 MATRIX
!C                ( 1. 0. 2. 0. 0.)
!C                ( 0. 3. 0. 0. 0.)
!C            M = ( 0. 4. 5. 6. 0.)
!C                ( 0. 0. 0. 7. 0.)
!C                ( 0. 0. 0. 8. 9.)
!C    WOULD BE STORED AS
!C               | 1  2  3  4  5  6  7  8  9
!C            ---+--------------------------
!C            IA | 1  3  4  7  8 10
!C            JA | 1  3  2  2  3  4  4  4  5
!C             A | 1. 2. 3. 4. 5. 6. 7. 8. 9.         .
!C
!C         THE STRICT TRIANGULAR PORTIONS OF THE MATRICES L AND U ARE
!C    STORED IN THE SAME FASHION USING THE ARRAYS  IL, JL, L  AND
!C    IU, JU, U  RESPECTIVELY.  THE DIAGONAL ENTRIES OF L AND U ARE
!C    ASSUMED TO BE EQUAL TO ONE AND ARE NOT STORED.  THE ARRAY D
!C    CONTAINS THE RECIPROCALS OF THE DIAGONAL ENTRIES OF THE MATRIX D.
!C
!C    III. ADDITIONAL STORAGE SAVINGS
!C         IN dnsf, R AND IC CAN BE THE SAME ARRAY IN THE CALLING
!C    SEQUENCE IF NO REORDERING OF THE COEFFICIENT MATRIX HAS BEEN DONE.
!C         IN dnnf, R, C AND IC CAN ALL BE THE SAME ARRAY IF NO REORDERING
!C    HAS BEEN DONE.  IF ONLY THE ROWS HAVE BEEN REORDERED, THEN C AND IC
!C    CAN BE THE SAME ARRAY.  IF THE ROW AND COLUMN ORDERINGS ARE THE
!C    SAME, THEN R AND C CAN BE THE SAME ARRAY.  Z AND ROW CAN BE THE
!C    SAME ARRAY.
!C         IN dnns OR dnnt, R AND C CAN BE THE SAME ARRAY IF NO REORDERING
!C    HAS BEEN DONE OR IF THE ROW AND COLUMN ORDERINGS ARE THE SAME.  Z
!C    AND B CAN BE THE SAME ARRAY;  HOWEVER, THEN B WILL BE DESTROYED.
!C
!C    IV.  PARAMETERS
!C         FOLLOWING IS A LIST OF PARAMETERS TO THE PROGRAMS.  NAMES ARE
!C    UNIFORM AMONG THE VARIOUS SUBROUTINES.  CLASS ABBREVIATIONS ARE --
!C       N - INTEGER VARIABLE
!C       F - REAL VARIABLE
!C       V - SUPPLIES A VALUE TO A SUBROUTINE
!C       R - RETURNS A RESULT FROM A SUBROUTINE
!C       I - USED INTERNALLY BY A SUBROUTINE
!C       A - ARRAY
!C
!C CLASS | PARAMETER
!C ------+----------
!C FVA   | A     - NONZERO ENTRIES OF THE COEFFICIENT MATRIX M, STORED
!C       |           BY ROWS.
!C       |           SIZE = NUMBER OF NONZERO ENTRIES IN M.
!C FVA   | B     - RIGHT-HAND SIDE B.
!C       |           SIZE = N.
!C NVA   | C     - ORDERING OF THE COLUMNS OF M.
!C       |           SIZE = N.
!C FVRA  | D     - RECIPROCALS OF THE DIAGONAL ENTRIES OF THE MATRIX D.
!C       |           SIZE = N.
!C NR    | FLAG  - ERROR FLAG;  VALUES AND THEIR MEANINGS ARE --
!C       |            0     NO ERRORS DETECTED
!C       |            N+K   NULL ROW IN A  --  ROW = K
!C       |           2N+K   DUPLICATE ENTRY IN A  --  ROW = K
!C       |           3N+K   INSUFFICIENT STORAGE FOR JL  --  ROW = K
!C       |           4N+1   INSUFFICIENT STORAGE FOR L
!C       |           5N+K   NULL PIVOT  --  ROW = K
!C       |           6N+K   INSUFFICIENT STORAGE FOR JU  --  ROW = K
!C       |           7N+1   INSUFFICIENT STORAGE FOR U
!C       |           8N+K   ZERO PIVOT  --  ROW = K
!C NVA   | IA    - POINTERS TO DELIMIT THE ROWS IN A.
!C       |           SIZE = N+1.
!C NVA   | IC    - INVERSE OF THE ORDERING OF THE COLUMNS OF M;  I.E.,
!C       |           IC(C(I) = I  FOR I=1,...N.
!C       |           SIZE = N.
!C NVRA  | IL    - POINTERS TO DELIMIT THE ROWS IN L.
!C       |           SIZE = N+1.
!C NVRA  | IU    - POINTERS TO DELIMIT THE ROWS IN U.
!C       |           SIZE = N+1.
!C NVA   | JA    - COLUMN NUMBERS CORRESPONDING TO THE ELEMENTS OF A.
!C       |           SIZE = SIZE OF A.
!C NVRA  | JL    - COLUMN NUMBERS CORRESPONDING TO THE ELEMENTS OF L.
!C       |           SIZE = JLMAX.
!C NV    | JLMAX - DECLARED DIMENSION OF JL;  JLMAX MUST BE LARGER THAN
!C       |           THE NUMBER OF NONZERO ENTRIES IN THE STRICT LOWER
!C       |           TRIANGLE OF M PLUS FILLIN  (IL(N+1)-1 AFTER dnsf).
!C NVRA  | JU    - COLUMN NUMBERS CORRESPONDING TO THE ELEMENTS OF U.
!C       |           SIZE = JUMAX.
!C NV    | JUMAX - DECLARED DIMENSION OF JU;  JUMAX MUST BE LARGER THAN
!C       |           THE NUMBER OF NONZERO ENTRIES IN THE STRICT UPPER
!C       |           TRIANGLE OF M PLUS FILLIN  (IU(N+1)-1 AFTER dnsf).
!C FVRA  | L     - NONZERO ENTRIES IN THE STRICT LOWER TRIANGULAR PORTION
!C       |           OF THE MATRIX L, STORED BY ROWS.
!C       |           SIZE = LMAX
!C NV    | LMAX  - DECLARED DIMENSION OF L;  LMAX MUST BE LARGER THAN
!C       |           THE NUMBER OF NONZERO ENTRIES IN THE STRICT LOWER
!C       |           TRIANGLE OF M PLUS FILLIN  (IL(N+1)-1 AFTER dnsf).
!C NV    | N     - NUMBER OF VARIABLES/EQUATIONS.
!C NVA   | R     - ORDERING OF THE ROWS OF M.
!C       |           SIZE = N.
!C FVRA  | U     - NONZERO ENTRIES IN THE STRICT UPPER TRIANGULAR PORTION
!C       |           OF THE MATRIX U, STORED BY ROWS.
!C       |           SIZE = UMAX.
!C NV    | UMAX  - DECLARED DIMENSION OF U;  UMAX MUST BE LARGER THAN
!C       |           THE NUMBER OF NONZERO ENTRIES IN THE STRICT UPPER
!C       |           TRIANGLE OF M PLUS FILLIN  (IU(N+1)-1 AFTER dnsf).
!C FRA   | Z     - SOLUTION X.
!C       |           SIZE = N.
!C
!C
!C       ----------------------------------------------------------------
!C
!C*** SUBROUTINE dnsf
!C*** SYMBOLIC LDU-FACTORIZATION OF A NONSYMMETRIC SPARSE MATRIX
!C      (UNCOMPRESSED POINTER STORAGE)
!C
        SUBROUTINE  dnsf
     *     (N, R,IC, IA,JA, IL,JL,JLMAX, IU,JU,JUMAX, Q, IM, FLAG)
!C
!C       INPUT VARIABLES:   N, R,IC, IA,JA, JLMAX, JUMAX.
!C       OUTPUT VARIABLES:  IL,JL, IU,JU, FLAG.
!C
!C       PARAMETERS USED INTERNALLY:
!C NIA   | Q     - SUPPOSE M' IS THE RESULT OF REORDERING M;  IF
!C       |           PROCESSING OF THE KTH ROW OF M' (HENCE THE KTH ROWS
!C       |           OF L AND U) IS BEING DONE, THEN Q(J) IS INITIALLY
!C       |           NONZERO IF M'(K,J) IS NONZERO;  SINCE VALUES NEED
!C       |           NOT BE STORED, EACH ENTRY POINTS TO THE NEXT
!C       |           NONZERO;  FOR EXAMPLE, IF  N=9  AND THE 5TH ROW OF
!C       |           M' IS
!C       |                   0 X X 0 X 0 0 X 0,
!C       |           THEN Q WILL INITIALLY BE
!C       |                   A 3 5 A 8 A A 10 A 2        (A - ARBITRARY);
!C       |           Q(N+1) POINTS TO THE FIRST NONZERO IN THE ROW AND
!C       |           THE LAST NONZERO POINTS TO  N+1;  AS THE ALGORITHM
!C       |           PROCEEDS, OTHER ELEMENTS OF Q ARE INSERTED IN THE
!C       |           LIST BECAUSE OF FILLIN.
!C       |           SIZE = N+1.
!C NIA   | IM    - AT EACH STEP IN THE FACTORIZATION, IM(I) IS THE LAST
!C       |           ELEMENT IN THE ITH ROW OF U WHICH NEEDS TO BE
!C       |           CONSIDERED IN COMPUTING FILLIN.
!C       |           SIZE = N.
!C
!C  INTERNAL VARIABLES--
!C    JLPTR - POINTS TO THE LAST POSITION USED IN  JL.
!C    JUPTR - POINTS TO THE LAST POSITION USED IN  JU.
!C
        INTEGER  R(1), IC(1),  IA(1), JA(1),  IL(1), JL(1),
     *     IU(1), JU(1),  Q(1),  IM(1),  FLAG,  QM, VJ
        INTEGER  JMAX, JMIN, J, I, M, K, JLMAX, N, JUMAX, JUPTR, JLPTR
!C
!C  ******  INITIALIZE POINTERS  ****************************************
        JLPTR = 0
        IL(1) = 1
        JUPTR = 0
        IU(1) = 1
!C
!C  ******  FOR EACH ROW OF L AND U  ************************************
        DO 10 K=1,N
!C  ******  SET Q TO THE REORDERED ROW OF A  ****************************
          Q(N+1) = N+1
          JMIN = IA(R(K))
          JMAX = IA(R(K)+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 101
          DO 2 J=JMIN,JMAX
            VJ = IC(JA(J))
            QM = N+1
   1        M = QM
            QM = Q(M)
            IF (QM.LT.VJ)  GO TO 1
            IF (QM.EQ.VJ)  GO TO 102
              Q(M) = VJ
              Q(VJ) = QM
   2        CONTINUE
!C
!C  ******  FOR EACH ENTRY IN THE LOWER TRIANGLE  ***********************
          I = N+1
   3      I = Q(I)
          IF (I.GE.K)  GO TO 7
!C  ******  L(K,I) WILL BE NONZERO, SO ADD IT TO JL  ********************
            JLPTR = JLPTR+1
            IF (JLPTR.GT.JLMAX)  GO TO 103
            JL(JLPTR) = I
            QM = I
!C  ******  INSPECT ITH ROW FOR FILLIN, ADJUST IM IF POSSIBLE  **********
            JMIN = IU(I)
            JMAX = IM(I)
            IF (JMIN.GT.JMAX)  GO TO 6
            DO 5 J=JMIN,JMAX
              VJ = JU(J)
              IF (VJ.EQ.K)  IM(I) = J
   4          M = QM
              QM = Q(M)
              IF (QM.LT.VJ)  GO TO 4
              IF (QM.EQ.VJ)  GO TO 5
                Q(M) = VJ
                Q(VJ) = QM
                QM = VJ
   5          CONTINUE
   6        GO TO 3
!C
!C  ******  CHECK FOR NULL PIVOT  ***************************************
   7      IF (I.NE.K)  GO TO 105
!C  ******  REMAINING ELEMENTS OF Q DEFINE STRUCTURE OF U(K, )  *********
   8      I = Q(I)
          IF (I.GT.N)  GO TO 9
            JUPTR = JUPTR+1
            IF (JUPTR.GT.JUMAX)  GO TO 106
            JU(JUPTR) = I
            GO TO 8
!C  ******  GET READY FOR NEXT ROW  *************************************
   9      IM(K) = JUPTR
          IL(K+1) = JLPTR+1
  10      IU(K+1) = JUPTR+1
!C
        FLAG = 0
        RETURN
!C
!C ** ERROR:  NULL ROW IN A
 101    FLAG = N + R(K)
        RETURN
!C ** ERROR:  DUPLICATE ENTRY IN A
 102    FLAG = 2*N + R(K)
        RETURN
!C ** ERROR:  INSUFFICIENT STORAGE FOR JL
 103    FLAG = 3*N + K
        RETURN
!C ** ERROR:  NULL PIVOT
 105    FLAG = 5*N + K
        RETURN
!C ** ERROR:  INSUFFICIENT STORAGE FOR JU
 106    FLAG = 6*N + K
        RETURN
        END
!C                            APPENDIX 1                          1/15/81
!C
!C        SUBROUTINES FOR SOLVING SPARSE NONSYMMETRIC SYSTEMS
!C        OF LINEAR EQUATIONS  (UNCOMPRESSED POINTER STORAGE)
!C
!C
!C*** SUBROUTINE sndrv
!C*** DRIVER FOR SUBROUTINES FOR SOLVING SPARSE NONSYMMETRIC SYSTEMS OF
!C       LINEAR EQUATIONS (UNCOMPRESSED POINTER STORAGE)
!C
        SUBROUTINE  sndrv
     *     (N, R,C,IC, IA,JA,A, B, Z, NSP,ISP,RSP,ESP, PATH, FLAG)
!C
!C    PARAMETERS
!C    CLASS ABBREVIATIONS ARE --
!C       N - INTEGER VARIABLE
!C       F - REAL VARIABLE
!C       V - SUPPLIES A VALUE TO THE DRIVER
!C       R - RETURNS A RESULT FROM THE DRIVER
!C       I - USED INTERNALLY BY THE DRIVER
!C       A - ARRAY
!C
!C CLASS | PARAMETER
!C ------+----------
!C       |
!C         THE NONZERO ENTRIES OF THE COEFFICIENT MATRIX M ARE STORED
!C    ROW-BY-ROW IN THE ARRAY A.  TO IDENTIFY THE INDIVIDUAL NONZERO
!C    ENTRIES IN EACH ROW, WE NEED TO KNOW IN WHICH COLUMN EACH ENTRY
!C    LIES.  THE COLUMN INDICES WHICH CORRESPOND TO THE NONZERO ENTRIES
!C    OF M ARE STORED IN THE ARRAY JA;  I.E., IF  A(K) = M(I,J),  THEN
!C    JA(K) = J.  IN ADDITION, WE NEED TO KNOW WHERE EACH ROW STARTS AND
!C    HOW LONG IT IS.  THE INDEX POSITIONS IN JA AND A WHERE THE ROWS OF
!C    M BEGIN ARE STORED IN THE ARRAY IA;  I.E., IF M(I,J) IS THE FIRST
!C    NONZERO ENTRY (STORED) IN THE I-TH ROW AND A(K) = M(I,J),  THEN
!C    IA(I) = K.  MOREOVER, THE INDEX IN JA AND A OF THE FIRST LOCATION
!c
!C    FOLLOWING THE LAST ELEMENT IN THE LAST ROW IS STORED IN IA(N+1).
!C    THUS, THE NUMBER OF ENTRIES IN THE I-TH ROW IS GIVEN BY
!C    IA(I+1) - IA(I),  THE NONZERO ENTRIES OF THE I-TH ROW ARE STORED
!C    CONSECUTIVELY IN
!C            A(IA(I)),  A(IA(I)+1),  ..., A(IA(I+1)-1),
!C    AND THE CORRESPONDING COLUMN INDICES ARE STORED CONSECUTIVELY IN
!C            JA(IA(I)), JA(IA(I)+1), ..., JA(IA(I+1)-1).
!C    FOR EXAMPLE, THE 5 BY 5 MATRIX
!C                ( 1. 0. 2. 0. 0.)
!C                ( 0. 3. 0. 0. 0.)
!C            M = ( 0. 4. 5. 6. 0.)
!C                ( 0. 0. 0. 7. 0.)
!C                ( 0. 0. 0. 8. 9.)
!C    WOULD BE STORED AS
!C               | 1  2  3  4  5  6  7  8  9
!C            ---+--------------------------
!C            IA | 1  3  4  7  8 10
!C            JA | 1  3  2  2  3  4  4  4  5
!C             A | 1. 2. 3. 4. 5. 6. 7. 8. 9.         .
!C
!C NV    | N     - NUMBER OF VARIABLES/EQUATIONS.
!C FVA   | A     - NONZERO ENTRIES OF THE COEFFICIENT MATRIX M, STORED
!C       |           BY ROWS.
!C       |           SIZE = NUMBER OF NONZERO ENTRIES IN M.
!C NVA   | IA    - POINTERS TO DELIMIT THE ROWS IN A.
!C       |           SIZE = N+1.
!C NVA   | JA    - COLUMN NUMBERS CORRESPONDING TO THE ELEMENTS OF A.
!C       |           SIZE = SIZE OF A.
!C FVA   | B     - RIGHT-HAND SIDE B;  B AND Z CAN THE SAME ARRAY.
!C       |           SIZE = N.
!C FRA   | Z     - SOLUTION X;  B AND Z CAN BE THE SAME ARRAY.
!C       |           SIZE = N.
!C
!C         THE ROWS AND COLUMNS OF THE ORIGINAL MATRIX M CAN BE
!C    REORDERED (E.G., TO REDUCE FILLIN OR ENSURE NUMERICAL STABILITY)
!C    BEFORE CALLING THE DRIVER.  IF NO REORDERING IS DONE, THEN SET
!C    R(I) = C(I) = IC(I) = I  FOR I=1,...,N.  THE SOLUTION Z IS RETURNED
!C    IN THE ORIGINAL ORDER.
!C
!C NVA   | R     - ORDERING OF THE ROWS OF M.
!C       |           SIZE = N.
!C NVA   | C     - ORDERING OF THE COLUMNS OF M.
!C       |           SIZE = N.
!C NVA   | IC    - INVERSE OF THE ORDERING OF THE COLUMNS OF M;  I.E.,
!C       |           IC(C(I)) = I  FOR I=1,...,N.
!C       |           SIZE = N.
!C
!C         THE SOLUTION OF THE SYSTEM OF LINEAR EQUATIONS IS DIVIDED INTO
!C    THREE STAGES --
!C      snsf -- THE MATRIX M IS PROCESSED SYMBOLICALLY TO DETERMINE WHERE
!C              FILLIN WILL OCCUR DURING THE NUMERIC FACTORIZATION.
!C      snnf -- THE MATRIX M IS FACTORED NUMERICALLY INTO THE PRODUCT LDU
!C              OF A UNIT LOWER TRIANGULAR MATRIX L, A DIAGONAL MATRIX D,
!C              AND A UNIT UPPER TRIANGULAR MATRIX U, AND THE SYSTEM
!C              MX = B  IS SOLVED.
!C      snns -- THE LINEAR SYSTEM  MX = B  IS SOLVED USING THE LDU
!C  OR          FACTORIZATION FROM snnf.
!C      snnt -- THE TRANSPOSED LINEAR SYSTEM  MT X = B  IS SOLVED USING
!C              THE LDU FACTORIZATION FROM snnf.
!C    FOR SEVERAL SYSTEMS WHOSE COEFFICIENT MATRICES HAVE THE SAME
!C    NONZERO STRUCTURE, snsf NEED BE DONE ONLY ONCE (FOR THE FIRST
!C    SYSTEM);  THEN snnf IS DONE ONCE FOR EACH ADDITIONAL SYSTEM.  FOR
!C    SEVERAL SYSTEMS WITH THE SAME COEFFICIENT MATRIX, snsf AND snnf NEED
!C    BE DONE ONLY ONCE (FOR THE FIRST SYSTEM);  THEN snns OR snnt IS DONE
!C    ONCE FOR EACH ADDITIONAL RIGHT-HAND SIDE.
!C
!C NV    | PATH  - PATH SPECIFICATION;  VALUES AND THEIR MEANINGS ARE --
!C       |           1  PERFORM snsf AND snnf.
!C       |           2  PERFORM snnf ONLY  (snsf IS ASSUMED TO HAVE BEEN
!C       |               DONE IN A MANNER COMPATIBLE WITH THE STORAGE
!C       |               ALLOCATION USED IN THE DRIVER).
!C       |           3  PERFORM snns ONLY  (snsf AND snnf ARE ASSUMED TO
!C       |               HAVE BEEN DONE IN A MANNER COMPATIBLE WITH THE
!C       |               STORAGE ALLOCATION USED IN THE DRIVER).
!C       |           4  PERFORM snnt ONLY  (snsf AND snnf ARE ASSUMED TO
!C       |               HAVE BEEN DONE IN A MANNER COMPATIBLE WITH THE
!C       |               STORAGE ALLOCATION USED IN THE DRIVER).
!C       |           5  PERFORM snsf ONLY.
!C
!C         VARIOUS ERRORS ARE DETECTED BY THE DRIVER AND THE INDIVIDUAL
!C    SUBROUTINES.
!C
!C NR    | FLAG  - ERROR FLAG;  VALUES AND THEIR MEANINGS ARE --
!C       |             0     NO ERRORS DETECTED
!C       |             N+K   NULL ROW IN A  --  ROW = K
!C       |            2N+K   DUPLICATE ENTRY IN A  --  ROW = K
!C       |            3N+K   INSUFFICIENT STORAGE IN snsf  --  ROW = K
!C       |            4N+1   INSUFFICIENT STORAGE IN snnf
!C       |            5N+K   NULL PIVOT  --  ROW = K
!C       |            6N+K   INSUFFICIENT STORAGE IN snsf  --  ROW = K
!C       |            7N+1   INSUFFICIENT STORAGE IN snnf
!C       |            8N+K   ZERO PIVOT  --  ROW = K
!C       |           10N+1   INSUFFICIENT STORAGE IN sndrv
!C       |           11N+1   ILLEGAL PATH SPECIFICATION
!C
!C         WORKING STORAGE IS NEEDED FOR THE FACTORED FORM OF THE MATRIX
!C    M PLUS VARIOUS TEMPORARY VECTORS.  THE ARRAYS ISP AND RSP SHOULD BE
!C    EQUIVALENCED;  INTEGER STORAGE IS ALLOCATED FROM THE BEGINNING OF
!C    ISP AND REAL STORAGE FROM THE END OF RSP.
!C
!C NV    | NSP   - DECLARED DIMENSION OF RSP;  NSP GENERALLY MUST
!C       |           BE LARGER THAN  5N+3 + 2K  (WHERE  K = (NUMBER OF
!C       |           NONZERO ENTRIES IN M)).
!C NVIRA | ISP   - INTEGER WORKING STORAGE DIVIDED UP INTO VARIOUS ARRAYS
!C       |           NEEDED BY THE SUBROUTINES;  ISP AND RSP SHOULD BE
!C       |           EQUIVALENCED.
!C       |           SIZE = 1*NSP
!C FVIRA | RSP   - REAL WORKING STORAGE DIVIDED UP INTO VARIOUS ARRAYS
!C       |           NEEDED BY THE SUBROUTINES;  ISP AND RSP SHOULD BE
!C       |           EQUIVALENCED.
!C       |           SIZE = NSP.
!C NR    | ESP   - IF SUFFICIENT STORAGE WAS AVAILABLE TO PERFORM THE
!C       |           SYMBOLIC FACTORIZATION (snsf), THEN ESP IS SET TO THE
!C       |           AMOUNT OF EXCESS STORAGE PROVIDED (NEGATIVE IF
!C       |           INSUFFICIENT STORAGE WAS AVAILABLE TO PERFORM THE
!C       |           NUMERIC FACTORIZATION (snnf)).
!C
!C
!C  CONVERSION TO DOUBLE PRECISION
!C
!C    TO CONVERT THESE ROUTINES FOR DOUBLE PRECISION ARRAYS, SIMPLY USE
!C    THE DOUBLE PRECISION DECLARATIONS IN PLACE OF THE REAL DECLARATIONS
!C    IN EACH SUBPROGRAM;  IN ADDITION, THE DATA VALUE OF THE INTEGER
!C    VARIABLE 1 MUST BE SET AS INDICATED IN SUBROUTINE sndrv
!C
        INTEGER  R(1), C(1), IC(1),  IA(1), JA(1),  ISP(1), ESP,
     *     PATH, FLAG,  Q, IM, D, U, ROW, TMP,  UMAX
        real*8    A(1),  B(1),  Z(1),  RSP(1)
        INTEGER  JU, JUMAX, JUTMP, LMAX, L, J, JLMAX, IL, NSP, N, MAX,
     *           JL, IU
!C
!C  SET 1 EQUAL TO THE RATIO BETWEEN THE LENGTH OF FLOATING POINT
!C  AND INTEGER ARRAY DATA;  E. G., 1 = 1 FOR (REAL, INTEGER),
!C  1 = 2 FOR (DOUBLE PRECISION, INTEGER)
!C
!c       DATA 1/2/
!C
        IF (PATH.LT.1 .OR. 5.LT.PATH)  GO TO 111
!C  ******  INITIALIZE AND DIVIDE UP TEMPORARY STORAGE  *****************
        IL = 1
        IU = IL + N+1
        JL = IU + N+1
!C
!C  ******  CALL snsf IF FLAG IS SET  ************************************
        IF ((PATH-1) * (PATH-5) .NE. 0)  GO TO 2
          MAX = (1*NSP + 1 - JL) - (N+1) - N
          JLMAX = MAX/2
          Q     = JL  + JLMAX
          IM    = Q   + (N+1)
          JUTMP = IM  +   N
          JUMAX = 1*NSP + 1 - JUTMP
          ESP = MAX/1
          IF (JLMAX.LE.0 .OR. JUMAX.LE.0)  GO TO 110
          CALL  snsf
     *       (N,  R, IC,  IA, JA,
     *        ISP(IL), ISP(JL), JLMAX,  ISP(IU), ISP(JUTMP), JUMAX,
     *        ISP(Q),  ISP(IM),  FLAG)
          IF (FLAG.NE.0)  GO TO 100
!C  ******  MOVE JU NEXT TO JL  *****************************************
          JLMAX = ISP(IL+N)-1
          JU    = JL + JLMAX
          JUMAX = ISP(IU+N)-1
          IF (JUMAX.LE.0)  GO TO 2
          DO 1 J=1,JUMAX
   1        ISP(JU+J-1) = ISP(JUTMP+J-1)
!C
!C  ******  CALL REMAINING SUBROUTINES  *********************************
   2    JLMAX = ISP(IL+N)-1
        JU    = JL  + JLMAX
        JUMAX = ISP(IU+N)-1
        L     = (JU + JUMAX - 2 + 1)  /  1    +    1
        LMAX  = JLMAX
        D     = L   + LMAX
        U     = D   + N
        ROW   = NSP + 1 - N
        TMP   = ROW - N
        UMAX  = TMP - U
        ESP = UMAX - JUMAX
!C
        IF ((PATH-1) * (PATH-2) .NE. 0)  GO TO 3
          IF (UMAX.LE.0)  GO TO 110
          CALL  snnf
     *       (N,  R, C, IC,  IA, JA, A,  Z,  B,
     *        ISP(IL), ISP(JL), RSP(L), LMAX,   RSP(D),
     *           ISP(IU), ISP(JU), RSP(U), UMAX,
     *        RSP(ROW),  RSP(TMP),  FLAG)
          IF (FLAG.NE.0)  GO TO 100
          RETURN
!C
   3    IF ((PATH-3) .NE. 0)  GO TO 4
          CALL  snns
     *       (N,  R, C,
     *        ISP(IL), ISP(JL), RSP(L),  RSP(D),
     *          ISP(IU), ISP(JU), RSP(U),
     *        Z,  B,  RSP(TMP))
!C
   4    IF ((PATH-4) .NE. 0)  GO TO 5
          CALL  snnt
     *       (N,  R, C,
     *        ISP(IL), ISP(JL), RSP(L),  RSP(D),
     *          ISP(IU), ISP(JU), RSP(U),
     *        Z,  B,  RSP(TMP))
   5    RETURN
!C
!C ** ERROR:  ERROR DETECTED IN snsf, snnf, snns, OR snnt
 100    RETURN
!C ** ERROR:  INSUFFICIENT STORAGE
 110    FLAG = 10*N + 1
        RETURN
!C ** ERROR:  ILLEGAL PATH SPECIFICATION
 111    FLAG = 11*N + 1
        RETURN
        END
!C
!C       ----------------------------------------------------------------
!C
!C*** SUBROUTINE snnf
!C*** NUMERIC LDU-FACTORIZATION OF SPARSE NONSYMMETRIC MATRIX AND
!C      SOLUTION OF SYSTEM OF LINEAR EQUATIONS (UNCOMPRESSED POINTER
!C      STORAGE)
!C
        SUBROUTINE  snnf
     *     (N, R,C,IC, IA,JA,A, Z, B, IL,JL,L,LMAX, D, IU,JU,U,UMAX,
     *      ROW, TMP, FLAG)
!C
!C       INPUT VARIABLES:   N, R,C,IC, IA,JA,A, B, IL,JL,LMAX, IU,JU,UMAX
!C       OUTPUT VARIABLES:  Z, L,D,U, FLAG
!C
!C       PARAMETERS USED INTERNALLY:
!C FIA   | ROW   - HOLDS INTERMEDIATE VALUES IN CALCULATION OF L, D, U.
!C       |           SIZE = N.
!C FIA   | TMP   - HOLDS NEW RIGHT-HAND SIDE B' FOR SOLUTION OF THE
!C       |           EQUATION  UX = B'.
!C       |           SIZE = N.
!C
        INTEGER  R(1), C(1), IC(1),  IA(1), JA(1),
     *     IL(1), JL(1), LMAX,  IU(1), JU(1), UMAX,  FLAG
        real*8    A(1), Z(1), B(1),  L(1), D(1), U(1),
     *     ROW(1), TMP(1),  LI, SUM, DK
        INTEGER  IMIN, J, I, IMAX, K, N, JMAX, JMIN
!C
!C  ******  CHECK STORAGE  **********************************************
        IF (IL(N+1)-1 .GT. LMAX)  GO TO 104
        IF (IU(N+1)-1 .GT. UMAX)  GO TO 107
!C
!C  ******  FOR EACH ROW  ***********************************************
        DO 10 K=1,N
!C  ******  SET THE INITIAL STRUCTURE OF ROW  ***************************
          JMIN = IL(K)
          JMAX = IL(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 2
!C  ******  IF L(K,M) .NE. 0, ROW(M)=0  *********************************
          DO 1 J=JMIN,JMAX
   1        ROW(JL(J)) = 0
   2      ROW(K) = 0
          JMIN = IU(K)
          JMAX = IU(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 4
!C  ******  IF U(K,M) .NE. 0, ROW(M)=0  *********************************
          DO 3 J=JMIN,JMAX
   3        ROW(JU(J)) = 0
   4      JMIN = IA(R(K))
          JMAX = IA(R(K)+1) - 1
!C  ******  SET ROW TO KTH ROW OF REORDERED A  **************************
          DO 5 J=JMIN,JMAX
   5        ROW(IC(JA(J))) = A(J)
!C  ******  INITIALIZE SUM  *********************************************
          SUM = B(R(K))
!C
!C  ******  ASSIGN THE KTH ROW OF L AND ADJUST ROW, SUM  ****************
          IMIN = IL(K)
          IMAX = IL(K+1) - 1
          IF (IMIN.GT.IMAX)  GO TO 8
          DO 7 I=IMIN,IMAX
            LI = - ROW(JL(I))
!C  ******  IF L IS NOT REQUIRED, THEN COMMENT OUT THE FOLLOWING LINE  **
            L(I) = - LI
            SUM = SUM + LI * TMP(JL(I))
            JMIN = IU(JL(I))
            JMAX = IU(JL(I)+1) - 1
            IF (JMIN.GT.JMAX)  GO TO 7
            DO 6 J=JMIN,JMAX
   6          ROW(JU(J)) = ROW(JU(J)) + LI * U(J)
   7        CONTINUE
!C
!C  ******  ASSIGN DIAGONAL D AND KTH ROW OF U, SET TMP(K)  *************
   8      IF (ROW(K).EQ.0)  GO TO 108
          DK = 1 / ROW(K)
          D(K) = DK
          TMP(K) = SUM * DK
          JMIN = IU(K)
          JMAX = IU(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 10
          DO 9 J=JMIN,JMAX
   9        U(J) = ROW(JU(J)) * DK
  10      CONTINUE
!C
!C  ******  SOLVE  UX = TMP  BY BACK SUBSTITUTION  **********************
        K = N
        DO 13 I=1,N
          SUM = TMP(K)
          JMIN = IU(K)
          JMAX = IU(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 12
          DO 11 J=JMIN,JMAX
  11        SUM = SUM - U(J) * TMP(JU(J))
  12      TMP(K) = SUM
          Z(C(K)) = SUM
  13      K = K-1
!C
        FLAG = 0
        RETURN
!C
!C ** ERROR:  INSUFFICIENT STORAGE FOR L
 104    FLAG = 4*N + 1
        RETURN
!C ** ERROR:  INSUFFICIENT STORAGE FOR U
 107    FLAG = 7*N + 1
        RETURN
!C ** ERROR:  ZERO PIVOT
 108    FLAG = 8*N + K
        RETURN
        END
!C
!C       ----------------------------------------------------------------
!C
!C*** SUBROUTINE snns
!C*** NUMERIC SOLUTION OF A SPARSE NONSYMMETRIC SYSTEM OF LINEAR
!C      EQUATIONS GIVEN LDU-FACTORIZATION (UNCOMPRESSED POINTER STORAGE)
!C
        SUBROUTINE  snns
     *     (N, R,C, IL,JL,L, D, IU,JU,U, Z, B, TMP)
!C
!C       INPUT VARIABLES:   N, R,C, IL,JL,L, D, IU,JU,U, B
!C       OUTPUT VARIABLES:  Z
!C
!C       PARAMETERS USED INTERNALLY:
!C FIA   | TMP   - HOLDS NEW RIGHT-HAND SIDE B' FOR SOLUTION OF THE
!C       |           EQUATION UX = B'.
!C       |           SIZE = N.
!C
        INTEGER  R(1), C(1),  IL(1), JL(1),  IU(1), JU(1)
        real*8    L(1), D(1), U(1),  Z(1), B(1),
     *           TMP(1), SUM
        INTEGER  JMAX, J, I, N, K, JMIN
!C
!C  ******  SOLVE LDY = B  BY FORWARD SUBSTITUTION  *********************
        DO 2 K=1,N
          SUM = B(R(K))
          JMIN = IL(K)
          JMAX = IL(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 2
          DO 1 J=JMIN,JMAX
   1        SUM = SUM - L(J) * TMP(JL(J))
   2      TMP(K) = SUM * D(K)
!C
!C  ******  SOLVE  UX = Y  BY BACK SUBSTITUTION  ************************
        K = N
        DO 5 I=1,N
          SUM = TMP(K)
          JMIN = IU(K)
          JMAX = IU(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 4
          DO 3 J=JMIN,JMAX
   3        SUM = SUM - U(J) * TMP(JU(J))
   4      TMP(K) = SUM
          Z(C(K)) = SUM
   5      K = K-1
        RETURN
        END
!C       ----------------------------------------------------------------
!C
!C*** SUBROUTINE snnt
!C*** NUMERIC SOLUTION OF THE TRANSPOSE OF A SPARSE NONSYMMETRIC SYSTEM
!C      OF LINEAR EQUATIONS GIVEN LDU-FACTORIZATION (UNCOMPRESSED POINTER
!C      STORAGE)
!C
        SUBROUTINE  snnt
     *     (N, R,C, IL,JL,L, D, IU,JU,U, Z, B, TMP)
!C
!C       INPUT VARIABLES:   N, R,C, IL,JL,L, D, IU,JU,U, B
!C       OUTPUT VARIABLES:  Z
!C
!C       PARAMETERS USED INTERNALLY:
!C FIA   | TMP   - HOLDS NEW RIGHT-HAND SIDE B' FOR SOLUTION OF THE
!C       |           EQUATION LX = B'.
!C       |           SIZE = N.
!C
        INTEGER  R(1), C(1),  IL(1), JL(1),  IU(1), JU(1)
        real*8    L(1), D(1), U(1),  Z(1), B(1),
     *           TMP(1), TMPK
        INTEGER  JMAX, J, I, N, K, JMIN
!C
!C  ******  SOLVE  UT Y = B  BY FORWARD SUBSTITUTION  *******************
        DO 1 K=1,N
   1      TMP(K) = B(C(K))
        DO 3 K=1,N
          TMPK = - TMP(K)
          JMIN = IU(K)
          JMAX = IU(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 3
          DO 2 J=JMIN,JMAX
   2        TMP(JU(J)) = TMP(JU(J)) + U(J) * TMPK
   3      CONTINUE
!C
!C  ******  SOLVE  D LT X = Y  BY BACK SUBSTITUTION  ********************
        K = N
        DO 6 I=1,N
          TMPK = - TMP(K) * D(K)
          JMIN = IL(K)
          JMAX = IL(K+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 5
          DO 4 J=JMIN,JMAX
   4        TMP(JL(J)) = TMP(JL(J)) + L(J) * TMPK
   5      Z(R(K)) = - TMPK
   6      K = K-1
        RETURN
        END
!C
!C       ----------------------------------------------------------------
!C
!C               YALE SPARSE MATRIX PACKAGE - NONSYMMETRIC CODES
!C                    SOLVING THE SYSTEM OF EQUATIONS MX = B
!C                        (UNCOMPRESSED POINTER STORAGE)
!C
!C    I.   CALLING SEQUENCES
!C         THE COEFFICIENT MATRIX CAN BE PROCESSED BY AN ORDERING ROUTINE
!C    (E.G., TO REDUCE FILLIN OR ENSURE NUMERICAL STABILITY) BEFORE USING
!C    THE REMAINING SUBROUTINES.  IF NO REORDERING IS DONE, THEN SET
!C    R(I) = C(I) = IC(I) = I  FOR I=1,...,N.  THE CALLING SEQUENCE IS --
!C        (      (MATRIX ORDERING))
!C         snsf   (SYMBOLIC FACTORIZATION TO DETERMINE WHERE FILLIN WILL
!C                 OCCUR DURING NUMERIC FACTORIZATION)
!C         snnf   (NUMERIC FACTORIZATION INTO PRODUCT LDU OF UNIT LOWER
!C                 TRIANGULAR MATRIX L, DIAGONAL MATRIX D, AND UNIT UPPER
!C                 TRIANGULAR MATRIX U, AND SOLUTION OF LINEAR SYSTEM)
!C         snns   (SOLUTION OF LINEAR SYSTEM FOR ADDITIONAL RIGHT-HAND
!C     OR          SIDE USING LDU FACTORIZATION FROM snnf)
!C         snnt   (SOLUTION OF TRANSPOSED LINEAR SYSTEM FOR ADDITIONAL
!C                 RIGHT-HAND SIDE USING LDU FACTORIZATION FROM snnf)
!C
!C    II.  STORAGE OF SPARSE MATRICES
!C         THE NONZERO ENTRIES OF THE COEFFICIENT MATRIX M ARE STORED
!C    ROW-BY-ROW IN THE ARRAY A.  TO IDENTIFY THE INDIVIDUAL NONZERO
!C    ENTRIES IN EACH ROW, WE NEED TO KNOW IN WHICH COLUMN EACH ENTRY
!C    LIES.  THE COLUMN INDICES WHICH CORRESPOND TO THE NONZERO ENTRIES
!C    OF M ARE STORED IN THE ARRAY JA;  I.E., IF  A(K) = M(I,J),  THEN
!C    JA(K) = J.  IN ADDITION, WE NEED TO KNOW WHERE EACH ROW STARTS AND
!C    HOW LONG IT IS.  THE INDEX POSITIONS IN JA AND A WHERE THE ROWS OF
!C    M BEGIN ARE STORED IN THE ARRAY IA;  I.E., IF M(I,J) IS THE FIRST
!C    IA(I) = K.  MOREOVER, THE INDEX IN JA AND A OF THE FIRST LOCATION
!C    FOLLOWING THE LAST ELEMENT IN THE LAST ROW IS STORED IN IA(N+1).
!C    THUS, THE NUMBER OF ENTRIES IN THE I-TH ROW IS GIVEN BY
!C    IA(I+1) - IA(I),  THE NONZERO ENTRIES OF THE I-TH ROW ARE STORED
!C    CONSECUTIVELY IN
!C            A(IA(I)),  A(IA(I)+1),  ..., A(IA(I+1)-1),
!C    AND THE CORRESPONDING COLUMN INDICES ARE STORED CONSECUTIVELY IN
!C            JA(IA(I)), JA(IA(I)+1), ..., JA(IA(I+1)-1).
!C    FOR EXAMPLE, THE 5 BY 5 MATRIX
!C                ( 1. 0. 2. 0. 0.)
!C                ( 0. 3. 0. 0. 0.)
!C            M = ( 0. 4. 5. 6. 0.)
!C                ( 0. 0. 0. 7. 0.)
!C                ( 0. 0. 0. 8. 9.)
!C    WOULD BE STORED AS
!C               | 1  2  3  4  5  6  7  8  9
!C            ---+--------------------------
!C            IA | 1  3  4  7  8 10
!C            JA | 1  3  2  2  3  4  4  4  5
!C             A | 1. 2. 3. 4. 5. 6. 7. 8. 9.         .
!C
!C         THE STRICT TRIANGULAR PORTIONS OF THE MATRICES L AND U ARE
!C    STORED IN THE SAME FASHION USING THE ARRAYS  IL, JL, L  AND
!C    IU, JU, U  RESPECTIVELY.  THE DIAGONAL ENTRIES OF L AND U ARE
!C    ASSUMED TO BE EQUAL TO ONE AND ARE NOT STORED.  THE ARRAY D
!C    CONTAINS THE RECIPROCALS OF THE DIAGONAL ENTRIES OF THE MATRIX D.
!C
!C    III. ADDITIONAL STORAGE SAVINGS
!C         IN snsf, R AND IC CAN BE THE SAME ARRAY IN THE CALLING
!C    SEQUENCE IF NO REORDERING OF THE COEFFICIENT MATRIX HAS BEEN DONE.
!C         IN snnf, R, C AND IC CAN ALL BE THE SAME ARRAY IF NO REORDERING
!C    HAS BEEN DONE.  IF ONLY THE ROWS HAVE BEEN REORDERED, THEN C AND IC
!C    CAN BE THE SAME ARRAY.  IF THE ROW AND COLUMN ORDERINGS ARE THE
!C    SAME, THEN R AND C CAN BE THE SAME ARRAY.  Z AND ROW CAN BE THE
!C    SAME ARRAY.
!C         IN snns OR snnt, R AND C CAN BE THE SAME ARRAY IF NO REORDERING
!C    HAS BEEN DONE OR IF THE ROW AND COLUMN ORDERINGS ARE THE SAME.  Z
!C    AND B CAN BE THE SAME ARRAY;  HOWEVER, THEN B WILL BE DESTROYED.
!C
!C    IV.  PARAMETERS
!C         FOLLOWING IS A LIST OF PARAMETERS TO THE PROGRAMS.  NAMES ARE
!C    UNIFORM AMONG THE VARIOUS SUBROUTINES.  CLASS ABBREVIATIONS ARE --
!C       N - INTEGER VARIABLE
!C       F - REAL VARIABLE
!C       V - SUPPLIES A VALUE TO A SUBROUTINE
!C       R - RETURNS A RESULT FROM A SUBROUTINE
!C       I - USED INTERNALLY BY A SUBROUTINE
!C       A - ARRAY
!C
!C CLASS | PARAMETER
!C ------+----------
!C FVA   | A     - NONZERO ENTRIES OF THE COEFFICIENT MATRIX M, STORED
!C       |           BY ROWS.
!C       |           SIZE = NUMBER OF NONZERO ENTRIES IN M.
!C FVA   | B     - RIGHT-HAND SIDE B.
!C       |           SIZE = N.
!C NVA   | C     - ORDERING OF THE COLUMNS OF M.
!C       |           SIZE = N.
!C FVRA  | D     - RECIPROCALS OF THE DIAGONAL ENTRIES OF THE MATRIX D.
!C       |           SIZE = N.
!C NR    | FLAG  - ERROR FLAG;  VALUES AND THEIR MEANINGS ARE --
!C       |            0     NO ERRORS DETECTED
!C       |            N+K   NULL ROW IN A  --  ROW = K
!C       |           2N+K   DUPLICATE ENTRY IN A  --  ROW = K
!C       |           3N+K   INSUFFICIENT STORAGE FOR JL  --  ROW = K
!C       |           4N+1   INSUFFICIENT STORAGE FOR L
!C       |           5N+K   NULL PIVOT  --  ROW = K
!C       |           6N+K   INSUFFICIENT STORAGE FOR JU  --  ROW = K
!C       |           7N+1   INSUFFICIENT STORAGE FOR U
!C       |           8N+K   ZERO PIVOT  --  ROW = K
!C NVA   | IA    - POINTERS TO DELIMIT THE ROWS IN A.
!C       |           SIZE = N+1.
!C NVA   | IC    - INVERSE OF THE ORDERING OF THE COLUMNS OF M;  I.E.,
!C       |           IC(C(I) = I  FOR I=1,...N.
!C       |           SIZE = N.
!C NVRA  | IL    - POINTERS TO DELIMIT THE ROWS IN L.
!C       |           SIZE = N+1.
!C NVRA  | IU    - POINTERS TO DELIMIT THE ROWS IN U.
!C       |           SIZE = N+1.
!C NVA   | JA    - COLUMN NUMBERS CORRESPONDING TO THE ELEMENTS OF A.
!C       |           SIZE = SIZE OF A.
!C NVRA  | JL    - COLUMN NUMBERS CORRESPONDING TO THE ELEMENTS OF L.
!C       |           SIZE = JLMAX.
!C NV    | JLMAX - DECLARED DIMENSION OF JL;  JLMAX MUST BE LARGER THAN
!C       |           THE NUMBER OF NONZERO ENTRIES IN THE STRICT LOWER
!C       |           TRIANGLE OF M PLUS FILLIN  (IL(N+1)-1 AFTER snsf).
!C NVRA  | JU    - COLUMN NUMBERS CORRESPONDING TO THE ELEMENTS OF U.
!C       |           SIZE = JUMAX.
!C NV    | JUMAX - DECLARED DIMENSION OF JU;  JUMAX MUST BE LARGER THAN
!C       |           THE NUMBER OF NONZERO ENTRIES IN THE STRICT UPPER
!C       |           TRIANGLE OF M PLUS FILLIN  (IU(N+1)-1 AFTER snsf).
!C FVRA  | L     - NONZERO ENTRIES IN THE STRICT LOWER TRIANGULAR PORTION
!C       |           OF THE MATRIX L, STORED BY ROWS.
!C       |           SIZE = LMAX
!C NV    | LMAX  - DECLARED DIMENSION OF L;  LMAX MUST BE LARGER THAN
!C       |           THE NUMBER OF NONZERO ENTRIES IN THE STRICT LOWER
!C       |           TRIANGLE OF M PLUS FILLIN  (IL(N+1)-1 AFTER snsf).
!C NV    | N     - NUMBER OF VARIABLES/EQUATIONS.
!C NVA   | R     - ORDERING OF THE ROWS OF M.
!C       |           SIZE = N.
!C FVRA  | U     - NONZERO ENTRIES IN THE STRICT UPPER TRIANGULAR PORTION
!C       |           OF THE MATRIX U, STORED BY ROWS.
!C       |           SIZE = UMAX.
!C NV    | UMAX  - DECLARED DIMENSION OF U;  UMAX MUST BE LARGER THAN
!C       |           THE NUMBER OF NONZERO ENTRIES IN THE STRICT UPPER
!C       |           TRIANGLE OF M PLUS FILLIN  (IU(N+1)-1 AFTER snsf).
!C FRA   | Z     - SOLUTION X.
!C       |           SIZE = N.
!C
!C
!C       ----------------------------------------------------------------
!C
!C*** SUBROUTINE snsf
!C*** SYMBOLIC LDU-FACTORIZATION OF A NONSYMMETRIC SPARSE MATRIX
!C      (UNCOMPRESSED POINTER STORAGE)
!C
        SUBROUTINE  snsf
     *     (N, R,IC, IA,JA, IL,JL,JLMAX, IU,JU,JUMAX, Q, IM, FLAG)
!C
!C       INPUT VARIABLES:   N, R,IC, IA,JA, JLMAX, JUMAX.
!C       OUTPUT VARIABLES:  IL,JL, IU,JU, FLAG.
!C
!C       PARAMETERS USED INTERNALLY:
!C NIA   | Q     - SUPPOSE M' IS THE RESULT OF REORDERING M;  IF
!C       |           PROCESSING OF THE KTH ROW OF M' (HENCE THE KTH ROWS
!C       |           OF L AND U) IS BEING DONE, THEN Q(J) IS INITIALLY
!C       |           NONZERO IF M'(K,J) IS NONZERO;  SINCE VALUES NEED
!C       |           NOT BE STORED, EACH ENTRY POINTS TO THE NEXT
!C       |           NONZERO;  FOR EXAMPLE, IF  N=9  AND THE 5TH ROW OF
!C       |           M' IS
!C       |                   0 X X 0 X 0 0 X 0,
!C       |           THEN Q WILL INITIALLY BE
!C       |                   A 3 5 A 8 A A 10 A 2        (A - ARBITRARY);
!C       |           Q(N+1) POINTS TO THE FIRST NONZERO IN THE ROW AND
!C       |           THE LAST NONZERO POINTS TO  N+1;  AS THE ALGORITHM
!C       |           PROCEEDS, OTHER ELEMENTS OF Q ARE INSERTED IN THE
!C       |           LIST BECAUSE OF FILLIN.
!C       |           SIZE = N+1.
!C NIA   | IM    - AT EACH STEP IN THE FACTORIZATION, IM(I) IS THE LAST
!C       |           ELEMENT IN THE ITH ROW OF U WHICH NEEDS TO BE
!C       |           CONSIDERED IN COMPUTING FILLIN.
!C       |           SIZE = N.
!C
!C  INTERNAL VARIABLES--
!C    JLPTR - POINTS TO THE LAST POSITION USED IN  JL.
!C    JUPTR - POINTS TO THE LAST POSITION USED IN  JU.
!C
        INTEGER  R(1), IC(1),  IA(1), JA(1),  IL(1), JL(1),
     *     IU(1), JU(1),  Q(1),  IM(1),  FLAG,  QM, VJ
        INTEGER  JMAX, JMIN, J, I, M, K, JLMAX, N, JUMAX, JUPTR, JLPTR
!C
!C  ******  INITIALIZE POINTERS  ****************************************
        JLPTR = 0
        IL(1) = 1
        JUPTR = 0
        IU(1) = 1
!C
!C  ******  FOR EACH ROW OF L AND U  ************************************
        DO 10 K=1,N
!C  ******  SET Q TO THE REORDERED ROW OF A  ****************************
          Q(N+1) = N+1
          JMIN = IA(R(K))
          JMAX = IA(R(K)+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 101
          DO 2 J=JMIN,JMAX
            VJ = IC(JA(J))
            QM = N+1
   1        M = QM
            QM = Q(M)
            IF (QM.LT.VJ)  GO TO 1
            IF (QM.EQ.VJ)  GO TO 102
              Q(M) = VJ
              Q(VJ) = QM
   2        CONTINUE
!C
!C  ******  FOR EACH ENTRY IN THE LOWER TRIANGLE  ***********************
          I = N+1
   3      I = Q(I)
          IF (I.GE.K)  GO TO 7
!C  ******  L(K,I) WILL BE NONZERO, SO ADD IT TO JL  ********************
            JLPTR = JLPTR+1
            IF (JLPTR.GT.JLMAX)  GO TO 103
            JL(JLPTR) = I
            QM = I
!C  ******  INSPECT ITH ROW FOR FILLIN, ADJUST IM IF POSSIBLE  **********
            JMIN = IU(I)
            JMAX = IM(I)
            IF (JMIN.GT.JMAX)  GO TO 6
            DO 5 J=JMIN,JMAX
              VJ = JU(J)
              IF (VJ.EQ.K)  IM(I) = J
   4          M = QM
              QM = Q(M)
              IF (QM.LT.VJ)  GO TO 4
              IF (QM.EQ.VJ)  GO TO 5
                Q(M) = VJ
                Q(VJ) = QM
                QM = VJ
   5          CONTINUE
   6        GO TO 3
!C
!C  ******  CHECK FOR NULL PIVOT  ***************************************
   7      IF (I.NE.K)  GO TO 105
!C  ******  REMAINING ELEMENTS OF Q DEFINE STRUCTURE OF U(K, )  *********
   8      I = Q(I)
          IF (I.GT.N)  GO TO 9
            JUPTR = JUPTR+1
            IF (JUPTR.GT.JUMAX)  GO TO 106
            JU(JUPTR) = I
            GO TO 8
!C  ******  GET READY FOR NEXT ROW  *************************************
   9      IM(K) = JUPTR
          IL(K+1) = JLPTR+1
  10      IU(K+1) = JUPTR+1
!C
        FLAG = 0
        RETURN
!C
!C ** ERROR:  NULL ROW IN A
 101    FLAG = N + R(K)
        RETURN
!C ** ERROR:  DUPLICATE ENTRY IN A
 102    FLAG = 2*N + R(K)
        RETURN
!C ** ERROR:  INSUFFICIENT STORAGE FOR JL
 103    FLAG = 3*N + K
        RETURN
!C ** ERROR:  NULL PIVOT
 105    FLAG = 5*N + K
        RETURN
!C ** ERROR:  INSUFFICIENT STORAGE FOR JU
 106    FLAG = 6*N + K
        RETURN
        END

!c **************************************************************************************


!C                                                                1/15/81
!C***********************************************************************
!C  zodrv -- DRIVER FOR SPARSE MATRIX REORDERING ROUTINES
!C***********************************************************************
        SUBROUTINE  zodrv
     *     (N, IA,JA,A, P,IP, NSP,ISP, PATH, FLAG)
!C
!C  DESCRIPTION
!C
!C    zodrv FINDS A MINIMUM DEGREE ORDERING OF THE ROWS AND COLUMNS OF A
!C    SYMMETRIC MATRIX M STORED IN (IA,JA,A) FORMAT (SEE BELOW).  FOR THE
!C    REORDERED MATRIX, THE WORK AND STORAGE REQUIRED TO PERFORM GAUSSIAN
!C    ELIMINATION IS (USUALLY) SIGNIFICANTLY LESS.
!C
!C    IF ONLY THE NONZERO ENTRIES IN THE UPPER TRIANGLE OF M ARE BEING
!C    STORED, THEN zodrv SYMMETRICALLY REORDERS (IA,JA,A), (OPTIONALLY)
!C    WITH THE DIAGONAL ENTRIES PLACED FIRST IN EACH ROW.  THIS IS TO
!C    ENSURE THAT IF M(I,J) WILL BE IN THE UPPER TRIANGLE OF M WITH
!C    RESPECT TO THE NEW ORDERING, THEN M(I,J) IS STORED IN ROW I (AND
!C    THUS M(J,I) IS NOT STORED);  WHEREAS IF M(I,J) WILL BE IN THE
!C    STRICT LOWER TRIANGLE OF M, THEN M(J,I) IS STORED IN ROW J (AND
!C    THUS M(I,J) IS NOT STORED).
!C
!C
!C  STORAGE OF SPARSE MATRICES
!C
!C    THE NONZERO ENTRIES OF THE MATRIX M ARE STORED ROW-BY-ROW IN THE
!C    ARRAY A.  TO IDENTIFY THE INDIVIDUAL NONZERO ENTRIES IN EACH ROW,
!C    WE NEED TO KNOW IN WHICH COLUMN EACH ENTRY LIES.  THESE COLUMN
!C    INDICES ARE STORED IN THE ARRAY JA;  I.E., IF  A(K) = M(I,J),  THEN
!C    JA(K) = J.  TO IDENTIFY THE INDIVIDUAL ROWS, WE NEED TO KNOW WHERE
!C    EACH ROW STARTS.  THESE ROW POINTERS ARE STORED IN THE ARRAY IA;
!C    I.E., IF M(I,J) IS THE FIRST NONZERO ENTRY (STORED) IN THE I-TH ROW
!C    AND  A(K) = M(I,J),  THEN  IA(I) = K.  MOREOVER, IA(N+1) POINTS TO
!C    THE FIRST LOCATION FOLLOWING THE LAST ELEMENT IN THE LAST ROW.
!C    THUS, THE NUMBER OF ENTRIES IN THE I-TH ROW IS  IA(I+1) - IA(I),
!C    THE NONZERO ENTRIES IN THE I-TH ROW ARE STORED CONSECUTIVELY IN
!C
!C            A(IA(I)),  A(IA(I)+1),  ..., A(IA(I+1)-1),
!C
!C    AND THE CORRESPONDING COLUMN INDICES ARE STORED CONSECUTIVELY IN
!C
!C            JA(IA(I)), JA(IA(I)+1), ..., JA(IA(I+1)-1).
!C
!C    SINCE THE COEFFICIENT MATRIX IS SYMMETRIC, ONLY THE NONZERO ENTRIES
!C    IN THE UPPER TRIANGLE NEED BE STORED.  FOR EXAMPLE, THE MATRIX
!C
!C             ( 1  0  2  3  0 )
!C             ( 0  4  0  0  0 )
!C         M = ( 2  0  5  6  0 )
!C             ( 3  0  6  7  8 )
!C             ( 0  0  0  8  9 )
!C
!C    COULD BE STORED AS
!C
!C            | 1  2  3  4  5  6  7  8  9 10 11 12 13
!C         ---+--------------------------------------
!C         IA | 1  4  5  8 12 14
!C         JA | 1  3  4  2  1  3  4  1  3  4  5  4  5
!C          A | 1  2  3  4  2  5  6  3  6  7  8  8  9
!C
!C    OR (SYMMETRICALLY) AS
!C
!C            | 1  2  3  4  5  6  7  8  9
!C         ---+--------------------------
!C         IA | 1  4  5  7  9 10
!C         JA | 1  3  4  2  3  4  4  5  5
!C          A | 1  2  3  4  5  6  7  8  9          .
!C
!C
!C  PARAMETERS
!C
!C    N    - ORDER OF THE MATRIX
!C
!C    IA   - INTEGER ONE-DIMENSIONAL ARRAY CONTAINING POINTERS TO DELIMIT
!C           ROWS IN JA AND A;  DIMENSION = N+1
!C
!C    JA   - INTEGER ONE-DIMENSIONAL ARRAY CONTAINING THE COLUMN INDICES
!C           CORRESPONDING TO THE ELEMENTS OF A;  DIMENSION = NUMBER OF
!C           NONZERO ENTRIES IN (THE UPPER TRIANGLE OF) M
!C
!C    A    - REAL ONE-DIMENSIONAL ARRAY CONTAINING THE NONZERO ENTRIES IN
!C           (THE UPPER TRIANGLE OF) M, STORED BY ROWS;  DIMENSION =
!C           NUMBER OF NONZERO ENTRIES IN (THE UPPER TRIANGLE OF) M
!C
!C    P    - INTEGER ONE-DIMENSIONAL ARRAY USED TO RETURN THE PERMUTATION
!C           OF THE ROWS AND COLUMNS OF M CORRESPONDING TO THE MINIMUM
!C           DEGREE ORDERING;  DIMENSION = N
!C
!C    IP   - INTEGER ONE-DIMENSIONAL ARRAY USED TO RETURN THE INVERSE OF
!C           THE PERMUTATION RETURNED IN P;  DIMENSION = N
!C
!C    NSP  - DECLARED DIMENSION OF THE ONE-DIMENSIONAL ARRAY ISP;  NSP
!C           MUST BE AT LEAST  3N+4K,  WHERE K IS THE NUMBER OF NONZEROES
!C           IN THE STRICT UPPER TRIANGLE OF M
!C
!C    ISP  - INTEGER ONE-DIMENSIONAL ARRAY USED FOR WORKING STORAGE;
!C           DIMENSION = NSP
!C
!C    PATH - INTEGER PATH SPECIFICATION;  VALUES AND THEIR MEANINGS ARE -
!C             1  FIND MINIMUM DEGREE ORDERING ONLY
!C             2  FIND MINIMUM DEGREE ORDERING AND REORDER SYMMETRICALLY
!C                  STORED MATRIX (USED WHEN ONLY THE NONZERO ENTRIES IN
!C                  THE UPPER TRIANGLE OF M ARE BEING STORED)
!C             3  REORDER SYMMETRICALLY STORED MATRIX AS SPECIFIED BY
!C                  INPUT PERMUTATION (USED WHEN AN ORDERING HAS ALREADY
!C                  BEEN DETERMINED AND ONLY THE NONZERO ENTRIES IN THE
!C                  UPPER TRIANGLE OF M ARE BEING STORED)
!C             4  SAME AS 2 BUT PUT DIAGONAL ENTRIES AT START OF EACH ROW
!C             5  SAME AS 3 BUT PUT DIAGONAL ENTRIES AT START OF EACH ROW
!C
!C    FLAG - INTEGER ERROR FLAG;  VALUES AND THEIR MEANINGS ARE -
!C               0    NO ERRORS DETECTED
!C              9N+K  INSUFFICIENT STORAGE IN MD
!C             10N+1  INSUFFICIENT STORAGE IN zodrv
!C             11N+1  ILLEGAL PATH SPECIFICATION
!C
!C
!C  CONVERSION FROM REAL TO DOUBLE PRECISION
!C
!C    CHANGE THE REAL DECLARATIONS IN zodrv AND zsro TO DOUBLE PRECISION
!C    DECLARATIONS.
!C
!C-----------------------------------------------------------------------
!C
        INTEGER  IA(1), JA(1),  P(1), IP(1),  ISP(1),  PATH,  FLAG,
     *     V, L, HEAD,  TMP, Q
        double complex    A(1)
        LOGICAL  DFLAG
        INTEGER  MAX, NEXT, N, NSP
!C
!C----INITIALIZE ERROR FLAG AND VALIDATE PATH SPECIFICATION
        FLAG = 0
        IF (PATH.LT.1 .OR. 5.LT.PATH)  GO TO 111
!C
!C----ALLOCATE STORAGE AND FIND MINIMUM DEGREE ORDERING
        IF ((PATH-1) * (PATH-2) * (PATH-4) .NE. 0)  GO TO 1
          MAX = (NSP-N)/2
          V    = 1
          L    = V     +  MAX
          HEAD = L     +  MAX
          NEXT = HEAD  +  N
          IF (MAX.LT.N)  GO TO 110
!C
          CALL  MD
     *       (N, IA,JA, MAX,ISP(V),ISP(L), ISP(HEAD),P,IP, ISP(V), FLAG)
          IF (FLAG.NE.0)  GO TO 100
!C
!C----ALLOCATE STORAGE AND SYMMETRICALLY REORDER MATRIX
   1    IF ((PATH-2) * (PATH-3) * (PATH-4) * (PATH-5) .NE. 0)  GO TO 2
          TMP = (NSP+1) -      N
          Q   = TMP     - (IA(N+1)-1)
          IF (Q.LT.1)  GO TO 110
!C
          DFLAG = PATH.EQ.4 .OR. PATH.EQ.5
          CALL zsro
     *       (N,  IP,  IA, JA, A,  ISP(TMP),  ISP(Q),  DFLAG)
!C
   2    RETURN
!C
!C ** ERROR -- ERROR DETECTED IN MD
 100    RETURN
!C ** ERROR -- INSUFFICIENT STORAGE
 110    FLAG = 10*N + 1
        RETURN
!C ** ERROR -- ILLEGAL PATH SPECIFIED
 111    FLAG = 11*N + 1
        RETURN
        END
!C                                                                1/15/81
!C***********************************************************************
!C  codrv -- DRIVER FOR SPARSE MATRIX REORDERING ROUTINES
!C***********************************************************************
        SUBROUTINE  codrv
     *     (N, IA,JA,A, P,IP, NSP,ISP, PATH, FLAG)
!C
!C  DESCRIPTION
!C
!C    codrv FINDS A MINIMUM DEGREE ORDERING OF THE ROWS AND COLUMNS OF A
!C    SYMMETRIC MATRIX M STORED IN (IA,JA,A) FORMAT (SEE BELOW).  FOR THE
!C    REORDERED MATRIX, THE WORK AND STORAGE REQUIRED TO PERFORM GAUSSIAN
!C    ELIMINATION IS (USUALLY) SIGNIFICANTLY LESS.
!C
!C    IF ONLY THE NONZERO ENTRIES IN THE UPPER TRIANGLE OF M ARE BEING
!C    STORED, THEN codrv SYMMETRICALLY REORDERS (IA,JA,A), (OPTIONALLY)
!C    WITH THE DIAGONAL ENTRIES PLACED FIRST IN EACH ROW.  THIS IS TO
!C    ENSURE THAT IF M(I,J) WILL BE IN THE UPPER TRIANGLE OF M WITH
!C    RESPECT TO THE NEW ORDERING, THEN M(I,J) IS STORED IN ROW I (AND
!C    THUS M(J,I) IS NOT STORED);  WHEREAS IF M(I,J) WILL BE IN THE
!C    STRICT LOWER TRIANGLE OF M, THEN M(J,I) IS STORED IN ROW J (AND
!C    THUS M(I,J) IS NOT STORED).
!C
!C
!C  STORAGE OF SPARSE MATRICES
!C
!C    THE NONZERO ENTRIES OF THE MATRIX M ARE STORED ROW-BY-ROW IN THE
!C    ARRAY A.  TO IDENTIFY THE INDIVIDUAL NONZERO ENTRIES IN EACH ROW,
!C    WE NEED TO KNOW IN WHICH COLUMN EACH ENTRY LIES.  THESE COLUMN
!C    INDICES ARE STORED IN THE ARRAY JA;  I.E., IF  A(K) = M(I,J),  THEN
!C    JA(K) = J.  TO IDENTIFY THE INDIVIDUAL ROWS, WE NEED TO KNOW WHERE
!C    EACH ROW STARTS.  THESE ROW POINTERS ARE STORED IN THE ARRAY IA;
!C    I.E., IF M(I,J) IS THE FIRST NONZERO ENTRY (STORED) IN THE I-TH ROW
!C    AND  A(K) = M(I,J),  THEN  IA(I) = K.  MOREOVER, IA(N+1) POINTS TO
!C    THE FIRST LOCATION FOLLOWING THE LAST ELEMENT IN THE LAST ROW.
!C    THUS, THE NUMBER OF ENTRIES IN THE I-TH ROW IS  IA(I+1) - IA(I),
!C    THE NONZERO ENTRIES IN THE I-TH ROW ARE STORED CONSECUTIVELY IN
!C
!C            A(IA(I)),  A(IA(I)+1),  ..., A(IA(I+1)-1),
!C
!C    AND THE CORRESPONDING COLUMN INDICES ARE STORED CONSECUTIVELY IN
!C
!C            JA(IA(I)), JA(IA(I)+1), ..., JA(IA(I+1)-1).
!C
!C    SINCE THE COEFFICIENT MATRIX IS SYMMETRIC, ONLY THE NONZERO ENTRIES
!C    IN THE UPPER TRIANGLE NEED BE STORED.  FOR EXAMPLE, THE MATRIX
!C
!C             ( 1  0  2  3  0 )
!C             ( 0  4  0  0  0 )
!C         M = ( 2  0  5  6  0 )
!C             ( 3  0  6  7  8 )
!C             ( 0  0  0  8  9 )
!C
!C    COULD BE STORED AS
!C
!C            | 1  2  3  4  5  6  7  8  9 10 11 12 13
!C         ---+--------------------------------------
!C         IA | 1  4  5  8 12 14
!C         JA | 1  3  4  2  1  3  4  1  3  4  5  4  5
!C          A | 1  2  3  4  2  5  6  3  6  7  8  8  9
!C
!C    OR (SYMMETRICALLY) AS
!C
!C            | 1  2  3  4  5  6  7  8  9
!C         ---+--------------------------
!C         IA | 1  4  5  7  9 10
!C         JA | 1  3  4  2  3  4  4  5  5
!C          A | 1  2  3  4  5  6  7  8  9          .
!C
!C
!C  PARAMETERS
!C
!C    N    - ORDER OF THE MATRIX
!C
!C    IA   - INTEGER ONE-DIMENSIONAL ARRAY CONTAINING POINTERS TO DELIMIT
!C           ROWS IN JA AND A;  DIMENSION = N+1
!C
!C    JA   - INTEGER ONE-DIMENSIONAL ARRAY CONTAINING THE COLUMN INDICES
!C           CORRESPONDING TO THE ELEMENTS OF A;  DIMENSION = NUMBER OF
!C           NONZERO ENTRIES IN (THE UPPER TRIANGLE OF) M
!C
!C    A    - REAL ONE-DIMENSIONAL ARRAY CONTAINING THE NONZERO ENTRIES IN
!C           (THE UPPER TRIANGLE OF) M, STORED BY ROWS;  DIMENSION =
!C           NUMBER OF NONZERO ENTRIES IN (THE UPPER TRIANGLE OF) M
!C
!C    P    - INTEGER ONE-DIMENSIONAL ARRAY USED TO RETURN THE PERMUTATION
!C           OF THE ROWS AND COLUMNS OF M CORRESPONDING TO THE MINIMUM
!C           DEGREE ORDERING;  DIMENSION = N
!C
!C    IP   - INTEGER ONE-DIMENSIONAL ARRAY USED TO RETURN THE INVERSE OF
!C           THE PERMUTATION RETURNED IN P;  DIMENSION = N
!C
!C    NSP  - DECLARED DIMENSION OF THE ONE-DIMENSIONAL ARRAY ISP;  NSP
!C           MUST BE AT LEAST  3N+4K,  WHERE K IS THE NUMBER OF NONZEROES
!C           IN THE STRICT UPPER TRIANGLE OF M
!C
!C    ISP  - INTEGER ONE-DIMENSIONAL ARRAY USED FOR WORKING STORAGE;
!C           DIMENSION = NSP
!C
!C    PATH - INTEGER PATH SPECIFICATION;  VALUES AND THEIR MEANINGS ARE -
!C             1  FIND MINIMUM DEGREE ORDERING ONLY
!C             2  FIND MINIMUM DEGREE ORDERING AND REORDER SYMMETRICALLY
!C                  STORED MATRIX (USED WHEN ONLY THE NONZERO ENTRIES IN
!C                  THE UPPER TRIANGLE OF M ARE BEING STORED)
!C             3  REORDER SYMMETRICALLY STORED MATRIX AS SPECIFIED BY
!C                  INPUT PERMUTATION (USED WHEN AN ORDERING HAS ALREADY
!C                  BEEN DETERMINED AND ONLY THE NONZERO ENTRIES IN THE
!C                  UPPER TRIANGLE OF M ARE BEING STORED)
!C             4  SAME AS 2 BUT PUT DIAGONAL ENTRIES AT START OF EACH ROW
!C             5  SAME AS 3 BUT PUT DIAGONAL ENTRIES AT START OF EACH ROW
!C
!C    FLAG - INTEGER ERROR FLAG;  VALUES AND THEIR MEANINGS ARE -
!C               0    NO ERRORS DETECTED
!C              9N+K  INSUFFICIENT STORAGE IN MD
!C             10N+1  INSUFFICIENT STORAGE IN codrv
!C             11N+1  ILLEGAL PATH SPECIFICATION
!C
!C
!C  CONVERSION FROM REAL TO DOUBLE PRECISION
!C
!C    CHANGE THE REAL DECLARATIONS IN codrv AND csro TO DOUBLE PRECISION
!C    DECLARATIONS.
!C
!C-----------------------------------------------------------------------
!C
        INTEGER  IA(1), JA(1),  P(1), IP(1),  ISP(1),  PATH,  FLAG,
     *     V, L, HEAD,  TMP, Q
        complex*16    A(1)
        LOGICAL  DFLAG
        INTEGER  MAX, NEXT, N, NSP
!C
!C----INITIALIZE ERROR FLAG AND VALIDATE PATH SPECIFICATION
        FLAG = 0
        IF (PATH.LT.1 .OR. 5.LT.PATH)  GO TO 111
!C
!C----ALLOCATE STORAGE AND FIND MINIMUM DEGREE ORDERING
        IF ((PATH-1) * (PATH-2) * (PATH-4) .NE. 0)  GO TO 1
          MAX = (NSP-N)/2
          V    = 1
          L    = V     +  MAX
          HEAD = L     +  MAX
          NEXT = HEAD  +  N
          IF (MAX.LT.N)  GO TO 110
!C
          CALL  MD
     *       (N, IA,JA, MAX,ISP(V),ISP(L), ISP(HEAD),P,IP, ISP(V), FLAG)
          IF (FLAG.NE.0)  GO TO 100
!C
!C----ALLOCATE STORAGE AND SYMMETRICALLY REORDER MATRIX
   1    IF ((PATH-2) * (PATH-3) * (PATH-4) * (PATH-5) .NE. 0)  GO TO 2
          TMP = (NSP+1) -      N
          Q   = TMP     - (IA(N+1)-1)
          IF (Q.LT.1)  GO TO 110
!C
          DFLAG = PATH.EQ.4 .OR. PATH.EQ.5
          CALL csro
     *       (N,  IP,  IA, JA, A,  ISP(TMP),  ISP(Q),  DFLAG)
!C
   2    RETURN
!C
!C ** ERROR -- ERROR DETECTED IN MD
 100    RETURN
!C ** ERROR -- INSUFFICIENT STORAGE
 110    FLAG = 10*N + 1
        RETURN
!C ** ERROR -- ILLEGAL PATH SPECIFIED
 111    FLAG = 11*N + 1
        RETURN
        END
!C                                                                1/15/81
!C***********************************************************************
!C  dodrv -- DRIVER FOR SPARSE MATRIX REORDERING ROUTINES
!C***********************************************************************
        SUBROUTINE  dodrv
     *     (N, IA,JA,A, P,IP, NSP,ISP, PATH, FLAG)
!C
!C  DESCRIPTION
!C
!C    dodrv FINDS A MINIMUM DEGREE ORDERING OF THE ROWS AND COLUMNS OF A
!C    SYMMETRIC MATRIX M STORED IN (IA,JA,A) FORMAT (SEE BELOW).  FOR THE
!C    REORDERED MATRIX, THE WORK AND STORAGE REQUIRED TO PERFORM GAUSSIAN
!c    ELIMINATION IS (USUALLY) SIGNIFICANTLY LESS.
!C
!C    IF ONLY THE NONZERO ENTRIES IN THE UPPER TRIANGLE OF M ARE BEING
!C    STORED, THEN dodrv SYMMETRICALLY REORDERS (IA,JA,A), (OPTIONALLY)
!C    WITH THE DIAGONAL ENTRIES PLACED FIRST IN EACH ROW.  THIS IS TO
!C    ENSURE THAT IF M(I,J) WILL BE IN THE UPPER TRIANGLE OF M WITH
!C    RESPECT TO THE NEW ORDERING, THEN M(I,J) IS STORED IN ROW I (AND
!C    THUS M(J,I) IS NOT STORED);  WHEREAS IF M(I,J) WILL BE IN THE
!C    STRICT LOWER TRIANGLE OF M, THEN M(J,I) IS STORED IN ROW J (AND
!C    THUS M(I,J) IS NOT STORED).
!C
!C
!C  STORAGE OF SPARSE MATRICES
!C
!C    THE NONZERO ENTRIES OF THE MATRIX M ARE STORED ROW-BY-ROW IN THE
!C    ARRAY A.  TO IDENTIFY THE INDIVIDUAL NONZERO ENTRIES IN EACH ROW,
!C    WE NEED TO KNOW IN WHICH COLUMN EACH ENTRY LIES.  THESE COLUMN
!C    INDICES ARE STORED IN THE ARRAY JA;  I.E., IF  A(K) = M(I,J),  THEN
!C    JA(K) = J.  TO IDENTIFY THE INDIVIDUAL ROWS, WE NEED TO KNOW WHERE
!C    EACH ROW STARTS.  THESE ROW POINTERS ARE STORED IN THE ARRAY IA;
!C    I.E., IF M(I,J) IS THE FIRST NONZERO ENTRY (STORED) IN THE I-TH ROW
!C    AND  A(K) = M(I,J),  THEN  IA(I) = K.  MOREOVER, IA(N+1) POINTS TO
!C    THE FIRST LOCATION FOLLOWING THE LAST ELEMENT IN THE LAST ROW.
!C    THUS, THE NUMBER OF ENTRIES IN THE I-TH ROW IS  IA(I+1) - IA(I),
!C    THE NONZERO ENTRIES IN THE I-TH ROW ARE STORED CONSECUTIVELY IN
!C
!C            A(IA(I)),  A(IA(I)+1),  ..., A(IA(I+1)-1),
!C
!C    AND THE CORRESPONDING COLUMN INDICES ARE STORED CONSECUTIVELY IN
!C
!C            JA(IA(I)), JA(IA(I)+1), ..., JA(IA(I+1)-1).
!C
!C    SINCE THE COEFFICIENT MATRIX IS SYMMETRIC, ONLY THE NONZERO ENTRIES
!C    IN THE UPPER TRIANGLE NEED BE STORED.  FOR EXAMPLE, THE MATRIX
!C
!C             ( 1  0  2  3  0 )
!C             ( 0  4  0  0  0 )
!C         M = ( 2  0  5  6  0 )
!C             ( 3  0  6  7  8 )
!C             ( 0  0  0  8  9 )
!C
!C    COULD BE STORED AS
!C
!C            | 1  2  3  4  5  6  7  8  9 10 11 12 13
!C         ---+--------------------------------------
!C         IA | 1  4  5  8 12 14
!C         JA | 1  3  4  2  1  3  4  1  3  4  5  4  5
!C          A | 1  2  3  4  2  5  6  3  6  7  8  8  9
!C
!C    OR (SYMMETRICALLY) AS
!C
!C            | 1  2  3  4  5  6  7  8  9
!C         ---+--------------------------
!C         IA | 1  4  5  7  9 10
!C         JA | 1  3  4  2  3  4  4  5  5
!C          A | 1  2  3  4  5  6  7  8  9          .
!C
!C
!C  PARAMETERS
!C
!C    N    - ORDER OF THE MATRIX
!C
!C    IA   - INTEGER ONE-DIMENSIONAL ARRAY CONTAINING POINTERS TO DELIMIT
!C           ROWS IN JA AND A;  DIMENSION = N+1
!C
!C    JA   - INTEGER ONE-DIMENSIONAL ARRAY CONTAINING THE COLUMN INDICES
!C           CORRESPONDING TO THE ELEMENTS OF A;  DIMENSION = NUMBER OF
!C           NONZERO ENTRIES IN (THE UPPER TRIANGLE OF) M
!C
!C    A    - REAL ONE-DIMENSIONAL ARRAY CONTAINING THE NONZERO ENTRIES IN
!C           (THE UPPER TRIANGLE OF) M, STORED BY ROWS;  DIMENSION =
!C           NUMBER OF NONZERO ENTRIES IN (THE UPPER TRIANGLE OF) M
!C
!C    P    - INTEGER ONE-DIMENSIONAL ARRAY USED TO RETURN THE PERMUTATION
!C           OF THE ROWS AND COLUMNS OF M CORRESPONDING TO THE MINIMUM
!C           DEGREE ORDERING;  DIMENSION = N
!C
!C    IP   - INTEGER ONE-DIMENSIONAL ARRAY USED TO RETURN THE INVERSE OF
!C           THE PERMUTATION RETURNED IN P;  DIMENSION = N
!C
!C    NSP  - DECLARED DIMENSION OF THE ONE-DIMENSIONAL ARRAY ISP;  NSP
!C           MUST BE AT LEAST  3N+4K,  WHERE K IS THE NUMBER OF NONZEROES
!C           IN THE STRICT UPPER TRIANGLE OF M
!C
!C    ISP  - INTEGER ONE-DIMENSIONAL ARRAY USED FOR WORKING STORAGE;
!C           DIMENSION = NSP
!C
!C    PATH - INTEGER PATH SPECIFICATION;  VALUES AND THEIR MEANINGS ARE -
!C             1  FIND MINIMUM DEGREE ORDERING ONLY
!C             2  FIND MINIMUM DEGREE ORDERING AND REORDER SYMMETRICALLY
!C                  STORED MATRIX (USED WHEN ONLY THE NONZERO ENTRIES IN
!C                  THE UPPER TRIANGLE OF M ARE BEING STORED)
!C             3  REORDER SYMMETRICALLY STORED MATRIX AS SPECIFIED BY
!C                  INPUT PERMUTATION (USED WHEN AN ORDERING HAS ALREADY
!C                  BEEN DETERMINED AND ONLY THE NONZERO ENTRIES IN THE
!C                  UPPER TRIANGLE OF M ARE BEING STORED)
!C             4  SAME AS 2 BUT PUT DIAGONAL ENTRIES AT START OF EACH ROW
!C             5  SAME AS 3 BUT PUT DIAGONAL ENTRIES AT START OF EACH ROW
!C
!C    FLAG - INTEGER ERROR FLAG;  VALUES AND THEIR MEANINGS ARE -
!C               0    NO ERRORS DETECTED
!C              9N+K  INSUFFICIENT STORAGE IN MD
!C             10N+1  INSUFFICIENT STORAGE IN dodrv
!C             11N+1  ILLEGAL PATH SPECIFICATION
!C
!C
!C  CONVERSION FROM REAL TO DOUBLE PRECISION
!C
!C    CHANGE THE REAL DECLARATIONS IN dodrv AND dsro TO DOUBLE PRECISION
!C    DECLARATIONS.
!C
!C-----------------------------------------------------------------------
!C
        INTEGER  IA(1), JA(1),  P(1), IP(1),  ISP(1),  PATH,  FLAG,
     *     V, L, HEAD,  TMP, Q
        double precision    A(1)
        LOGICAL  DFLAG
        INTEGER  MAX, NEXT, N, NSP
!C
!C----INITIALIZE ERROR FLAG AND VALIDATE PATH SPECIFICATION
        FLAG = 0
        IF (PATH.LT.1 .OR. 5.LT.PATH)  GO TO 111
!C
!C----ALLOCATE STORAGE AND FIND MINIMUM DEGREE ORDERING
        IF ((PATH-1) * (PATH-2) * (PATH-4) .NE. 0)  GO TO 1
          MAX = (NSP-N)/2
          V    = 1
          L    = V     +  MAX
          HEAD = L     +  MAX
          NEXT = HEAD  +  N
          IF (MAX.LT.N)  GO TO 110
!C
          CALL  MD
     *       (N, IA,JA, MAX,ISP(V),ISP(L), ISP(HEAD),P,IP, ISP(V), FLAG)
          IF (FLAG.NE.0)  GO TO 100
!C
!C----ALLOCATE STORAGE AND SYMMETRICALLY REORDER MATRIX
   1    IF ((PATH-2) * (PATH-3) * (PATH-4) * (PATH-5) .NE. 0)  GO TO 2
          TMP = (NSP+1) -      N
          Q   = TMP     - (IA(N+1)-1)
          IF (Q.LT.1)  GO TO 110
!C
          DFLAG = PATH.EQ.4 .OR. PATH.EQ.5
          CALL dsro
     *       (N,  IP,  IA, JA, A,  ISP(TMP),  ISP(Q),  DFLAG)
!C
   2    RETURN
!C
!C ** ERROR -- ERROR DETECTED IN MD
 100    RETURN
!C ** ERROR -- INSUFFICIENT STORAGE
 110    FLAG = 10*N + 1
        RETURN
!C ** ERROR -- ILLEGAL PATH SPECIFIED
 111    FLAG = 11*N + 1
        RETURN
        END
!C                                                                1/15/81
!C***********************************************************************
!C  sodrv -- DRIVER FOR SPARSE MATRIX REORDERING ROUTINES
!C***********************************************************************
        SUBROUTINE  sodrv
     *     (N, IA,JA,A, P,IP, NSP,ISP, PATH, FLAG)
!C
!C  DESCRIPTION
!C
!C    sodrv FINDS A MINIMUM DEGREE ORDERING OF THE ROWS AND COLUMNS OF A
!C    SYMMETRIC MATRIX M STORED IN (IA,JA,A) FORMAT (SEE BELOW).  FOR THE
!C    REORDERED MATRIX, THE WORK AND STORAGE REQUIRED TO PERFORM GAUSSIAN
!C    ELIMINATION IS (USUALLY) SIGNIFICANTLY LESS.
!C
!C    IF ONLY THE NONZERO ENTRIES IN THE UPPER TRIANGLE OF M ARE BEING
!C    STORED, THEN sodrv SYMMETRICALLY REORDERS (IA,JA,A), (OPTIONALLY)
!C    WITH THE DIAGONAL ENTRIES PLACED FIRST IN EACH ROW.  THIS IS TO
!C    ENSURE THAT IF M(I,J) WILL BE IN THE UPPER TRIANGLE OF M WITH
!C    RESPECT TO THE NEW ORDERING, THEN M(I,J) IS STORED IN ROW I (AND
!C    THUS M(J,I) IS NOT STORED);  WHEREAS IF M(I,J) WILL BE IN THE
!C    STRICT LOWER TRIANGLE OF M, THEN M(J,I) IS STORED IN ROW J (AND
!C    THUS M(I,J) IS NOT STORED).
!C
!C
!C  STORAGE OF SPARSE MATRICES
!C
!C    THE NONZERO ENTRIES OF THE MATRIX M ARE STORED ROW-BY-ROW IN THE
!C    ARRAY A.  TO IDENTIFY THE INDIVIDUAL NONZERO ENTRIES IN EACH ROW,
!C    WE NEED TO KNOW IN WHICH COLUMN EACH ENTRY LIES.  THESE COLUMN
!C    INDICES ARE STORED IN THE ARRAY JA;  I.E., IF  A(K) = M(I,J),  THEN
!C    JA(K) = J.  TO IDENTIFY THE INDIVIDUAL ROWS, WE NEED TO KNOW WHERE
!C    EACH ROW STARTS.  THESE ROW POINTERS ARE STORED IN THE ARRAY IA;
!C    I.E., IF M(I,J) IS THE FIRST NONZERO ENTRY (STORED) IN THE I-TH ROW
!C    AND  A(K) = M(I,J),  THEN  IA(I) = K.  MOREOVER, IA(N+1) POINTS TO
!C    THE FIRST LOCATION FOLLOWING THE LAST ELEMENT IN THE LAST ROW.
!C    THUS, THE NUMBER OF ENTRIES IN THE I-TH ROW IS  IA(I+1) - IA(I),
!C    THE NONZERO ENTRIES IN THE I-TH ROW ARE STORED CONSECUTIVELY IN
!C
!C            A(IA(I)),  A(IA(I)+1),  ..., A(IA(I+1)-1),
!C
!C    AND THE CORRESPONDING COLUMN INDICES ARE STORED CONSECUTIVELY IN
!C
!C            JA(IA(I)), JA(IA(I)+1), ..., JA(IA(I+1)-1).
!C
!C    SINCE THE COEFFICIENT MATRIX IS SYMMETRIC, ONLY THE NONZERO ENTRIES
!C    IN THE UPPER TRIANGLE NEED BE STORED.  FOR EXAMPLE, THE MATRIX
!C
!C             ( 1  0  2  3  0 )
!C             ( 0  4  0  0  0 )
!C         M = ( 2  0  5  6  0 )
!C             ( 3  0  6  7  8 )
!C             ( 0  0  0  8  9 )
!C
!C    COULD BE STORED AS
!C
!C            | 1  2  3  4  5  6  7  8  9 10 11 12 13
!C         ---+--------------------------------------
!C         IA | 1  4  5  8 12 14
!C         JA | 1  3  4  2  1  3  4  1  3  4  5  4  5
!C          A | 1  2  3  4  2  5  6  3  6  7  8  8  9
!C
!C    OR (SYMMETRICALLY) AS
!C
!C            | 1  2  3  4  5  6  7  8  9
!C         ---+--------------------------
!C         IA | 1  4  5  7  9 10
!C         JA | 1  3  4  2  3  4  4  5  5
!C          A | 1  2  3  4  5  6  7  8  9          .
!C
!C
!C  PARAMETERS
!C
!C    N    - ORDER OF THE MATRIX
!C
!C    IA   - INTEGER ONE-DIMENSIONAL ARRAY CONTAINING POINTERS TO DELIMIT
!C           ROWS IN JA AND A;  DIMENSION = N+1
!C
!C    JA   - INTEGER ONE-DIMENSIONAL ARRAY CONTAINING THE COLUMN INDICES
!C           CORRESPONDING TO THE ELEMENTS OF A;  DIMENSION = NUMBER OF
!C           NONZERO ENTRIES IN (THE UPPER TRIANGLE OF) M
!C
!C    A    - REAL ONE-DIMENSIONAL ARRAY CONTAINING THE NONZERO ENTRIES IN
!C           (THE UPPER TRIANGLE OF) M, STORED BY ROWS;  DIMENSION =
!C           NUMBER OF NONZERO ENTRIES IN (THE UPPER TRIANGLE OF) M
!C
!C    P    - INTEGER ONE-DIMENSIONAL ARRAY USED TO RETURN THE PERMUTATION
!C           OF THE ROWS AND COLUMNS OF M CORRESPONDING TO THE MINIMUM
!C           DEGREE ORDERING;  DIMENSION = N
!C
!C    IP   - INTEGER ONE-DIMENSIONAL ARRAY USED TO RETURN THE INVERSE OF
!C           THE PERMUTATION RETURNED IN P;  DIMENSION = N
!C
!C    NSP  - DECLARED DIMENSION OF THE ONE-DIMENSIONAL ARRAY ISP;  NSP
!C           MUST BE AT LEAST  3N+4K,  WHERE K IS THE NUMBER OF NONZEROES
!C           IN THE STRICT UPPER TRIANGLE OF M
!C
!C    ISP  - INTEGER ONE-DIMENSIONAL ARRAY USED FOR WORKING STORAGE;
!C           DIMENSION = NSP
!C
!C    PATH - INTEGER PATH SPECIFICATION;  VALUES AND THEIR MEANINGS ARE -
!C             1  FIND MINIMUM DEGREE ORDERING ONLY
!C             2  FIND MINIMUM DEGREE ORDERING AND REORDER SYMMETRICALLY
!C                  STORED MATRIX (USED WHEN ONLY THE NONZERO ENTRIES IN
!C                  THE UPPER TRIANGLE OF M ARE BEING STORED)
!C             3  REORDER SYMMETRICALLY STORED MATRIX AS SPECIFIED BY
!C                  INPUT PERMUTATION (USED WHEN AN ORDERING HAS ALREADY
!C                  BEEN DETERMINED AND ONLY THE NONZERO ENTRIES IN THE
!C                  UPPER TRIANGLE OF M ARE BEING STORED)
!C             4  SAME AS 2 BUT PUT DIAGONAL ENTRIES AT START OF EACH ROW
!C             5  SAME AS 3 BUT PUT DIAGONAL ENTRIES AT START OF EACH ROW
!C
!C    FLAG - INTEGER ERROR FLAG;  VALUES AND THEIR MEANINGS ARE -
!C               0    NO ERRORS DETECTED
!C              9N+K  INSUFFICIENT STORAGE IN MD
!C             10N+1  INSUFFICIENT STORAGE IN sodrv
!C             11N+1  ILLEGAL PATH SPECIFICATION
!C
!C
!C  CONVERSION FROM REAL TO DOUBLE PRECISION
!C
!C    CHANGE THE REAL DECLARATIONS IN sodrv AND ssro TO DOUBLE PRECISION
!C    DECLARATIONS.
!C
!C-----------------------------------------------------------------------
!C
        INTEGER  IA(1), JA(1),  P(1), IP(1),  ISP(1),  PATH,  FLAG,
     *     V, L, HEAD,  TMP, Q
        real*8    A(1)
        LOGICAL  DFLAG
        INTEGER  MAX, NEXT, N, NSP
!C
!C----INITIALIZE ERROR FLAG AND VALIDATE PATH SPECIFICATION
        FLAG = 0
        IF (PATH.LT.1 .OR. 5.LT.PATH)  GO TO 111
!C
!C----ALLOCATE STORAGE AND FIND MINIMUM DEGREE ORDERING
        IF ((PATH-1) * (PATH-2) * (PATH-4) .NE. 0)  GO TO 1
          MAX = (NSP-N)/2
          V    = 1
          L    = V     +  MAX
          HEAD = L     +  MAX
          NEXT = HEAD  +  N
          IF (MAX.LT.N)  GO TO 110
!C
          CALL  MD
     *       (N, IA,JA, MAX,ISP(V),ISP(L), ISP(HEAD),P,IP, ISP(V), FLAG)
          IF (FLAG.NE.0)  GO TO 100
!C
!C----ALLOCATE STORAGE AND SYMMETRICALLY REORDER MATRIX
   1    IF ((PATH-2) * (PATH-3) * (PATH-4) * (PATH-5) .NE. 0)  GO TO 2
          TMP = (NSP+1) -      N
          Q   = TMP     - (IA(N+1)-1)
          IF (Q.LT.1)  GO TO 110
!C
          DFLAG = PATH.EQ.4 .OR. PATH.EQ.5
          CALL ssro
     *       (N,  IP,  IA, JA, A,  ISP(TMP),  ISP(Q),  DFLAG)
!C
   2    RETURN
!C
!C ** ERROR -- ERROR DETECTED IN MD
 100    RETURN
!C ** ERROR -- INSUFFICIENT STORAGE
 110    FLAG = 10*N + 1
        RETURN
!C ** ERROR -- ILLEGAL PATH SPECIFIED
 111    FLAG = 11*N + 1
        RETURN
        END
!C***********************************************************************
!C***********************************************************************
!C  zsro -- SYMMETRIC REORDERING OF SPARSE SYMMETRIC MATRIX
!C***********************************************************************
        SUBROUTINE  zsro
     *     (N, IP, IA,JA,A, Q, R, DFLAG)
!C
!C  DESCRIPTION
!C
!C    THE NONZERO ENTRIES OF THE MATRIX M ARE ASSUMED TO BE STORED
!C    SYMMETRICALLY IN (IA,JA,A) FORMAT (I.E., NOT BOTH M(I,J) AND M(J,I)
!C    ARE STORED IF I NE J).
!C
!C    zsro DOES NOT REARRANGE THE ORDER OF THE ROWS, BUT DOES MOVE
!C    NONZEROES FROM ONE ROW TO ANOTHER TO ENSURE THAT IF M(I,J) WILL BE
!C    IN THE UPPER TRIANGLE OF M WITH RESPECT TO THE NEW ORDERING, THEN
!C    M(I,J) IS STORED IN ROW I (AND THUS M(J,I) IS NOT STORED);  WHEREAS
!C    IF M(I,J) WILL BE IN THE STRICT LOWER TRIANGLE OF M, THEN M(J,I) IS
!C    STORED IN ROW J (AND THUS M(I,J) IS NOT STORED).
!C
!C
!C  ADDITIONAL PARAMETERS
!C
!C    Q     - INTEGER ONE-DIMENSIONAL WORK ARRAY;  DIMENSION = N
!C
!C    R     - INTEGER ONE-DIMENSIONAL WORK ARRAY;  DIMENSION = NUMBER OF
!C            NONZERO ENTRIES IN THE UPPER TRIANGLE OF M
!C
!C    DFLAG - LOGICAL VARIABLE;  IF DFLAG = .TRUE., THEN STORE NONZERO
!C            DIAGONAL ELEMENTS AT THE BEGINNING OF THE ROW
!C
!C-----------------------------------------------------------------------
!C
        INTEGER  IP(1),  IA(1), JA(1),  Q(1), R(1)
        double complex    A(1),  AK
        LOGICAL  DFLAG
        INTEGER  ILAST, K, JAK, JDUMMY, J, I, N, JMAX, JMIN
!C
!C
!C--PHASE 1 -- FIND ROW IN WHICH TO STORE EACH NONZERO
!C----INITIALIZE COUNT OF NONZEROES TO BE STORED IN EACH ROW
        DO 1 I=1,N
  1       Q(I) = 0
!C
!C----FOR EACH NONZERO ELEMENT A(J)
        DO 3 I=1,N
          JMIN = IA(I)
          JMAX = IA(I+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 3
          DO 2 J=JMIN,JMAX
!C
!C--------FIND ROW (=R(J)) AND COLUMN (=JA(J)) IN WHICH TO STORE A(J) ...
            K = JA(J)
            IF (IP(K).LT.IP(I))  JA(J) = I
            IF (IP(K).GE.IP(I))  K = I
            R(J) = K
!C
!C--------... AND INCREMENT COUNT OF NONZEROES (=Q(R(J)) IN THAT ROW
  2         Q(K) = Q(K) + 1
  3       CONTINUE
!C
!C
!C--PHASE 2 -- FIND NEW IA AND PERMUTATION TO APPLY TO (JA,A)
!C----DETERMINE POINTERS TO DELIMIT ROWS IN PERMUTED (JA,A)
        DO 4 I=1,N
          IA(I+1) = IA(I) + Q(I)
  4       Q(I) = IA(I+1)
!C
!C----DETERMINE WHERE EACH (JA(J),A(J)) IS STORED IN PERMUTED (JA,A)
!C----FOR EACH NONZERO ELEMENT (IN REVERSE ORDER)
        ILAST = 0
        JMIN = IA(1)
        JMAX = IA(N+1) - 1
        J = JMAX
        DO 6 JDUMMY=JMIN,JMAX
          I = R(J)
          IF (.NOT.DFLAG .OR. JA(J).NE.I .OR. I.EQ.ILAST)  GO TO 5
!C
!C------IF DFLAG, THEN PUT DIAGONAL NONZERO AT BEGINNING OF ROW
            R(J) = IA(I)
            ILAST = I
            GO TO 6
!C
!C------PUT (OFF-DIAGONAL) NONZERO IN LAST UNUSED LOCATION IN ROW
  5         Q(I) = Q(I) - 1
            R(J) = Q(I)
!C
  6       J = J-1
!C
!C
!C--PHASE 3 -- PERMUTE (JA,A) TO UPPER TRIANGULAR FORM (WRT NEW ORDERING)
        DO 8 J=JMIN,JMAX
  7       IF (R(J).EQ.J)  GO TO 8
            K = R(J)
            R(J) = R(K)
            R(K) = K
            JAK = JA(K)
            JA(K) = JA(J)
            JA(J) = JAK
            AK = A(K)
            A(K) = A(J)
            A(J) = AK
            GO TO 7
  8       CONTINUE
!C
        RETURN
        END
!C***********************************************************************
!C***********************************************************************
!C  csro -- SYMMETRIC REORDERING OF SPARSE SYMMETRIC MATRIX
!C***********************************************************************
        SUBROUTINE  csro
     *     (N, IP, IA,JA,A, Q, R, DFLAG)
!C
!C  DESCRIPTION
!C
!C    THE NONZERO ENTRIES OF THE MATRIX M ARE ASSUMED TO BE STORED
!C    SYMMETRICALLY IN (IA,JA,A) FORMAT (I.E., NOT BOTH M(I,J) AND M(J,I)
!C    ARE STORED IF I NE J).
!C
!C    csro DOES NOT REARRANGE THE ORDER OF THE ROWS, BUT DOES MOVE
!C    NONZEROES FROM ONE ROW TO ANOTHER TO ENSURE THAT IF M(I,J) WILL BE
!C    IN THE UPPER TRIANGLE OF M WITH RESPECT TO THE NEW ORDERING, THEN
!C    M(I,J) IS STORED IN ROW I (AND THUS M(J,I) IS NOT STORED);  WHEREAS
!C    IF M(I,J) WILL BE IN THE STRICT LOWER TRIANGLE OF M, THEN M(J,I) IS
!C    STORED IN ROW J (AND THUS M(I,J) IS NOT STORED).
!C
!C
!C  ADDITIONAL PARAMETERS
!C
!C    Q     - INTEGER ONE-DIMENSIONAL WORK ARRAY;  DIMENSION = N
!C
!C    R     - INTEGER ONE-DIMENSIONAL WORK ARRAY;  DIMENSION = NUMBER OF
!C            NONZERO ENTRIES IN THE UPPER TRIANGLE OF M
!C
!C    DFLAG - LOGICAL VARIABLE;  IF DFLAG = .TRUE., THEN STORE NONZERO
!C            DIAGONAL ELEMENTS AT THE BEGINNING OF THE ROW
!C
!C-----------------------------------------------------------------------
!C
        INTEGER  IP(1),  IA(1), JA(1),  Q(1), R(1)
        complex*16    A(1),  AK
        LOGICAL  DFLAG
        INTEGER  ILAST, K, JAK, JDUMMY, J, I, N, JMAX, JMIN
!C
!C
!C--PHASE 1 -- FIND ROW IN WHICH TO STORE EACH NONZERO
!C----INITIALIZE COUNT OF NONZEROES TO BE STORED IN EACH ROW
        DO 1 I=1,N
  1       Q(I) = 0
!C
!C----FOR EACH NONZERO ELEMENT A(J)
        DO 3 I=1,N
          JMIN = IA(I)
          JMAX = IA(I+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 3
          DO 2 J=JMIN,JMAX
!C
!C--------FIND ROW (=R(J)) AND COLUMN (=JA(J)) IN WHICH TO STORE A(J) ...
            K = JA(J)
            IF (IP(K).LT.IP(I))  JA(J) = I
            IF (IP(K).GE.IP(I))  K = I
            R(J) = K
!C
!C--------... AND INCREMENT COUNT OF NONZEROES (=Q(R(J)) IN THAT ROW
  2         Q(K) = Q(K) + 1
  3       CONTINUE
!C
!C
!C--PHASE 2 -- FIND NEW IA AND PERMUTATION TO APPLY TO (JA,A)
!C----DETERMINE POINTERS TO DELIMIT ROWS IN PERMUTED (JA,A)
        DO 4 I=1,N
          IA(I+1) = IA(I) + Q(I)
  4       Q(I) = IA(I+1)
!C
!C----DETERMINE WHERE EACH (JA(J),A(J)) IS STORED IN PERMUTED (JA,A)
!C----FOR EACH NONZERO ELEMENT (IN REVERSE ORDER)
        ILAST = 0
        JMIN = IA(1)
        JMAX = IA(N+1) - 1
        J = JMAX
        DO 6 JDUMMY=JMIN,JMAX
          I = R(J)
          IF (.NOT.DFLAG .OR. JA(J).NE.I .OR. I.EQ.ILAST)  GO TO 5
!C
!C------IF DFLAG, THEN PUT DIAGONAL NONZERO AT BEGINNING OF ROW
            R(J) = IA(I)
            ILAST = I
            GO TO 6
!C
!C------PUT (OFF-DIAGONAL) NONZERO IN LAST UNUSED LOCATION IN ROW
  5         Q(I) = Q(I) - 1
            R(J) = Q(I)
!C
  6       J = J-1
!C
!C
!C--PHASE 3 -- PERMUTE (JA,A) TO UPPER TRIANGULAR FORM (WRT NEW ORDERING)
        DO 8 J=JMIN,JMAX
  7       IF (R(J).EQ.J)  GO TO 8
            K = R(J)
            R(J) = R(K)
            R(K) = K
            JAK = JA(K)
            JA(K) = JA(J)
            JA(J) = JAK
            AK = A(K)
            A(K) = A(J)
            A(J) = AK
            GO TO 7
  8       CONTINUE
!C
        RETURN
        END
!C***********************************************************************
!C***********************************************************************
!C  dsro -- SYMMETRIC REORDERING OF SPARSE SYMMETRIC MATRIX
!C***********************************************************************
        SUBROUTINE  dsro
     *     (N, IP, IA,JA,A, Q, R, DFLAG)
!C
!C  DESCRIPTION
!C
!C    THE NONZERO ENTRIES OF THE MATRIX M ARE ASSUMED TO BE STORED
!C    SYMMETRICALLY IN (IA,JA,A) FORMAT (I.E., NOT BOTH M(I,J) AND M(J,I)
!C    ARE STORED IF I NE J).
!C
!C    dsro DOES NOT REARRANGE THE ORDER OF THE ROWS, BUT DOES MOVE
!C    NONZEROES FROM ONE ROW TO ANOTHER TO ENSURE THAT IF M(I,J) WILL BE
!C    IN THE UPPER TRIANGLE OF M WITH RESPECT TO THE NEW ORDERING, THEN
!C    M(I,J) IS STORED IN ROW I (AND THUS M(J,I) IS NOT STORED);  WHEREAS
!C    IF M(I,J) WILL BE IN THE STRICT LOWER TRIANGLE OF M, THEN M(J,I) IS
!C    STORED IN ROW J (AND THUS M(I,J) IS NOT STORED).
!C
!C
!C  ADDITIONAL PARAMETERS
!C
!C    Q     - INTEGER ONE-DIMENSIONAL WORK ARRAY;  DIMENSION = N
!C
!C    R     - INTEGER ONE-DIMENSIONAL WORK ARRAY;  DIMENSION = NUMBER OF
!C            NONZERO ENTRIES IN THE UPPER TRIANGLE OF M
!C
!C    DFLAG - LOGICAL VARIABLE;  IF DFLAG = .TRUE., THEN STORE NONZERO
!C            DIAGONAL ELEMENTS AT THE BEGINNING OF THE ROW
!C
!C-----------------------------------------------------------------------
!C
        INTEGER  IP(1),  IA(1), JA(1),  Q(1), R(1)
        double precision    A(1),  AK
        LOGICAL  DFLAG
        INTEGER  ILAST, K, JAK, JDUMMY, J, I, N, JMAX, JMIN
!C
!C
!C--PHASE 1 -- FIND ROW IN WHICH TO STORE EACH NONZERO
!C----INITIALIZE COUNT OF NONZEROES TO BE STORED IN EACH ROW
        DO 1 I=1,N
  1       Q(I) = 0
!C
!C----FOR EACH NONZERO ELEMENT A(J)
        DO 3 I=1,N
          JMIN = IA(I)
          JMAX = IA(I+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 3
          DO 2 J=JMIN,JMAX
!C
!C--------FIND ROW (=R(J)) AND COLUMN (=JA(J)) IN WHICH TO STORE A(J) ...
            K = JA(J)
            IF (IP(K).LT.IP(I))  JA(J) = I
            IF (IP(K).GE.IP(I))  K = I
            R(J) = K
!C
!C--------... AND INCREMENT COUNT OF NONZEROES (=Q(R(J)) IN THAT ROW
  2         Q(K) = Q(K) + 1
  3       CONTINUE
!C
!C
!C--PHASE 2 -- FIND NEW IA AND PERMUTATION TO APPLY TO (JA,A)
!C----DETERMINE POINTERS TO DELIMIT ROWS IN PERMUTED (JA,A)
        DO 4 I=1,N
          IA(I+1) = IA(I) + Q(I)
  4       Q(I) = IA(I+1)
!C
!C----DETERMINE WHERE EACH (JA(J),A(J)) IS STORED IN PERMUTED (JA,A)
!C----FOR EACH NONZERO ELEMENT (IN REVERSE ORDER)
        ILAST = 0
        JMIN = IA(1)
        JMAX = IA(N+1) - 1
        J = JMAX
        DO 6 JDUMMY=JMIN,JMAX
          I = R(J)
          IF (.NOT.DFLAG .OR. JA(J).NE.I .OR. I.EQ.ILAST)  GO TO 5
!C
!C------IF DFLAG, THEN PUT DIAGONAL NONZERO AT BEGINNING OF ROW
            R(J) = IA(I)
            ILAST = I
            GO TO 6
!C
!C------PUT (OFF-DIAGONAL) NONZERO IN LAST UNUSED LOCATION IN ROW
  5         Q(I) = Q(I) - 1
            R(J) = Q(I)
!C
  6       J = J-1
!C
!C
!C--PHASE 3 -- PERMUTE (JA,A) TO UPPER TRIANGULAR FORM (WRT NEW ORDERING)
        DO 8 J=JMIN,JMAX
  7       IF (R(J).EQ.J)  GO TO 8
            K = R(J)
            R(J) = R(K)
            R(K) = K
            JAK = JA(K)
            JA(K) = JA(J)
            JA(J) = JAK
            AK = A(K)
            A(K) = A(J)
            A(J) = AK
            GO TO 7
  8       CONTINUE
!C
        RETURN
        END
!C***********************************************************************
!C***********************************************************************
!C  ssro -- SYMMETRIC REORDERING OF SPARSE SYMMETRIC MATRIX
!C***********************************************************************
        SUBROUTINE  ssro
     *     (N, IP, IA,JA,A, Q, R, DFLAG)
!C
!C  DESCRIPTION
!C
!C    THE NONZERO ENTRIES OF THE MATRIX M ARE ASSUMED TO BE STORED
!C    SYMMETRICALLY IN (IA,JA,A) FORMAT (I.E., NOT BOTH M(I,J) AND M(J,I)
!C    ARE STORED IF I NE J).
!C
!C    ssro DOES NOT REARRANGE THE ORDER OF THE ROWS, BUT DOES MOVE
!C    NONZEROES FROM ONE ROW TO ANOTHER TO ENSURE THAT IF M(I,J) WILL BE
!C    IN THE UPPER TRIANGLE OF M WITH RESPECT TO THE NEW ORDERING, THEN
!C    M(I,J) IS STORED IN ROW I (AND THUS M(J,I) IS NOT STORED);  WHEREAS
!C    IF M(I,J) WILL BE IN THE STRICT LOWER TRIANGLE OF M, THEN M(J,I) IS
!C    STORED IN ROW J (AND THUS M(I,J) IS NOT STORED).
!C
!C
!C  ADDITIONAL PARAMETERS
!C
!C    Q     - INTEGER ONE-DIMENSIONAL WORK ARRAY;  DIMENSION = N
!C
!C    R     - INTEGER ONE-DIMENSIONAL WORK ARRAY;  DIMENSION = NUMBER OF
!C            NONZERO ENTRIES IN THE UPPER TRIANGLE OF M
!C
!C    DFLAG - LOGICAL VARIABLE;  IF DFLAG = .TRUE., THEN STORE NONZERO
!C            DIAGONAL ELEMENTS AT THE BEGINNING OF THE ROW
!C
!C-----------------------------------------------------------------------
!C
        INTEGER  IP(1),  IA(1), JA(1),  Q(1), R(1)
        real*8    A(1),  AK
        LOGICAL  DFLAG
        INTEGER  ILAST, K, JAK, JDUMMY, J, I, N, JMAX, JMIN
!C
!C
!C--PHASE 1 -- FIND ROW IN WHICH TO STORE EACH NONZERO
!C----INITIALIZE COUNT OF NONZEROES TO BE STORED IN EACH ROW
        DO 1 I=1,N
  1       Q(I) = 0
!C
!C----FOR EACH NONZERO ELEMENT A(J)
        DO 3 I=1,N
          JMIN = IA(I)
          JMAX = IA(I+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 3
          DO 2 J=JMIN,JMAX
!C
!C--------FIND ROW (=R(J)) AND COLUMN (=JA(J)) IN WHICH TO STORE A(J) ...
            K = JA(J)
            IF (IP(K).LT.IP(I))  JA(J) = I
            IF (IP(K).GE.IP(I))  K = I
            R(J) = K
!C
!C--------... AND INCREMENT COUNT OF NONZEROES (=Q(R(J)) IN THAT ROW
  2         Q(K) = Q(K) + 1
  3       CONTINUE
!C
!C
!C--PHASE 2 -- FIND NEW IA AND PERMUTATION TO APPLY TO (JA,A)
!C----DETERMINE POINTERS TO DELIMIT ROWS IN PERMUTED (JA,A)
        DO 4 I=1,N
          IA(I+1) = IA(I) + Q(I)
  4       Q(I) = IA(I+1)
!C
!C----DETERMINE WHERE EACH (JA(J),A(J)) IS STORED IN PERMUTED (JA,A)
!C----FOR EACH NONZERO ELEMENT (IN REVERSE ORDER)
        ILAST = 0
        JMIN = IA(1)
        JMAX = IA(N+1) - 1
        J = JMAX
        DO 6 JDUMMY=JMIN,JMAX
          I = R(J)
          IF (.NOT.DFLAG .OR. JA(J).NE.I .OR. I.EQ.ILAST)  GO TO 5
!C
!C------IF DFLAG, THEN PUT DIAGONAL NONZERO AT BEGINNING OF ROW
            R(J) = IA(I)
            ILAST = I
            GO TO 6
!C
!C------PUT (OFF-DIAGONAL) NONZERO IN LAST UNUSED LOCATION IN ROW
  5         Q(I) = Q(I) - 1
            R(J) = Q(I)
!C
  6       J = J-1
!C
!C
!C--PHASE 3 -- PERMUTE (JA,A) TO UPPER TRIANGULAR FORM (WRT NEW ORDERING)
        DO 8 J=JMIN,JMAX
  7       IF (R(J).EQ.J)  GO TO 8
            K = R(J)
            R(J) = R(K)
            R(K) = K
            JAK = JA(K)
            JA(K) = JA(J)
            JA(J) = JAK
            AK = A(K)
            A(K) = A(J)
            A(J) = AK
            GO TO 7
  8       CONTINUE
!C
        RETURN
        END
!C***********************************************************************
!C***********************************************************************
!C  MD -- MINIMUM DEGREE ALGORITHM (BASED ON ELEMENT MODEL)
!C***********************************************************************
        SUBROUTINE  MD
     *     (N, IA,JA, MAX, V,L, HEAD,LAST,NEXT, MARK, FLAG)
!C
!C  DESCRIPTION
!C
!C    MD FINDS A MINIMUM DEGREE ORDERING OF THE ROWS AND COLUMNS OF A
!C    SYMMETRIC MATRIX M STORED IN (IA,JA,A) FORMAT.
!C
!C
!C  ADDITIONAL PARAMETERS
!C
!C    MAX  - DECLARED DIMENSION OF THE ONE-DIMENSIONAL ARRAYS V AND L;
!C           MAX MUST BE AT LEAST  N+2K,  WHERE K IS THE NUMBER OF
!C           NONZEROES IN THE STRICT UPPER TRIANGLE OF M
!C
!C    V    - INTEGER ONE-DIMENSIONAL WORK ARRAY;  DIMENSION = MAX
!C
!C    L    - INTEGER ONE-DIMENSIONAL WORK ARRAY;  DIMENSION = MAX
!C
!C    HEAD - INTEGER ONE-DIMENSIONAL WORK ARRAY;  DIMENSION = N
!C
!C    LAST - INTEGER ONE-DIMENSIONAL ARRAY USED TO RETURN THE PERMUTATION
!C           OF THE ROWS AND COLUMNS OF M CORRESPONDING TO THE MINIMUM
!C           DEGREE ORDERING;  DIMENSION = N
!C
!C    NEXT - INTEGER ONE-DIMENSIONAL ARRAY USED TO RETURN THE INVERSE OF
!C           THE PERMUTATION RETURNED IN LAST;  DIMENSION = N
!C
!C    MARK - INTEGER ONE-DIMENSIONAL WORK ARRAY (MAY BE THE SAME AS V);
!C           DIMENSION = N
!C
!C    FLAG - INTEGER ERROR FLAG;  VALUES AND THEIR MEANINGS ARE -
!C             0      NO ERRORS DETECTED
!C             11N+1  INSUFFICIENT STORAGE IN MD
!C
!C
!C  DEFINITIONS OF INTERNAL PARAMETERS
!C
!C    ---------+---------------------------------------------------------
!C    V(S)     | VALUE FIELD OF LIST ENTRY
!C    ---------+---------------------------------------------------------
!C    L(S)     | LINK FIELD OF LIST ENTRY  (0 => END OF LIST)
!C    ---------+---------------------------------------------------------
!C    L(VI)    | POINTER TO ELEMENT LIST OF UNELIMINATED VERTEX VI
!C    ---------+---------------------------------------------------------
!C    L(EJ)    | POINTER TO BOUNDARY LIST OF ACTIVE ELEMENT EJ
!C    ---------+---------------------------------------------------------
!C    HEAD(D)  | VJ => VJ HEAD OF D-LIST D
!C             |  0 => NO VERTEX IN D-LIST D
!C
!C
!C             |                  VI UNELIMINATED VERTEX
!C             |          VI IN EK           |       VI NOT IN EK
!C    ---------+-----------------------------+---------------------------
!C    NEXT(VI) | UNDEFINED BUT NONNEGATIVE   | VJ => VJ NEXT IN D-LIST
!C             |                             |  0 => VI TAIL OF D-LIST
!C    ---------+-----------------------------+---------------------------
!C    LAST(VI) | (NOT SET UNTIL MDP)         | -D => VI HEAD OF D-LIST D
!C             |-VK => COMPUTE DEGREE        | VJ => VJ LAST IN D-LIST
!C             | EJ => VI PROTOTYPE OF EJ    |  0 => VI NOT IN ANY D-LIST
!C             |  0 => DO NOT COMPUTE DEGREE |
!C    ---------+-----------------------------+---------------------------
!C    MARK(VI) | MARK(VK)                    | NONNEGATIVE TAG < MARK(VK)
!C
!C
!C             |                   VI ELIMINATED VERTEX
!C             |      EI ACTIVE ELEMENT      |           OTHERWISE
!C    ---------+-----------------------------+---------------------------
!C    NEXT(VI) | -J => VI WAS J-TH VERTEX    | -J => VI WAS J-TH VERTEX
!C             |       TO BE ELIMINATED      |       TO BE ELIMINATED
!C    ---------+-----------------------------+---------------------------
!C    LAST(VI) |  M => SIZE OF EI = M        | UNDEFINED
!C    ---------+-----------------------------+---------------------------
!C    MARK(VI) | -M => OVERLAP COUNT OF EI   | UNDEFINED
!C             |       WITH EK = M           |
!C             | OTHERWISE NONNEGATIVE TAG   |
!C             |       < MARK(VK)            |
!C
!C-----------------------------------------------------------------------
!C
        INTEGER  IA(1), JA(1),  V(1), L(1),  HEAD(1), LAST(1), NEXT(1),
     *     MARK(1),  FLAG,  TAG, DMIN, VK,EK, TAIL
        EQUIVALENCE  (VK,EK)
        INTEGER  K, MAX, N
!C
!C----INITIALIZATION
        TAG = 0
        CALL  MDI
     *     (N, IA,JA, MAX,V,L, HEAD,LAST,NEXT, MARK,TAG, FLAG)
        IF (FLAG.NE.0)  RETURN
!C
        K = 0
        DMIN = 1
!C
!C----WHILE  K < N  DO
   1    IF (K.GE.N)  GO TO 4
!C
!C------SEARCH FOR VERTEX OF MINIMUM DEGREE
   2      IF (HEAD(DMIN).GT.0)  GO TO 3
            DMIN = DMIN + 1
            GO TO 2
!C
!C------REMOVE VERTEX VK OF MINIMUM DEGREE FROM DEGREE LIST
   3      VK = HEAD(DMIN)
          HEAD(DMIN) = NEXT(VK)
          IF (HEAD(DMIN).GT.0)  LAST(HEAD(DMIN)) = -DMIN
!C
!C------NUMBER VERTEX VK, ADJUST TAG, AND TAG VK
          K = K+1
          NEXT(VK) = -K
          LAST(EK) = DMIN - 1
          TAG = TAG + LAST(EK)
          MARK(VK) = TAG
!C
!C------FORM ELEMENT EK FROM UNELIMINATED NEIGHBORS OF VK
          CALL  MDM
     *       (VK,TAIL, V,L, LAST,NEXT, MARK)
!C
!C------PURGE INACTIVE ELEMENTS AND DO MASS ELIMINATION
          CALL  MDP
     *       (K,EK,TAIL, V,L, HEAD,LAST,NEXT, MARK)
!C
!C------UPDATE DEGREES OF UNELIMINATED VERTICES IN EK
          CALL  MDU
     *       (EK,DMIN, V,L, HEAD,LAST,NEXT, MARK)
!C
          GO TO 1
!C
!C----GENERATE INVERSE PERMUTATION FROM PERMUTATION
   4    DO 5 K=1,N
          NEXT(K) = -NEXT(K)
   5      LAST(NEXT(K)) = K
!C
        RETURN
        END
!C
!C***********************************************************************
!C  MDI -- INITIALIZATION
!C***********************************************************************
        SUBROUTINE  MDI
     *     (N, IA,JA, MAX,V,L, HEAD,LAST,NEXT, MARK,TAG, FLAG)
        INTEGER  IA(1), JA(1),  V(1), L(1),  HEAD(1), LAST(1), NEXT(1),
     *     MARK(1), TAG,  FLAG,  SFS, VI,DVI, VJ
        INTEGER  JMAX, J, JMIN, N, MAX
!C
!C----INITIALIZE DEGREES, ELEMENT LISTS, AND DEGREE LISTS
        DO 1 VI=1,N
          MARK(VI) = 1
          L(VI) = 0
   1      HEAD(VI) = 0
        SFS = N+1
!C
!C----CREATE NONZERO STRUCTURE
!C----FOR EACH NONZERO ENTRY A(VI,VJ) IN STRICT UPPER TRIANGLE
        DO 3 VI=1,N
          JMIN = IA(VI)
          JMAX = IA(VI+1) - 1
          IF (JMIN.GT.JMAX)  GO TO 3
          DO 2 J=JMIN,JMAX
            VJ = JA(J)
            IF (VI.GE.VJ)  GO TO 2
              IF (SFS.GE.MAX)  GO TO 101
!C
!C------ENTER VJ IN ELEMENT LIST FOR VI
              MARK(VI) = MARK(VI) + 1
              V(SFS) = VJ
              L(SFS) = L(VI)
              L(VI) = SFS
              SFS = SFS+1
!C
!C------ENTER VI IN ELEMENT LIST FOR VJ
              MARK(VJ) = MARK(VJ) + 1
              V(SFS) = VI
              L(SFS) = L(VJ)
              L(VJ) = SFS
              SFS = SFS+1
   2        CONTINUE
   3      CONTINUE
!C
!C----CREATE DEGREE LISTS AND INITIALIZE MARK VECTOR
        DO 4 VI=1,N
          DVI = MARK(VI)
          NEXT(VI) = HEAD(DVI)
          HEAD(DVI) = VI
          LAST(VI) = -DVI
          IF (NEXT(VI).GT.0)  LAST(NEXT(VI)) = VI
   4      MARK(VI) = TAG
!C
        RETURN
!C
!C ** ERROR -- INSUFFICIENT STORAGE
 101    FLAG = 9*N + VI
        RETURN
        END
!C
!C***********************************************************************
!C  MDM -- FORM ELEMENT FROM UNELIMINATED NEIGHBORS OF VK
!C***********************************************************************
        SUBROUTINE  MDM
     *     (VK,TAIL, V,L, LAST,NEXT, MARK)
        INTEGER  VK, TAIL,  V(1), L(1),   LAST(1), NEXT(1),   MARK(1),
     *     TAG, S,LS,VS,ES, B,LB,VB, BLP,BLPMAX
        EQUIVALENCE  (VS, ES)
!C
!C----INITIALIZE TAG AND LIST OF UNELIMINATED NEIGHBORS
        TAG = MARK(VK)
        TAIL = VK
!C
!C----FOR EACH VERTEX/ELEMENT VS/ES IN ELEMENT LIST OF VK
        LS = L(VK)
   1    S = LS
        IF (S.EQ.0)  GO TO 5
          LS = L(S)
          VS = V(S)
          IF (NEXT(VS).LT.0)  GO TO 2
!C
!C------IF VS IS UNELIMINATED VERTEX, THEN TAG AND APPEND TO LIST OF
!C------UNELIMINATED NEIGHBORS
            MARK(VS) = TAG
            L(TAIL) = S
            TAIL = S
            GO TO 4
!C
!C------IF ES IS ACTIVE ELEMENT, THEN ...
!C--------FOR EACH VERTEX VB IN BOUNDARY LIST OF ELEMENT ES
   2        LB = L(ES)
            BLPMAX = LAST(ES)
            DO 3 BLP=1,BLPMAX
              B = LB
              LB = L(B)
              VB = V(B)
!C
!C----------IF VB IS UNTAGGED VERTEX, THEN TAG AND APPEND TO LIST OF
!C----------UNELIMINATED NEIGHBORS
              IF (MARK(VB).GE.TAG)  GO TO 3
                MARK(VB) = TAG
                L(TAIL) = B
                TAIL = B
   3          CONTINUE
!C
!C--------MARK ES INACTIVE
            MARK(ES) = TAG
!C
   4      GO TO 1
!C
!C----TERMINATE LIST OF UNELIMINATED NEIGHBORS
   5    L(TAIL) = 0
!C
        RETURN
        END
!C
!C***********************************************************************
!C  MDP -- PURGE INACTIVE ELEMENTS AND DO MASS ELIMINATION
!C***********************************************************************
        SUBROUTINE  MDP
     *     (K,EK,TAIL, V,L, HEAD,LAST,NEXT, MARK)
        INTEGER  EK, TAIL,  V(1), L(1),  HEAD(1), LAST(1), NEXT(1),
     *     MARK(1),  TAG, FREE, LI,VI,LVI,EVI, S,LS,ES, ILP,ILPMAX
        INTEGER  I, K
!C
!C----INITIALIZE TAG
        TAG = MARK(EK)
!C
!C----FOR EACH VERTEX VI IN EK
        LI = EK
        ILPMAX = LAST(EK)
        IF (ILPMAX.LE.0)  GO TO 12
        DO 11 ILP=1,ILPMAX
          I = LI
          LI = L(I)
          VI = V(LI)
!C
!C------REMOVE VI FROM DEGREE LIST
          IF (LAST(VI).EQ.0)  GO TO 3
            IF (LAST(VI).GT.0)  GO TO 1
              HEAD(-LAST(VI)) = NEXT(VI)
              GO TO 2
   1          NEXT(LAST(VI)) = NEXT(VI)
   2        IF (NEXT(VI).GT.0)  LAST(NEXT(VI)) = LAST(VI)
!C
!C------REMOVE INACTIVE ITEMS FROM ELEMENT LIST OF VI
   3      LS = VI
   4      S = LS
          LS = L(S)
          IF (LS.EQ.0)  GO TO 6
            ES = V(LS)
            IF (MARK(ES).LT.TAG)  GO TO 5
              FREE = LS
              L(S) = L(LS)
              LS = S
   5        GO TO 4
!C
!C------IF VI IS INTERIOR VERTEX, THEN REMOVE FROM LIST AND ELIMINATE
   6      LVI = L(VI)
          IF (LVI.NE.0)  GO TO 7
            L(I) = L(LI)
            LI = I
!C
            K = K+1
            NEXT(VI) = -K
            LAST(EK) = LAST(EK) - 1
            GO TO 11
!C
!C------ELSE ...
!C--------CLASSIFY VERTEX VI
   7        IF (L(LVI).NE.0)  GO TO 9
              EVI = V(LVI)
              IF (NEXT(EVI).GE.0)  GO TO 9
                IF (MARK(EVI).LT.0)  GO TO 8
!C
!C----------IF VI IS PROTOTYPE VERTEX, THEN MARK AS SUCH, INITIALIZE
!C----------OVERLAP COUNT FOR CORRESPONDING ELEMENT, AND MOVE VI TO END
!C----------OF BOUNDARY LIST
                  LAST(VI) = EVI
                  MARK(EVI) = -1
                  L(TAIL) = LI
                  TAIL = LI
                  L(I) = L(LI)
                  LI = I
                  GO TO 10
!C
!C----------ELSE IF VI IS DUPLICATE VERTEX, THEN MARK AS SUCH AND ADJUST
!C----------OVERLAP COUNT FOR CORRESPONDING ELEMENT
   8              LAST(VI) = 0
                  MARK(EVI) = MARK(EVI) - 1
                  GO TO 10
!C
!C----------ELSE MARK VI TO COMPUTE DEGREE
   9              LAST(VI) = -EK
!C
!C--------INSERT EK IN ELEMENT LIST OF VI
  10        V(FREE) = EK
            L(FREE) = L(VI)
            L(VI) = FREE
  11      CONTINUE
!C
!C----TERMINATE BOUNDARY LIST
  12    L(TAIL) = 0
!C
        RETURN
        END
!C
!C***********************************************************************
!C  MDU -- UPDATE DEGREES OF UNELIMINATED VERTICES IN EK
!C***********************************************************************
        SUBROUTINE  MDU
     *     (EK,DMIN, V,L, HEAD,LAST,NEXT, MARK)
        INTEGER  EK, DMIN,  V(1), L(1),  HEAD(1), LAST(1), NEXT(1),
     *     MARK(1),  TAG, VI,EVI,DVI, S,VS,ES, B,VB, ILP,ILPMAX,
     *     BLP,BLPMAX
        EQUIVALENCE  (VS, ES)
        INTEGER  I
!C
!C----INITIALIZE TAG
        TAG = MARK(EK) - LAST(EK)
!C
!C----FOR EACH VERTEX VI IN EK
        I = EK
        ILPMAX = LAST(EK)
        IF (ILPMAX.LE.0)  GO TO 11
        DO 10 ILP=1,ILPMAX
          I = L(I)
          VI = V(I)
          IF (LAST(VI))  1, 10, 8
!C
!C------IF VI NEITHER PROTOTYPE NOR DUPLICATE VERTEX, THEN MERGE ELEMENTS
!C------TO COMPUTE DEGREE
   1        TAG = TAG + 1
            DVI = LAST(EK)
!C
!C--------FOR EACH VERTEX/ELEMENT VS/ES IN ELEMENT LIST OF VI
            S = L(VI)
   2        S = L(S)
            IF (S.EQ.0)  GO TO 9
              VS = V(S)
              IF (NEXT(VS).LT.0)  GO TO 3
!C
!C----------IF VS IS UNELIMINATED VERTEX, THEN TAG AND ADJUST DEGREE
                MARK(VS) = TAG
                DVI = DVI + 1
                GO TO 5
!C
!C----------IF ES IS ACTIVE ELEMENT, THEN EXPAND
!C------------CHECK FOR OUTMATCHED VERTEX
   3            IF (MARK(ES).LT.0)  GO TO 6
!C
!C------------FOR EACH VERTEX VB IN ES
                B = ES
                BLPMAX = LAST(ES)
                DO 4 BLP=1,BLPMAX
                  B = L(B)
                  VB = V(B)
!C
!C--------------IF VB IS UNTAGGED, THEN TAG AND ADJUST DEGREE
                  IF (MARK(VB).GE.TAG)  GO TO 4
                    MARK(VB) = TAG
                    DVI = DVI + 1
   4              CONTINUE
!C
   5          GO TO 2
!C
!C------ELSE IF VI IS OUTMATCHED VERTEX, THEN ADJUST OVERLAPS BUT DO NOT
!C------COMPUTE DEGREE
   6        LAST(VI) = 0
            MARK(ES) = MARK(ES) - 1
   7        S = L(S)
            IF (S.EQ.0)  GO TO 10
              ES = V(S)
              IF (MARK(ES).LT.0)  MARK(ES) = MARK(ES) - 1
              GO TO 7
!C
!C------ELSE IF VI IS PROTOTYPE VERTEX, THEN CALCULATE DEGREE BY
!C------INCLUSION/EXCLUSION AND RESET OVERLAP COUNT
   8        EVI = LAST(VI)
            DVI = LAST(EK) + LAST(EVI) + MARK(EVI)
            MARK(EVI) = 0
!C
!C------INSERT VI IN APPROPRIATE DEGREE LIST
   9      NEXT(VI) = HEAD(DVI)
          HEAD(DVI) = VI
          LAST(VI) = -DVI
          IF (NEXT(VI).GT.0)  LAST(NEXT(VI)) = VI
          IF (DVI.LT.DMIN)  DMIN = DVI
!C
  10      CONTINUE
!C
  11    RETURN
        END

      
       
       
       